/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5447 $
 * $Date: 2012-04-14 00:24:26 +0200 (Sa, 14. Apr 2012) $
 * $Author: posch $
 * 
 */

// TODO:  output of intermediate results for plugin
// TODO: optional: filename base fuer intermediate LS
// TODO: javadoc for all energy types
package cmdTools.segmentation;

import jargs.gnu.CmdLineParser;

import java.util.Vector;



import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.io.files.ImageReaderMTB;
import de.unihalle.informatik.MiToBo.io.files.ImageWriterMTB;
import de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.ui.LevelSet_Segmentation_NonPDE.EnergyType;
import de.unihalle.informatik.MiToBo.tools.system.UserTime;

/**
 * Commandline interface for levelset segmentation
 * using nonPDE optimization and optionally topology preserving.
 * For details see {@link de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.ui.LevelSet_Segmentation_NonPDE}
 *
 * @author Stefan Posch, Markus Glass
 *
 */


public class LevelSet_Segmentation_NonPDE {

    static final Integer DEFAULT_MAXITER = 100;
    static final Double DEFAULT_MU = 1.0;
    static final Double DEFAULT_LAMBDA_FG = 1.0;
    static final Double DEFAULT_LAMBDA_BG = 1.0;
    static final Double DEFAULT_NU = 0.0;
	static final Double DEFAULT_ALPHA = 0.0;
	static final Double DEFAULT_SIGMA = 20.0;
	static final Double DEFAULT_KAPPA = 0.0;

    // ---- command line interface ----

    private static void printUsage() {
        System.err.println( "Usage: LevelSet_Segmentation_NonPDE " +
                 "[--multi ] " +
                 "[-p preserveTopology] [--preserve preserveTopology] " +

                 "[-x maxIter] [--maxiter maxIter] " +
				 "\n" +
                 "[-i initBinName] [--initBinImg initBinImg] " +
                 "[-l initLabelName] [--initLabelImg initLabelName] " +
                 "[-k invalidImg] [--invalidImg invalidImg] " +
				 "\n" +
				 "[-e energytype] [--energytype --energytype]" +
                 "[-m mu] [--mu mu] " +
                 "[-n nu] [--nu nu] " +
                 "[-f lambdafg] [--lambdafg lambdafg] " +
                 "[-b lambdabg] [--lambdabg lambdabg] " +
                 "[--alpha alpha] " +
                 "[--sigma sigma] " +
                 "[--kappa kappa] " +
				 "\n" +
                 "[-v] [--verbose] " +
                 "[-d debug] [--debug debug] " +
                 "[-s s] [--spacing s] " +
                 "[--ibase basename] " +
                 "[--contourResult --contourResult] " +
				 "\n" +
                 "image resultImage"
            );
    }

    public static void main(String [] args) {
        de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.ui.LevelSet_Segmentation_NonPDE mainOp = null;
	try {
        	mainOp = new de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.ui.LevelSet_Segmentation_NonPDE();
	} catch( ALDOperatorException e) {
		e.printStackTrace();
		System.exit(0);
	}

        // create commandline parser object
        CmdLineParser parser = new CmdLineParser();
        CmdLineParser.Option oMulti = parser.addBooleanOption("multi");
        CmdLineParser.Option oEnergytype = parser.addStringOption('e',"energytype");
        CmdLineParser.Option oLambdaFg = parser.addDoubleOption('f',"lambdafg");
        CmdLineParser.Option oLambdaBg = parser.addDoubleOption('b',"lambdabg");
        CmdLineParser.Option oMu = parser.addDoubleOption('m',"mu");
        CmdLineParser.Option oNu = parser.addDoubleOption('n',"nu");
        CmdLineParser.Option oAlpha = parser.addDoubleOption("alpha");
        CmdLineParser.Option oSigma = parser.addDoubleOption("sigma");
        CmdLineParser.Option oKappa = parser.addDoubleOption("kappa");
        CmdLineParser.Option oMaxIter = parser.addIntegerOption('x',"maxiter");
        CmdLineParser.Option oSpacing = parser.addIntegerOption('s',"spacing");
        CmdLineParser.Option oBasename = parser.addStringOption("ibase");
        CmdLineParser.Option oContourResult = parser.addStringOption("contourResult");
        CmdLineParser.Option oInitBinName = parser.addStringOption('i',"initBinImg");
        CmdLineParser.Option oInitLabelName = parser.addStringOption('l',"initLabelImg");
        CmdLineParser.Option oInvalidName = parser.addStringOption('k',"invalidImg");
		CmdLineParser.Option oPreserve = parser.addBooleanOption('p',"preserve");
        CmdLineParser.Option oVerbose = parser.addBooleanOption('v',"verbose");
        CmdLineParser.Option oDebug = parser.addIntegerOption('d',"debug");

    
        // parse arguments
        try {
          parser.parse(args);
        }
        catch ( CmdLineParser.OptionException e ) {
            System.err.println(e.getMessage());
            printUsage();
            System.exit(2);
        }

        // remaining arguments ok?
        String[] otherArgs = parser.getRemainingArgs();
        if (otherArgs.length != 2) {
            printUsage();
            System.exit(2);
        }

        // retrieve options
        mainOp.setMultiphase( (Boolean)parser.getOptionValue( oMulti, java.lang.Boolean.FALSE));
        String energytype = (String)parser.getOptionValue( oEnergytype, "chanvese");
        mainOp.setLambdafg( (Double)parser.getOptionValue( oLambdaFg, new Double( DEFAULT_LAMBDA_FG)));
        mainOp.setLambdabg( (Double)parser.getOptionValue( oLambdaBg, new Double( DEFAULT_LAMBDA_BG)));
        mainOp.setMu( (Double)parser.getOptionValue( oMu, new Double( DEFAULT_MU)));
        mainOp.setNu( (Double)parser.getOptionValue( oNu, new Double( DEFAULT_NU)));
        mainOp.setAlpha( (Double)parser.getOptionValue( oAlpha, new Double( DEFAULT_ALPHA)));
        mainOp.setSigma( (Double)parser.getOptionValue( oSigma, new Double( DEFAULT_SIGMA)));
        mainOp.setKappa( (Double)parser.getOptionValue( oKappa, new Double( DEFAULT_KAPPA)));
        mainOp.setMaxIter( (Integer)parser.getOptionValue( oMaxIter, new Integer( DEFAULT_MAXITER)));
        mainOp.setSpacing( (Integer)parser.getOptionValue( oSpacing, new Integer( 0)));
		mainOp.setPreserveTopology( (Boolean)parser.getOptionValue( oPreserve, java.lang.Boolean.FALSE));
        Boolean verbose = (Boolean)parser.getOptionValue( oVerbose, java.lang.Boolean.FALSE);
        Integer debug = (Integer)parser.getOptionValue( oDebug, new Integer( 0));

        String contourResultName = (String)parser.getOptionValue( oContourResult, "contourImg.tif");

		try {
			ImageReaderMTB reader = new ImageReaderMTB( otherArgs[0]);
			reader.runOp(null);
			mainOp.setInImg( reader.getResultMTBImage()); // may we sould check correct type of image ???
	
	        // read initial binary image
	        String initBinName = (String)parser.getOptionValue( oInitBinName);
	        MTBImage initBinImg;
	        if ( initBinName != null ) {
				reader = new ImageReaderMTB( initBinName);
				reader.runOp(null);
				initBinImg = reader.getResultMTBImage(); // may we sould check correct type of image ???

	            // TODO: check correct size
	        } else {
	            initBinImg = null;
	        }
			mainOp.setInitBinImg( initBinImg);
	
	
	        // read initial label image
	        String initLabelName = (String)parser.getOptionValue( oInitLabelName);
	        MTBImage initLabelImg;
	        if ( initLabelName != null ) {

				reader = new ImageReaderMTB( initLabelName);
				reader.runOp(null);
	            initLabelImg = reader.getResultMTBImage(); // may we sould check correct type of image ???

	            // TODO: check correct size
	        } else {
	            initLabelImg = null;
	        }
			mainOp.setInitLabelImg( initLabelImg);
	
	        // read invalid pixel image
	        String invalidName = (String)parser.getOptionValue( oInvalidName);
	        MTBImage invalidImg;
	        if ( invalidName != null ) {
				reader = new ImageReaderMTB( invalidName);
				reader.runOp(null);
	            invalidImg = reader.getResultMTBImage(); // may we sould check correct type of image ???

	            // TODO: check correct size
	        } else {
	            invalidImg = null;
	        }
			mainOp.setInvalidImg( invalidImg);
	
System.out.println( "ENERGY: " + energytype);
			if ( energytype.equalsIgnoreCase( "chanvese") ||
			     energytype.equalsIgnoreCase( "chan-vese") ||
			     energytype.equalsIgnoreCase( "chan_vese")    )
				mainOp.setEnergytype( EnergyType.CHANVESE);
			else if ( energytype.equalsIgnoreCase( "chanvesefitting") ||
			     energytype.equalsIgnoreCase( "chan-vese-fitting") ||
			     energytype.equalsIgnoreCase( "CV-fitting")    )
				mainOp.setEnergytype( EnergyType.CV_FITTING);
			else if ( energytype.equalsIgnoreCase( "chanvese-gaussian") ||
			     energytype.equalsIgnoreCase( "chan-vese-gaussian") ||
			     energytype.equalsIgnoreCase( "chan_vese_gaussian")    )
				mainOp.setEnergytype( EnergyType.CHANVESE_GAUSSIAN);
			else if ( energytype.equalsIgnoreCase( "meijering") )
				mainOp.setEnergytype( EnergyType.MEIJERING);
			else if ( energytype.equalsIgnoreCase( "meanseparation") )
				mainOp.setEnergytype( EnergyType.MEANSEPARATION);
			else if ( energytype.equalsIgnoreCase( "curvature") )
				mainOp.setEnergytype( EnergyType.CURVATURE);
			else if ( energytype.equalsIgnoreCase( "cv-fitting-curvature") ) 
				mainOp.setEnergytype( EnergyType.CV_FITTING_CURVATURE);
			else {
        		System.err.println( "Unknown energytype: " + energytype);
            	printUsage();
            	System.exit(2);
			}

			mainOp.setVerbose( verbose);
			mainOp.setDebug( debug);
	
        	if ( verbose ) mainOp.print();
			// and do the work

			UserTime timer = null;
			if ( verbose ) timer = new UserTime();
			
        	mainOp.runOp( null);

			if ( verbose ) {
				System.out.println( "Solver did " + mainOp.getNumIterations() + " iterations");
				System.out.println( "User time (" + timer.getOperation() + "): " + timer.getElapsedTime());
			}

        	// write result image
        	MTBImage resultImg = mainOp.getResultImage();
        
        	//ImageSave writer= new ImageSave( otherArgs[1],resultImg, verbose, true);
        	//writer.runOp(null);

			ImageWriterMTB writer = new ImageWriterMTB( resultImg, otherArgs[1]);
			writer.setOverwrite( true);
			writer.runOp(null);


			// write optional intermediate levelset functions
        	if ( mainOp.getSpacing() > 0 ) {
            	String base = (String)parser.getOptionValue( oBasename, new String( "intermediateLS"));
            	Vector<MTBImage> intermediateLS = (Vector<MTBImage>)(mainOp.getIntermediateLS());

            	for ( int i = 0 ; i < intermediateLS.size() ; i++ ) {

					writer = new ImageWriterMTB( intermediateLS.elementAt(i), 
									base + "-" + String.format( "%04d", i) + ".tif");
					writer.setOverwrite( true);
					writer.runOp(null);
            	}
        	}

			// write optional contour image
			writer = new ImageWriterMTB( mainOp.getResultContourImage(), contourResultName);
			writer.setOverwrite( true);
        	writer.runOp( null);
		} catch (Exception e) {
			e.printStackTrace();
		}

    }

}
