/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5473 $
 * $Date: 2012-04-18 16:13:33 +0200 (Mi, 18. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio;

import  de.unihalle.informatik.Alida.annotations.ALDDataIOProvider;
import de.unihalle.informatik.Alida.dataio.ALDDataIOManagerCmdline;
import de.unihalle.informatik.Alida.dataio.provider.cmdline.ALDDataIOCmdline;
import de.unihalle.informatik.Alida.dataio.provider.swing.ALDDataIOSwing;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOManagerException;


import java.lang.reflect.Field;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Collection;
import java.lang.reflect.Method;

import net.java.sezpoz.Index;
import net.java.sezpoz.IndexItem;

/**
 * This class implements a DataIO Mangager for reading/writing from command line.
 * For reading and writing, it essentially looks up the correct provider for command line execution 
 * using the method of its super class and invokes its method.
 * <p>
 * It does its work in collaboration with {@link de.unihalle.informatik.Alida.dataio.provider.cmdline.ALDDataIOCmdline}.
 * 
 * @author posch
 *
 */


public class ALDDataIOManagerCmdline extends ALDDataIOManager {

	/** The singleton instance of this class
	 */
	static final ALDDataIOManagerCmdline instance;

	static {
		instance = new ALDDataIOManagerCmdline();
	}

	/** private constructor 
	 */
	private ALDDataIOManagerCmdline() {
		mapTable = initMapTable(ALDDataIOCmdline.class);
	}

	/** Return the single instance of this class
	 * @return single instance
	 */
	public static ALDDataIOManagerCmdline getInstance() {
		return instance;
	}

	/**
	 * Reads data of given class from a specified source.
     * If both <code>field</code> and <code>cl</code> are non-null, the class defined in <code>field</code> is used
     * and <code>cl</code> ignored.
     * 
     * If one of <code>field or</code> <code>cl</code> is null, the other non null argument will be used.
     * Some objects can only be read if <code>field</code> is supplied, e.g. Collections.
	 * <p>
	 * The <code>valueString</code> is used to actualy read the data. The interpretation is
	 * specific to the class to be read and defined by the corresponding provider class.
	 * As a convention, if <code>valueString</code> starts with <code>FILEIO_CHAR</code>  
     * (see {@link de.unihalle.informatik.Alida.dataio.provider.cmdline.ALDStandardizedDataIOCmdline})  data are to
	 * be read from file, where the remaining value of <code>valueString</code> defines the filename.
	 * Otherwise <code>valueString</code> is directly parsed for the value.
	 * 
	 * @param field	field of object to be returned.
	 * @param cl	Class of data to be read.
	 * @param valueString	Source or value string to read data.
	 * @return	Read data object of null if no provider was found.
	 * @throws ALDDataIOManagerException 
	 */
	public Object readData(Field field, Class<?> cl, String valueString) throws ALDDataIOManagerException {
		if ( field != null )
			cl = field.getType();

      	ALDDataIOCmdline provider = (ALDDataIOCmdline)getProvider( cl, ALDDataIOCmdline.class);
		if ( provider != null )
      		return provider.readData( field, cl, valueString);
		else
			return null;
	}
	
	/**
	 * Writes data to the specified location.
     * This method may return a String with a representation of parameters value
     * or may choose to write the value itself, e.g. to a file.
     * The latter will by convention be true, if <code>locatationString</code> starts with <code>cmdlineFILEIO_CHAR</code>  
     * (see {@link de.unihalle.informatik.Alida.dataio.provider.cmdline}).
	 * 
	 * @param obj	Object to write.
	 * @param locatationString	String indicated whether to return the value of where to write the value to.
	 * @return  String formated from obj
	 * @throws ALDDataIOManagerException 
	 */
	public String writeData(Object obj, String locatationString) throws ALDDataIOManagerException {
		Class cl = obj.getClass();
      	ALDDataIOCmdline provider = (ALDDataIOCmdline)getProvider( cl, ALDDataIOCmdline.class);
		if ( provider != null )
      		return provider.writeData(obj, locatationString);
		else
			return null;
	}
}
