/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5413 $
 * $Date: 2012-04-12 11:01:03 +0200 (Do, 12. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio;

import de.unihalle.informatik.Alida.dataio.provider.swing.ALDDataIOSwing;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOManagerException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOManagerException.ALDDataIOManagerExceptionType;


import java.lang.reflect.Field;

import javax.swing.*;

/**
 * This class implements a DataIO Mangager for reading/writing from a graphical
 * user interface.
 * <p>
 * For IO, it essentially looks up the correct provider for command line execution 
 * using the method of its super class and invokes its method.
 * <p>
 * It does its work in collaboration with {@link de.unihalle.informatik.Alida.dataio.provider.swing.ALDDataIOSwing}.
 * <p>
 * 
 * @author posch
 *
 */

public class ALDDataIOManagerSwing extends ALDDataIOManager {

    /** The singleton instance of this class
     */
    static final ALDDataIOManagerSwing instance;

    static {
        instance = new ALDDataIOManagerSwing();
    }

    /** private constructor 
     */
    private ALDDataIOManagerSwing() {
        this.mapTable = initMapTable(ALDDataIOSwing.class);
    }

    /** Return the single instance of this class
     * @return single instance
     */
    public static ALDDataIOManagerSwing getInstance() {
        return instance;
    }

	
	/** Create a gui element which may be used to request the user to
     * input data of class <code>cl</code>.
	 * This element may later be forwarded to <code>readData</code> to
	 * extract an object representig this data.
	 * Null may be returned if no provider was found.
     * Additionally an object may be supplied which is assumed to be of type <code>cl</code>.   
     * If it is non-null its value may be used to set up an inital value in the gui element.
	 * 
	 * @param cl	Class of data to be read.
	 * @param obj The value of this object to set as inital value
	 * @return	gui element of null if no provider was found
	 */
	public JComponent createGUIElement(Field field, Class<?> cl, Object obj) 
		throws ALDDataIOException {
      	ALDDataIOSwing dataLoader = (ALDDataIOSwing)getProvider( cl, ALDDataIOSwing.class);
		if ( dataLoader != null )
      		return dataLoader.createGUIElement(field, cl, obj);
		return null;
	}
	
	/**
	 * Update the value of a parameter in the GUI component.
	 * 
	 * @param field		Related field of parameter to be updated.
	 * @param cl			Class of object to be handled.
	 * @param guiElement	Corresponding GUI element.
	 * @param value		New value to be set.
	 * @throws ALDDataIOException
	 */
	public void setValue(
			Field field, Class<?> cl, JComponent guiElement, Object value)
	throws ALDDataIOException {
		// get a suitable provider
		ALDDataIOSwing dataProvider = 
			(ALDDataIOSwing)getProvider( cl, ALDDataIOSwing.class);
		// ask the provider to update the value
		if ( dataProvider != null )
			dataProvider.setValue(field, cl, guiElement, value);
	}
	
	/**
	 * Reads data of given class from a specified source.
	 * 
	 * @param cl	Class of data to be read.
	 * @param guiElement	gui element from where to read data.
	 * @return	data object read or null if no provider was found
	 */
	public Object readData(Field field, Class<?> cl, JComponent guiElement) 
		throws ALDDataIOException {
      	ALDDataIOSwing dataLoader = (ALDDataIOSwing)getProvider( cl, ALDDataIOSwing.class);
		
		if ( dataLoader != null )
      		return dataLoader.readData(field, cl,guiElement);
		return null;
	}
	
	/**
	 * Return a gui element displaying the <code>obj</code> or null.
	 * In the latter case the obj may have been shown otherwise, e.g. in an
	 * own window.
	 * 
	 * @param obj	Object to write.
	 * @return gui element created
	 */
	public JComponent writeData(Object obj) throws ALDDataIOException {
		// if object is null, there is nothing to write
		if (obj == null)
			return null;
		Class<? extends Object> cl = obj.getClass();
		ALDDataIOSwing dataWriter = 
			(ALDDataIOSwing)getProvider( cl, ALDDataIOSwing.class);
		if ( dataWriter != null ) 
			return dataWriter.writeData(obj);
		throw new ALDDataIOManagerException(
				ALDDataIOManagerExceptionType.NO_PROVIDER_FOUND,
				"No provider found for " + cl.toString());
	}
}
