/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5377 $
 * $Date: 2012-04-10 10:35:27 +0200 (Di, 10. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.helpers;

import javax.swing.*;

import java.util.*;

import de.unihalle.informatik.Alida.operator.*;


import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDParameterPanel;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;

/**	
 * Panel to configure an operator.
 * 
 * @author Birgit Moeller
 */

public class ALDOperatorParameterPanel extends JPanel {

	/**
	 * Local debug flag, not accessible from outside of the class.
	 */
	private boolean debug = false;
	
	/** 
	 * The operator associated with this panel.
	 */
	protected ALDOperator operator = null;
	
	/**
	 * Flag indicating if panel is used on top level or not.
	 * <p>
	 * Using the panel on top-level refers to usage with operators directly
	 * executed, while non-top-level use refers to usage with operators passed
	 * as parameters to other operators. In the second case no deep validation 
	 * is done, i.e. required parameters are allowed to have null values.
	 */
	private boolean topLevelCall = false;

	/**
	 * List of all parameter descriptors associated with the operator.
	 */
	private LinkedList<ALDOpParameterDescriptor> opParamDescrips;
	
	/**
	 * Sub-panel managing the configuration of required parameters.
	 */
	private ALDParameterPanel panelRequiredParams = null;
	
	/**
	 * Sub-panel managing the configuration of optional parameters.
	 */
	private ALDParameterPanel panelOptionalParams = null;

	/**
	 * Sub-panel managing the configuration of supplemental parameters.
	 */
	private ALDParameterPanel panelSupplementalParams = null;

	/**
	 * Mode for displaying parameters.
	 * <p>
	 * In standard mode the view is restricted to parameters annotated as
	 * standard, in advanced mode all annotated parameters of the operator are
	 * displayed in the GUI.
	 */
	private Parameter.ExpertMode displayMode = Parameter.ExpertMode.STANDARD;

	/**
	 * Default constructor.
	 * 
	 * @param op				Operator associated with this panel.
	 * @param mode			Display mode of the panel.
	 * @param topLevel	If true, a top-level context is assumed.
	 */
	public ALDOperatorParameterPanel(ALDOperator op, Parameter.ExpertMode mode,
			boolean topLevel) {
		super();
		this.operator = op;
		this.displayMode = mode;
		this.topLevelCall = topLevel;
		this.initPanel();
	}

	/** 
	 * Register a descriptor.
	 * <p>
	 * Registration is needed to allow the panel to get aware of the parameter.
	 * E.g., sometimes actions are required for certain parameter types on
	 * initialization, and this can be done within this function.
	 * <p>
	 * This method is intended to be overriden to customize parameter handling.
	 * 
	 * @param paramDescriptors	List to which the descriptor is added.	
	 * @param descr							Descriptor to be registered.
	 */
	protected void registerDescriptor( 
			LinkedList<ALDOpParameterDescriptor> paramDescriptors, 
			ALDOpParameterDescriptor descr) {
		paramDescriptors.add( descr);
	}

	/** 
	 * Get the default for a paramter.
	 * <p>
	 * Usually the default value of a parameter is fetched directly from the 
	 * parameter descriptor itself. However, sometimes default values depend on
	 * the system state or are taken from external devices (database, GUI). 
	 * This method allows to customize the reading of default parameters
	 * and is intended to be overriden by sub-classes.
	 * 
   * @param useRequired			Ignored.
   * @param useSupplemental	Ignored.
   * @param descr						Parameter descriptor in question. 
   * @return	Default value of requested parameter, might be null.
   */
  protected Object getDefaultValue( 
  		@SuppressWarnings("unused") Boolean useRequired, 
			@SuppressWarnings("unused") Boolean useSupplemental, 
			ALDOpParameterDescriptor descr) {
		try {
			return this.operator.getParameter( descr.getName());
		} catch (Exception ex) {
			return null;
		}
	}

	/**
	 * Associate the panel with the given operator and update the GUI.
	 * 
	 * @param op	New operator object to be associated with GUI.
	 */
	public void updateConfigPanel(ALDOperator op) {
		// check if new operator has same class like given one...
		if (!(op.getClass().equals(this.operator.getClass()))) {
			Object[] options = { "OK" };
			JOptionPane.showOptionDialog(null, 
					"You are in turn of updating the panel with\n" +
					"an object of a totally different class than before!\n" +
					"Old class: " + this.operator.getClass() + "\n" +
					"New class: " + op.getClass(), "Warning", JOptionPane.DEFAULT_OPTION,
					JOptionPane.WARNING_MESSAGE, null, options, options[0]);
		}
		this.operator = op;
		this.updatePanel();
	}
	
	/**
	 * Switches view mode of the panel.
	 * @param mode	Desired view mode.
	 */
	public void changeViewMode(Parameter.ExpertMode mode) {
		if (this.panelRequiredParams != null) {
			this.panelRequiredParams.changeViewMode(mode);
		}
		if (this.panelOptionalParams != null) {
			this.panelOptionalParams.changeViewMode(mode);
		}
		if (this.panelSupplementalParams != null) {
			this.panelSupplementalParams.changeViewMode(mode);
		}
	}
	
	/**
	 * Inits the panel according to formerly specified descriptors.
	 */
	private void initPanel() {
		// get operator parameters
		this.opParamDescrips = new LinkedList<ALDOpParameterDescriptor>();
		LinkedList<String> namesSorted = 
			new LinkedList<String>(this.operator.getParameterNames()) ;
		java.util.Collections.sort( namesSorted);
		for ( String pName : namesSorted ) {
			try {
				ALDOpParameterDescriptor descr = 
					this.operator.getParameterDescriptor(pName);
				// get parameter values (...and use as defaults)
//				descr.setDefaultValue(this.operator.getParameter(pName));
				registerDescriptor(this.opParamDescrips,descr);
				descr.setDefaultValue(this.getDefaultValue(null, null, descr));
			} catch (Exception e) {
				System.out.println("ALDOperatorDataIOSwing: " +
				" problems reading parameter descriptor... skipping!");
			}
		}
    BoxLayout ylayout = new BoxLayout(this, BoxLayout.Y_AXIS );
    this.setLayout( ylayout);
    // create panel with required parameters
    this.panelRequiredParams = 
    	this.createInParameterPanel(Boolean.TRUE, Boolean.FALSE);
    if (this.panelRequiredParams != null)
    	this.add(this.panelRequiredParams);
    // create panel with optional parameters
		this.panelOptionalParams =
			this.createInParameterPanel(Boolean.FALSE, Boolean.FALSE);
		if (this.panelOptionalParams != null)
			this.add(this.panelOptionalParams);
    // create panel with required parameters
		this.panelSupplementalParams =
			this.createInParameterPanel(null, Boolean.TRUE);
		if (this.panelSupplementalParams != null)
			this.add(this.panelSupplementalParams);
	}

	/**
	 * Updates panel according to values in formerly specified operator.
	 */
	private void updatePanel() {

		// update values in GUI of all parameter values
		for (ALDOpParameterDescriptor descr : this.opParamDescrips) {
			
			// get parameter value from operator
			Object value = null;
			try {
				value = this.operator.getParameter(descr.getName());
			} catch (ALDOperatorException e) {
				Object[] options = { "OK" };
				JOptionPane.showOptionDialog(null, 
						"Updating parameters in GUI failed! \n" + 
						"Could not read parameter \""+descr.getLabel()+"\", skipping.\n"+
						e.getCommentString(), "Warning", 
						JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
						null, options, options[0]);			
			}

      // ignore OUT parameters
      if (descr.getDirection().equals(Parameter.Direction.OUT))
      	continue;
      	
			// is parameter required?
			if (descr.isRequired()) {
				if (   this.panelRequiredParams != null ){
					try {
						this.panelRequiredParams.setParameter(descr, value);
					}	catch (ALDDataIOException e2) {
						Object[] options = { "OK" };
						JOptionPane.showOptionDialog(null, 
								"Updating GUI failed! \n" + 
								"Could not set parameter \""+descr.getLabel()+"\"\n"+
								e2.getCommentString(), "Warning", 
								JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
								null, options, options[0]);			
					}
				}
				else {
					Object[] options = { "OK" };
					JOptionPane.showOptionDialog(null, 
							"Problem in GUI update! \n" + 
							"Found required parameters, but no panel?!", "Warning", 
							JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
							null, options, options[0]);			
				}	
			}	
			// optional, but not supplemental?
			else if (!descr.getSupplemental().booleanValue()) {
				if (   this.panelOptionalParams != null ){
					try {
						this.panelOptionalParams.setParameter(descr, value);
					} catch (ALDDataIOException e2) {
						Object[] options = { "OK" };
						JOptionPane.showOptionDialog(null, 
								"Updating GUI failed! \n" + 
								"Could not set parameter \""+descr.getLabel()+"\"\n"+
								e2.getCommentString(), "Warning", 
								JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
								null, options, options[0]);			
					}
				}
				else {
					Object[] options = { "OK" };
					JOptionPane.showOptionDialog(null, 
							"Problem in GUI update! \n" + 
							"Found optional parameters, but no panel?!", "Warning", 
							JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
							null, options, options[0]);			
				}	
			}
			// must be supplemental...
			else {
				if (   this.panelSupplementalParams != null ){
					try {
						this.panelSupplementalParams.setParameter(descr, value);
					} catch (ALDDataIOException e2) {
						Object[] options = { "OK" };
						JOptionPane.showOptionDialog(null, 
								"Updating GUI failed! \n" + 
								"Could not set parameter \""+descr.getLabel()+"\"\n"+
								e2.getCommentString(), "Warning", 
								JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
								null, options, options[0]);
					}
				}
				else {
					Object[] options = { "OK" };
					JOptionPane.showOptionDialog(null, 
							"Problem in GUI update! \n" + 
							"Found supplemental parameters, but no panel?!", "Warning", 
							JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
							null, options, options[0]);			
				}
			}
		}
		// update the GUI, i.e. redraw all components
		if (this.panelRequiredParams != null)
			this.panelRequiredParams.updateUI();
		if (this.panelOptionalParams != null)
			this.panelOptionalParams.updateUI();
		if (this.panelSupplementalParams != null)
			this.panelSupplementalParams.updateUI();
	}

	/** 
	 * Create a panel with an entry for each input parameter as requested.
	 * <p>
	 * The latter is determined via <code>useRequired</code> and 
	 * <code>useSupplemental</code>.
	 * @param	useRequired 
	 * 				If non-null only parameters with a required flag are
	 * 				used; if null the required flag is ignored.
	 * @param	useSupplemental 
	 * 				If non-null use only parameters with supplemental flag, 
	 *        otherwise supplemental flag is ignored.
	 * @return Generated parameter panel.
	 */
	private ALDParameterPanel createInParameterPanel( 
  		Boolean useRequired, Boolean useSupplemental) {
		
		// find appropriate parameters
		LinkedList<ALDOpParameterDescriptor> paramDescriptors = 
			new LinkedList<ALDOpParameterDescriptor> ();
		for (ALDOpParameterDescriptor descr: this.opParamDescrips) {
			try {
				if (   (   descr.getDirection() ==  Parameter.Direction.IN 
						||     descr.getDirection() ==  Parameter.Direction.INOUT) 
						&& (   useRequired == null 
								|| descr.isRequired() == useRequired.booleanValue()) 
								&& (   useSupplemental == null 
										|| descr.getSupplemental().booleanValue() == 
											useSupplemental.booleanValue()) ) {
					paramDescriptors.add(descr);
				}
			} catch (Exception e) {
				System.out.println(e.getMessage());
				e.printStackTrace();
			}
		}
			
		// and setup GUI elements
		if ( paramDescriptors.size() > 0 ) {
			String title = (
				(useSupplemental == null || useSupplemental.booleanValue()) ? 
																										"Supplemental " : "") +
				"Parameters " + 
				(useRequired == null ? "" : ( useRequired.booleanValue() ? 
							                                  "(required)" : "(optional)"));
			
			ALDParameterPanel panel = 
				new ALDParameterPanel(paramDescriptors,title,this.topLevelCall);
			panel.changeViewMode(this.displayMode);
			return panel;
		}
		return null;
	}

	/** 
	 * Check if operator is ready to run.
	 * <p>
	 * The operator can be run if all GUI elements of required parameters
	 * return non-null values, and if furtheron the operator's validate()
	 * method returns true.
	 * 
	 * @return True, if operator is properly configured.
	 */
	public boolean validateOperatorParameters() {
		
		boolean valid = true;

		// first check if all GUI elements can deliver proper values.
		if (this.panelRequiredParams != null) {
			this.panelRequiredParams.unmarkAllParameters();
			valid = valid && this.panelRequiredParams.validateParameters();
		}
		if (this.panelOptionalParams != null) {
			this.panelOptionalParams.unmarkAllParameters();
			valid = valid && this.panelOptionalParams.validateParameters();
		}
		if (this.panelSupplementalParams != null) {
			this.panelSupplementalParams.unmarkAllParameters();
			valid = valid && this.panelSupplementalParams.validateParameters();
		}

		// check required operator parameters, but only on top level
		if (valid && this.topLevelCall) {
			for (ALDOpParameterDescriptor descr : this.opParamDescrips) {

				// try to read current parameter value
				@SuppressWarnings("unused")
				Object readValue = null; 
				if (descr.isRequired()) {

					// skip output parameters
					if (descr.getDirection() == Parameter.Direction.OUT)
						continue;
					
					try {
						readValue = this.panelRequiredParams.readParameter(descr);
					} catch (ALDDataIOException e) {
						if (this.panelOptionalParams != null) {
							// try to read from optional parameters panel
							try {
								readValue = this.panelOptionalParams.readParameter(descr);
								//	        		if (readValue != null)
								//	        			paramType = 1; // optional parameter
							} catch (ALDDataIOException e1) {
								if (this.panelSupplementalParams != null) {
									// if still null, try to read from supplemental parameters panel
									try {
										readValue = this.panelSupplementalParams.readParameter(descr);
										//	        				if (readValue != null)
										//	        					paramType = 2; // supplemental parameter
									} catch (ALDDataIOException e2) {
										Object[] options = { "OK" };
										JOptionPane.showOptionDialog(null, 
												"Reading parameter \"" +descr.getLabel()+ "\" failed!" +
												"Reason:\n" +	e2.getCommentString(), 
												"Warning", JOptionPane.DEFAULT_OPTION,
												JOptionPane.WARNING_MESSAGE,
												null, options, options[0]);
									}
								}
							}
						}
					}
				}
			}
			// validate the operator itself
			try {
				this.setOperatorParameters();
				this.operator.validate();
			} catch (ALDOperatorException e) {
				Object[] options = { "OK" };
				JOptionPane.showOptionDialog(null, 
						"Operator validation failed! Reason:\n" + e.getCommentString(), 
						"Warning", JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
						null, options, options[0]);
				return false;
			}
		}
		return valid;
	}

	/** 
	 * Retrieve parameter values from GUI and configure operator accordingly.
	 */
	public void setOperatorParameters() {
//		for ( Map.Entry<ALDOpParameterDescriptor,JComponent> entry : this.componentMap.entrySet() ) {
		for (ALDOpParameterDescriptor descr : this.opParamDescrips) {
//			ALDOpParameterDescriptor descr = entry.getKey();
//			JComponent guiElement = entry.getValue();
			
			// skip output parameters
			if (descr.getDirection().equals(Parameter.Direction.OUT))
				continue;
				
			Object value = null;
			// is parameter required?
			if (descr.isRequired()) {
				if (   this.panelRequiredParams != null ){
					try {
						value = this.panelRequiredParams.readParameter(descr);
					}	catch (ALDDataIOException e2) {
						Object[] options = { "OK" };
						JOptionPane.showOptionDialog(null, 
								"Configuration of operator failed! \n" + 
								"Could not read parameter \""+descr.getLabel()+"\"\n"+
								e2.getCommentString(), "Warning", 
								JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
								null, options, options[0]);			
					}
				}
				else {
					Object[] options = { "OK" };
					JOptionPane.showOptionDialog(null, 
							"Problem in operator configuration! \n" + 
							"Found required parameters, but no panel?!", "Warning", 
							JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
							null, options, options[0]);			
				}	
			}	
			// optional, but not supplemental?
			else if (!descr.getSupplemental().booleanValue()) {
				if (   this.panelOptionalParams != null ){
					try {
						value = this.panelOptionalParams.readParameter(descr);
					} catch (ALDDataIOException e2) {
						Object[] options = { "OK" };
						JOptionPane.showOptionDialog(null, 
								"Configuration of operator failed! \n" + 
								"Could not read parameter \""+descr.getLabel()+"\"\n"+
								e2.getCommentString(), "Warning", 
								JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
								null, options, options[0]);			
					}
				}
				else {
					Object[] options = { "OK" };
					JOptionPane.showOptionDialog(null, 
							"Problem in operator configuration! \n" + 
							"Found optional parameters, but no panel?!", "Warning", 
							JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
							null, options, options[0]);			
				}	
			}
			// must be supplemental...
			else {
				if (   this.panelSupplementalParams != null ){
					try {
						value = this.panelSupplementalParams.readParameter(descr);
					} catch (ALDDataIOException e2) {
						Object[] options = { "OK" };
						JOptionPane.showOptionDialog(null, 
								"Configuration of operator failed! \n" + 
								"Could not read parameter \""+descr.getLabel()+"\"\n"+
								e2.getCommentString(), "Warning", 
								JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
								null, options, options[0]);
					}
				}
				else {
					Object[] options = { "OK" };
					JOptionPane.showOptionDialog(null, 
							"Problem in operator configuration! \n" + 
							"Found supplemental parameters, but no panel?!", "Warning", 
							JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
							null, options, options[0]);			
				}
			}
			if ( this.debug )
				System.out.println( "Setting " + descr.getName() + " to " + value);
			try {
				this.operator.setParameter( descr.getName(), value);
			} catch (Exception ex) {
				System.out.println(ex.getMessage());
				ex.printStackTrace();
			}
		}
	}
}
