/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5532 $
 * $Date: 2012-04-24 19:34:54 +0200 (Di, 24. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.swing;

import javax.swing.*;

import java.awt.FlowLayout;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.IOException;
import java.lang.reflect.Field;
import java.util.*;

import de.unihalle.informatik.Alida.operator.*;
import de.unihalle.informatik.Alida.operator.gui.ALDOperatorControlFrame;
import de.unihalle.informatik.Alida.operator.gui.OnlineHelpDisplayer;


import de.unihalle.informatik.Alida.annotations.ALDDataIOProvider;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDOperatorParameterPanel;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException.ALDDataIOProviderExceptionType;
import de.unihalle.informatik.Alida.helpers.ALDClassInfo;

/**	
 * Data IO provider for operators in GUI.
 * <p>
 * This class provides GUI elements and functionality for configuring 
 * operators in GUI. Note that the class has to operation modes:
 * <ul>
 * <li> invocation from data IO manager:<br>
 * 	in this case a complete frame including buttons is returned
 * <li> call from {@link ALDOperatorControlFrame}:<br>
 * 	in this case just a panel is generated and returned, but no control buttons
 * </ul>
 * While in the first case the default constructor without arguments is used,
 * in the second case the constructor with boolean flag is used, followed by
 * a call of the `getConfigPanel()` method.
 * <p>
 * If you do any changes to this class, please make sure that the panel is 
 * always properly built, independent of how the class object in constructed! 
 * 
 * @author Stefan Posch
 * @author Birgit Moeller
 */
@ALDDataIOProvider

public class ALDOperatorDataIOSwing implements ALDDataIOSwing {

	/**
	 * Local debug flag, not accessible from outside of the class.
	 */
	@SuppressWarnings("unused")
  private boolean debug = false;
	
	/**
	 * Local flag to indicate if object is created on OpRunner top level or not;
	 * flag modules if deep validation is done or not.
	 */
	protected boolean topLevelCall = false;

	/**
	 * Panel to configure sub-classed or abstract operators as parameters. 
	 */
	private OperatorHierarchyConfigPanel subClassHandler = null;
	
	/** 
	 * The operator of this frame, may be abstract!
	 */
	protected ALDOperator op = null;
	
	/**
	 * List of all parameter descriptors associated with operator.
	 */
	protected LinkedList<ALDOpParameterDescriptor> opParamDescrips;
	
	/**
	 * Configure button for operator as parameter.
	 */
	OperatorConfigButton confButton;
	
	/**
	 * Default constructor.
	 */
	public ALDOperatorDataIOSwing() {
		// nothing to do here
	}

	@Override
  public JComponent createGUIElement(Field field, Class<?> cl, Object obj) {
		this.topLevelCall = false;
		try {
			if (obj != null) {
				// we have a default object given...
				if (cl.equals(obj.getClass())) {
					// same class like parameter, just use given object
					this.op = (ALDOperator)obj;
				}
				else {
					// class is different, should be a subclass
					this.op = (ALDOperator)obj;
		    	// instantiating operator failed, maybe an abstract class...
		    	this.subClassHandler = new OperatorHierarchyConfigPanel();
		    	return this.subClassHandler.createGUIElement(field, cl, obj);
				}
			}
			else {
				this.op = (ALDOperator)cl.newInstance();
			}
    } catch (Exception e) {
    	// instantiating operator failed, maybe an abstract class...
    	this.subClassHandler = new OperatorHierarchyConfigPanel();
    	return this.subClassHandler.createGUIElement(field, cl, obj);
    }
    // check if the operator has derived classes
		@SuppressWarnings("rawtypes")
    Collection<Class> extClasses = ALDClassInfo.lookupExtendingClasses(cl);
		if (extClasses.size() > 1) { // class itself is also found...
    	this.subClassHandler = new OperatorHierarchyConfigPanel();
			return this.subClassHandler.createGUIElement(field, cl, obj);
		}
    // return configuration button for this class
	  this.confButton = new OperatorConfigButton();
	  return this.confButton;
  }

	@Override
	public void setValue(Field field, Class<?> cl, JComponent guiElement, 
			Object value) 
	throws ALDDataIOProviderException {
		if (!(value instanceof ALDOperator)) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR,
					this.getClass().getName()+": got invalid object type in setValue()");
		}
		if (guiElement instanceof OperatorConfigButton) {
			OperatorConfigButton confB = (OperatorConfigButton)guiElement;
			this.op = (ALDOperator)value;
			confB.updateOperator(this.op);
		}
		else if (guiElement instanceof 
				OperatorHierarchyConfigPanel.AbstrOpClassPanel) {
			OperatorHierarchyConfigPanel.AbstrOpClassPanel p=
				(OperatorHierarchyConfigPanel.AbstrOpClassPanel)guiElement;
			p.setValue(field, cl, guiElement, value);
			guiElement.updateUI();
		}
	}

	@Override
  public Object readData(Field field, Class<?> cl, JComponent guiElement) 
		throws ALDDataIOProviderException {
		if (guiElement instanceof OperatorConfigButton) {
			OperatorConfigButton confB = (OperatorConfigButton)guiElement;
			return confB.readData(field, cl);
		}
		else if (guiElement instanceof 
				OperatorHierarchyConfigPanel.AbstrOpClassPanel) {
			OperatorHierarchyConfigPanel.AbstrOpClassPanel p=
				(OperatorHierarchyConfigPanel.AbstrOpClassPanel)guiElement;
			return p.readData(field, cl);
		}
		return null;
  }

	@Override
  public JComponent writeData(Object obj) {
	  // TODO Auto-generated method stub
	  return null;
  }

	@Override
	public Collection<Class<?>> providedClasses() {
		LinkedList<Class<?>> classes = new LinkedList<Class<?>>();
		classes.add(ALDOperator.class);
		return classes;
	}

	/**
	 * GUI element for configuring a single operator.
	 * <p>
	 * This button has an operator configuration window attached to it. 
	 * 
	 * @author moeller
	 */
	class OperatorConfigButton extends JButton {

		/**
		 * Operator configuration window.
		 */
		OperatorConfigWindow confWin;

		/**
		 * Constructor.
		 * 
		 * @param field	Field of collection.
		 * @param cl	Class of collection.
		 * @param obj	Default object.
		 */
		public OperatorConfigButton() {
			super("Configure Operator...");
			this.confWin = new OperatorConfigWindow(ALDOperatorDataIOSwing.this.op);
			this.setActionCommand("configButtonPressed");
			this.addActionListener(this.confWin);
			this.setVisible(true);
		}		
		
		/**
		 * Read operator parameter values.
		 * 
		 * @param field
		 * @param cl
		 * @return
		 * @throws ALDDataIOProviderException
		 */
		public Object readData(Field field, Class<?> cl) 
			throws ALDDataIOProviderException {
			return this.confWin.setOperatorParameters(field, cl);
		}

		/**
		 * Update window with new operator parameter values.
		 * 
		 * @param field
		 * @param cl
		 * @return
		 * @throws ALDDataIOProviderException
		 */
		public void updateOperator(ALDOperator op) 
			throws ALDDataIOProviderException {
			this.confWin.updateOperator(op);
		}
	}	

	/**
	 * Configuration window for operator.
	 * 
	 * @author moeller
	 */
	class OperatorConfigWindow implements ActionListener {

		/**
		 * Main configuration window.
		 */
		JFrame window;
		
		/**
		 * Fixed width of window.
		 */
		private final int frameWidth = 500;
		
		/**
		 * Main panel of main frame.
		 */
		private JPanel mainPanel = null;
		
		/**
		 * Sub-panel containing operator parameters.
		 */
		private ALDOperatorParameterPanel configPanel = null;

		/**
		 * Corresponding operator to be configured.
		 */
		private ALDOperator operator;
		
		/**
		 * Mode for displaying parameters.
		 */
		protected Parameter.ExpertMode displayMode = Parameter.ExpertMode.STANDARD;

		/**
		 * Default constructor.
		 * 
		 * @param ops		Corresponding operator to configure.
		 */
		public OperatorConfigWindow(ALDOperator ops) {
			this.operator = ops;
			this.buildWindow();
		}
		
    @SuppressWarnings("unused")
    public ALDOperator setOperatorParameters(Field field, Class<?> cl) {
			this.configPanel.setOperatorParameters();
			return this.operator;
		}
    
    public void updateOperator(ALDOperator oper) {
    	this.operator = oper;
    	this.configPanel.updateConfigPanel(this.operator);
    }
		
		@Override
    public void actionPerformed(ActionEvent e) {
			String cmd = e.getActionCommand(); 
			if (cmd.equals("close")) {
				this.window.setVisible(false);
			}
			else if (cmd.equals("validate") || cmd.equals("actionsM_validate")) {
				this.configPanel.validateOperatorParameters();
			}
			else if (cmd.equals("configButtonPressed")) {
				this.window.setVisible(true);
			}
			else if (cmd.equals("fileM_save")) {
				Object[] options = { "OK" };
				JOptionPane.showOptionDialog(null, 
						"Function will be provided soon...\n", "Warning",
						JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
						null, options, options[0]);
			}
			else if (cmd.equals("fileM_load")) {
				Object[] options = { "OK" };
				JOptionPane.showOptionDialog(null, 
						"Function will be provided soon...\n", "Warning",
						JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
						null, options, options[0]);
			}
			else if (cmd.equals("viewM_standard")) {
				this.configPanel.changeViewMode(Parameter.ExpertMode.STANDARD);
				this.window.repaint();
			}
			else if (cmd.equals("viewM_advanced")) {
				this.configPanel.changeViewMode(Parameter.ExpertMode.ADVANCED);
				this.window.repaint();
			}
			else if (cmd.equals("helpM_about")) {
				Object[] options = { "OK" };
				JOptionPane.showOptionDialog(null, 
						"Alida / MiToBo Graphical Operator Runner - \n" +
						"Chosen Operator: " + this.operator.getName() + "\n" +
						" @2012 Martin Luther University Halle-Wittenberg", 
						"About Alida / MiToBo",
						JOptionPane.DEFAULT_OPTION, JOptionPane.PLAIN_MESSAGE,
						null, options, options[0]);
			}
	  }		
		
		/**
		 * Generates the window.
		 */
		private void buildWindow() {
			this.window = new JFrame();
			this.window.setTitle("Configure operator: " + this.operator.getName());
	//		this.window.setSize(this.frameWidth,300);
			
			this.mainPanel = new JPanel();
			this.window.add(this.mainPanel);
			
			BoxLayout ylayout = new BoxLayout( this.mainPanel, BoxLayout.Y_AXIS );
	    this.mainPanel.setLayout( ylayout);

	    this.configPanel = 
	    	new ALDOperatorParameterPanel(this.operator, this.displayMode, false);
	    JScrollPane scroller = new JScrollPane(this.configPanel);  
//	    this.window.getContentPane().add(scroller, BorderLayout.CENTER);
			this.mainPanel.add(scroller);

			// add buttons
			JPanel buttonPanel = new JPanel(new GridLayout(1,2));
			JPanel valPanel = new JPanel();
			valPanel.setLayout(new BoxLayout(valPanel, BoxLayout.LINE_AXIS));
			valPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
			JButton validateButton = new JButton("Validate");
			validateButton.setActionCommand( "validate");
			validateButton.addActionListener( this);
			validateButton.setBounds(50, 60, 80, 30);
			valPanel.add(validateButton);
			
			JPanel closePanel = new JPanel();
			closePanel.setLayout(new BoxLayout(closePanel, BoxLayout.LINE_AXIS));
			closePanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
			JButton closeButton = new JButton("Close");
			closeButton.setActionCommand("close");
			closeButton.addActionListener(this);
			closePanel.add(closeButton);
//			buttonPanel.add(validateButton);
//			buttonPanel.add(closeButton);
			buttonPanel.add(valPanel);
			buttonPanel.add(closePanel);
			this.mainPanel.add(buttonPanel);
			
			// wrap up with a scrolling panel
//	    JScrollPane scrollPane = new JScrollPane(this.mainPanel);
//	    this.window.add(scrollPane);
			
	    // add a menu bar
			JMenuBar mainWindowMenu = new JMenuBar();
			JMenu fileM = new JMenu("File");
			JMenuItem itemSave = new JMenuItem("Save Settings");
			itemSave.setActionCommand("fileM_save");
			itemSave.addActionListener(this);
			JMenuItem itemLoad = new JMenuItem("Load Settings");
			itemLoad.setActionCommand("fileM_load");
			itemLoad.addActionListener(this);
			fileM.add(itemSave);
			fileM.add(itemLoad);
			
			JMenu actionsM = new JMenu("Actions");
			JMenuItem itemValidate = new JMenuItem("Validate");
			itemValidate.setActionCommand("actionsM_validate");
			itemValidate.addActionListener(this);
			actionsM.add(itemValidate);
			
			JMenu viewM = new JMenu("View");
			ButtonGroup operatorLevelGroup = new ButtonGroup();
			// default
			JRadioButtonMenuItem radioItemStandard = 
				new JRadioButtonMenuItem("Standard");
			if (this.displayMode.equals(Parameter.ExpertMode.STANDARD))
				radioItemStandard.setSelected(true);
			radioItemStandard.setActionCommand("viewM_standard");
			radioItemStandard.addActionListener(this);
			JRadioButtonMenuItem radioItemAdvanced =
				new JRadioButtonMenuItem("Advanced");
			if (this.displayMode.equals(Parameter.ExpertMode.ADVANCED))
				radioItemAdvanced.setSelected(true);
			radioItemAdvanced.setActionCommand("viewM_advanced");
			radioItemAdvanced.addActionListener(this);
			operatorLevelGroup.add(radioItemStandard);
			operatorLevelGroup.add(radioItemAdvanced);
			viewM.add(radioItemStandard);
			viewM.add(radioItemAdvanced);

			// generate help menu
			JMenu helpM = new JMenu("Help");
			JMenuItem itemHelp = new JMenuItem("Online Help");
			itemHelp.addActionListener(OnlineHelpDisplayer.getHelpActionListener(
					itemHelp,this.operator.getClass().getName()));
			JMenuItem itemAbout = new JMenuItem("About Alida");
			itemAbout.setActionCommand("helpM_about");
			itemAbout.addActionListener(this);
			helpM.add(itemHelp);
			helpM.add(itemAbout);

			mainWindowMenu.add(fileM);
			mainWindowMenu.add(actionsM);
			mainWindowMenu.add(viewM);
			mainWindowMenu.add(Box.createHorizontalGlue());
			mainWindowMenu.add(helpM);

			this.window.setJMenuBar(mainWindowMenu);
			this.window.pack();
			// resize window 
			this.window.setSize(this.frameWidth, 400);
		}
	}
	
	/**	
	 * Element for handling operator inheritance hierarchies in configuration.
	 * 
	 * @author Birgit Moeller
	 */
	private class OperatorHierarchyConfigPanel {

		/**
		 * GUI configuration element.
		 */
		private AbstrOpClassPanel paramPanel;

		/**
		 * Default constructor.
		 */
		public OperatorHierarchyConfigPanel() {
			// nothing to do here
		}

		public JComponent createGUIElement(Field field, Class<?> cl, Object obj) {
			this.paramPanel = new AbstrOpClassPanel(field, cl, obj);
			return this.paramPanel;
		}

		public Object readData(Field field, Class<?> cl, JComponent guiElement) 
		throws ALDDataIOProviderException {
			if (!(guiElement instanceof AbstrOpClassPanel))
				throw new ALDDataIOProviderException(
						ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT, 
						"OperatorDataIO: readData received invalid GUI element!");
			AbstrOpClassPanel paramGUIPanel = (AbstrOpClassPanel)guiElement;
			return paramGUIPanel.readData(field, cl);
		}

		public JComponent writeData(Object obj) {
			// TODO Auto-generated method stub
			return null;
		}

		/**
		 * GUI element class for parametrized class parameters.
		 * 
		 * @author moeller
		 */
		class AbstrOpClassPanel extends JPanel 
		implements ActionListener, ItemListener {

			/**
			 * Combobox containing available classes to select.
			 */
			private JComboBox classSelection;

			/**
			 * List of available classes.
			 */
			@SuppressWarnings("rawtypes")
			private Collection<Class> availableClasses;

			/**
			 * Hashmap to remember short class names.
			 */
			@SuppressWarnings("rawtypes")
			private HashMap<String, Class> shortNames = new HashMap<String, Class>();

			/**
			 * Hashmap to store configuration windows for classes.
			 */
			@SuppressWarnings("rawtypes")
			private HashMap<Class, OperatorConfigButton> configWins = 
				new HashMap<Class, OperatorConfigButton>();

			/**
			 * Checkbox to toggle between long and short class names.
			 */
			private JCheckBox longNameCheckBox;

			/**
			 * Default constructor.
			 */
			public AbstrOpClassPanel(Field field, Class<?> cl, Object obj) {
				super();
				this.setLayout(new FlowLayout());

				// checkbox for long names
				this.longNameCheckBox = new JCheckBox("Full names");
				this.longNameCheckBox.setSelected(false);
				this.longNameCheckBox.setActionCommand("changed");
				this.longNameCheckBox.addItemListener(this);

				// combo box for possible classes
				this.availableClasses = ALDClassInfo.lookupExtendingClasses(cl);
				Vector<Object> comboFields = new Vector<Object>();
				for (Class<?> c : this.availableClasses) {
					String shortName = c.getSimpleName();
					comboFields.add(shortName);
					this.shortNames.put(shortName, c);
					// generate configuration window for each class
					this.configWins.put(c, 
							(OperatorConfigButton)new ALDOperatorDataIOSwing().
									createGUIElement(field, c, null));
				}
				this.classSelection = new JComboBox(comboFields);
				if ( obj != null )
					this.classSelection.setSelectedItem( obj);

				// configuration button
				JButton configButton = new JButton("Configure...");
				configButton.setActionCommand("configure");
				configButton.addActionListener(this);

				this.add(this.longNameCheckBox);
				this.add(this.classSelection);
				this.add(configButton);
				
				// init with given value
				try {
					if (obj != null)
						this.setValue(field, cl, null, obj);
        } catch (ALDDataIOProviderException e) {
        	System.err.println(
        			"[ALDOperatorDataIOSwing] Warning! setValue() failed!");
        	return;
        }
			}

			public void setValue(Field field, Class<?> cl, JComponent guiElement, 
					Object value) throws ALDDataIOProviderException {
				// set selected item
				this.classSelection.setSelectedItem(value.getClass().getSimpleName());
				if (this.configWins.get(value.getClass()) == null) {
					throw new ALDDataIOProviderException(
							ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR, 
					"OperatorDataIO: readData found invalid class in GUI?!");
				}
				OperatorConfigButton cButton = this.configWins.get(value.getClass()); 
				cButton.updateOperator((ALDOperator)value);
			}

			/**
			 * Function to read parameter values from GUI. 
			 * 
			 * @param field		Field of object.
			 * @param cl			Class of object.
			 * @return	Current object value.
			 */
			public Object readData(Field field, 
					@SuppressWarnings("unused") Class<?> cl) 
			throws ALDDataIOProviderException {
				// get selected item
				Object item = this.classSelection.getSelectedItem();
				Class<?> selectedClass = null;
				if (item instanceof String)
					selectedClass = this.shortNames.get(item);
				else
					selectedClass = (Class<?>)item;
				if (this.configWins.get(selectedClass) == null) {
					throw new ALDDataIOProviderException(
							ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR, 
							"OperatorDataIO: readData found invalid class in GUI?!");
				}
				return new ALDOperatorDataIOSwing().readData(field, selectedClass, 
						this.configWins.get(selectedClass));
			}

			@Override
			public void actionPerformed(ActionEvent e) {

				String command = e.getActionCommand();
				// pop-up the parametrized class configuration window
				if (command.equals("configure")) {
					// get selected item
					Object item = this.classSelection.getSelectedItem();
					Class<?> selectedClass = null;
					if (item instanceof String)
						selectedClass = this.shortNames.get(item);
					else
						selectedClass = (Class<?>)item;
					this.configWins.get(selectedClass).confWin.window.setVisible(true);
				}
			}

			@Override
			public void itemStateChanged(ItemEvent e) {
				this.classSelection.removeAllItems();

				// check which name convention to use
				if (this.longNameCheckBox.isSelected()) {
					for (@SuppressWarnings("rawtypes") Class c: this.availableClasses) {
						this.classSelection.addItem(c);
					}
				}
				else {
					for (@SuppressWarnings("rawtypes") Class c: this.availableClasses) {
						this.classSelection.addItem(c.getSimpleName());
					}
				}
			}
		}
	}
}
