/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5335 $
 * $Date: 2012-04-03 15:13:31 +0200 (Di, 03. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.dataio.provider.swing;

import de.unihalle.informatik.Alida.helpers.ALDClassInfo;


import de.unihalle.informatik.Alida.annotations.ALDDataIOProvider;
import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDParametrizedClassConfigWindow;
import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDParametrizedClassDummy;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException.*;

import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JPanel;

/**
 * Class for generic handling of parameterized classes in GUI.
 * <p>
 * This class implements the <code>ALDDataIOSwing</code> interface for 
 * parametrized classes. It provides a button linked to a configuration 
 * window that subsumes all parameters of the class to configure.
 * 
 * @author moeller
 */
@ALDDataIOProvider

public class ALDParametrizedClassDataIOSwing implements ALDDataIOSwing {

	/**
	 * GUI configuration element.
	 */
	private ParametrizedClassPanel paramPanel;
	
	/**
	 * Default constructor.
	 */
	public ALDParametrizedClassDataIOSwing() {
		// nothing to do here
	}
	
	/**
	 * Interface method to announce class for which IO is provided for.
	 * 
	 * @return  Collection of classes provided.
	 */
	@Override
	public Collection<Class<?>> providedClasses() {
		LinkedList<Class<?>> classes = new LinkedList<Class<?>>();

		classes.add( ALDParametrizedClassDummy.class);

		return classes;
	}

	@Override
  public JComponent createGUIElement(Field field, Class<?> cl, Object obj) 
		throws ALDDataIOProviderException {
		try {
	    this.paramPanel = new ParametrizedClassPanel(field, cl, obj);
    } catch (ALDDataIOException e) {
    	throw new ALDDataIOProviderException(
    			ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR,
    			"Provider Error: " + e.getCommentString());
    }
		return this.paramPanel;
  }

	@Override
  public void setValue(Field field, Class<?> cl, JComponent guiElement,
      Object value) throws ALDDataIOProviderException {
		if (!(guiElement instanceof ParametrizedClassPanel))
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT, 
					"ParametrizedClassDataIO: setValue() received invalid GUI element!");
		ParametrizedClassPanel paramGUIPanel = (ParametrizedClassPanel)guiElement;
		try {
			paramGUIPanel.setValue(field, cl, value);
		} catch(ALDDataIOException exp) {
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR,
					"setValue() - setting parameter in GUI failed!");
		}
  }
	
	@Override
  public Object readData(Field field, Class<?> cl, JComponent guiElement) 
		throws ALDDataIOProviderException {
		if (!(guiElement instanceof ParametrizedClassPanel))
			throw new ALDDataIOProviderException(
					ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT, 
					"ParametrizedClassDataIO: readData received invalid GUI element!");
		ParametrizedClassPanel paramGUIPanel = (ParametrizedClassPanel)guiElement;
		try {
	    Object obj = paramGUIPanel.readData(field, cl);
	    return obj;
    } catch (ALDDataIOException e) {
    	// if exception came from provider, just throw it to parent
    	if (e instanceof ALDDataIOProviderException)
    		throw (ALDDataIOProviderException)e;
    	// otherwise embed it into our own exception
    	// (... necessary due to method signature)
			throw new ALDDataIOProviderException(
      		ALDDataIOProviderExceptionType.UNSPECIFIED_ERROR, 
      		"Data IO Manager returned an error: \n" +
      		e.getCommentString());
    }
  }

	@Override
  public JComponent writeData(Object obj) {
		return new ParametrizedClassShowButton(obj);
  }

	/**
	 * GUI element class for parametrized class parameters.
	 * 
	 * @author moeller
	 */
	private class ParametrizedClassPanel extends JPanel 
		implements ActionListener, ItemListener {
		
		/**
		 * Combobox containing available classes to select.
		 */
		private JComboBox classSelection;
		
		/**
		 * List of available classes.
		 */
		@SuppressWarnings("rawtypes")
	  private Collection<Class> availableClasses;

		/**
		 * Hashmap to remember short class names.
		 */
		@SuppressWarnings("rawtypes")
	  private HashMap<String, Class> shortNames = new HashMap<String, Class>();
		
		/**
		 * Hashmap to store configuration windows for classes.
		 */
	  @SuppressWarnings("rawtypes")
    private HashMap<Class, ALDParametrizedClassConfigWindow> configWins = 
	  	new HashMap<Class, ALDParametrizedClassConfigWindow>();

		/**
		 * Checkbox to toggle between long and short class names.
		 */
		private JCheckBox longNameCheckBox;
		
		/**
		 * Default constructor.
		 * @throws ALDDataIOException 
		 */
    public ParametrizedClassPanel(@SuppressWarnings("unused") Field field, 
				Class<?> cl, Object obj) throws ALDDataIOException {
			super();
			this.setLayout(new FlowLayout());

			// checkbox for long names
			this.longNameCheckBox = new JCheckBox("Full names");
			this.longNameCheckBox.setSelected(false);
			this.longNameCheckBox.setActionCommand("changed");
			this.longNameCheckBox.addItemListener(this);

			// combo box for possible classes
			this.availableClasses = ALDClassInfo.lookupExtendingClasses(cl);
			// add the class itself
			if (!(this.availableClasses.contains(cl)))
					this.availableClasses.add(cl);
			Vector<Object> comboFields = new Vector<Object>();
			for (Class<?> c : this.availableClasses) {
				String shortName = c.getSimpleName();
				comboFields.add(shortName);
				this.shortNames.put(shortName, c);
				// generate configuration window for each class
				this.configWins.put(c, new ALDParametrizedClassConfigWindow(c));
			}
			this.classSelection = new JComboBox(comboFields);
			// set the default value...
			if ( obj != null ) {
				// check which name convention to use
				this.classSelection.setSelectedItem(obj.getClass().getSimpleName());
				ALDParametrizedClassConfigWindow win =
					this.configWins.get(obj.getClass());
				try {
	        win.setValue(obj);
        } catch (NullPointerException e2) {
        	System.err.println("Attention! Default object has wrong type, " +
        			" class " + obj.getClass() + " not found!!!");
        }
			}

			// configuration button
			JButton confButton = new JButton("Configure...");
			confButton.setActionCommand("configure");
			confButton.addActionListener(this);

			this.add(this.longNameCheckBox);
			this.add(this.classSelection);
			this.add(confButton);
		}

		/**
		 * Function to update values of parametrized class object.
		 * 
		 * @param field
		 * @param cl
		 * @param value
		 * @throws ALDDataIOException 
		 */
		@SuppressWarnings("unused")
    public void setValue(Field field, Class<?> cl, Object value) 
			throws ALDDataIOException {
			Class<?> objClass = value.getClass();
			this.classSelection.setSelectedItem(objClass.getSimpleName());
			ALDParametrizedClassConfigWindow win= this.configWins.get(objClass);
			if (win != null) { 
				win.setValue(value);
			}
			else {
				System.err.println("ALDParametrizedClassDataIOSwing: \n" +
						"Configuration window for requested class does not exist...");
			}
		}
	      
	  /**
	   * Function to read parameter values from GUI. 
	   * 
	   * @param field		Field of object.
	   * @param cl			Class of object.
	   * @return	Current object value.
	   * @throws ALDDataIOException 
	   */
	  public Object readData(Field field,@SuppressWarnings("unused") Class<?> cl) 
	  	throws ALDDataIOException {
	  	// get selected item
	  	Object item = this.classSelection.getSelectedItem();
	  	Class<?> selectedClass = null;
	  	if (item instanceof String)
	  		selectedClass = this.shortNames.get(item);
	  	else
	  		selectedClass = (Class<?>)item;
	  	if (this.configWins.get(selectedClass) == null) {
	  		return null;
	  	}
		  return this.configWins.get(selectedClass).readData(field, selectedClass);
	  }

		@Override
	  public void actionPerformed(ActionEvent e) {
			
			String command = e.getActionCommand();
			// pop-up the parametrized class configuration window
			if (command.equals("configure")) {
				// get selected item
				Object item = this.classSelection.getSelectedItem();
				Class<?> selectedClass = null;
				if (item instanceof String)
					selectedClass = this.shortNames.get(item);
				else
					selectedClass = (Class<?>)item;
				this.configWins.get(selectedClass).setVisible(true);
			}
	  }
		
		@Override
	  public void itemStateChanged(ItemEvent e) {
			this.classSelection.removeAllItems();
			
			// check which name convention to use
			if (this.longNameCheckBox.isSelected()) {
				for (@SuppressWarnings("rawtypes") Class c: this.availableClasses) {
					this.classSelection.addItem(c);
				}
			}
			else {
				for (@SuppressWarnings("rawtypes") Class c: this.availableClasses) {
					this.classSelection.addItem(c.getSimpleName());
				}
			}
		}
	}
	
	/**
	 * Button to visualize parametrized class objects.
	 * 
	 * @author moeller
	 */
	private class ParametrizedClassShowButton extends JButton 
		implements ActionListener {
		
		/**
		 * Associated parameter window.
		 */
		private ALDParametrizedClassConfigWindow win;
		
		public ParametrizedClassShowButton(Object obj) {
			super("Show data...");
			this.setActionCommand("show");
			this.addActionListener(this);
			this.win = new ALDParametrizedClassConfigWindow(obj,true);
		}

		@Override
    public void actionPerformed(ActionEvent e) {
			String command = e.getActionCommand();
			if (command.equals("show")) {
				this.win.setVisible(true);
			}
    }
	}
}
