/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5426 $
 * $Date: 2012-04-12 16:03:28 +0200 (Do, 12. Apr 2012) $
 * $Author: moeller $
 * 
 */
//TODO: do we need unset..... if we  wompletely switch to annotations??

package de.unihalle.informatik.Alida.operator;

import de.unihalle.informatik.Alida.version.ALDVersionProvider;
import de.unihalle.informatik.Alida.version.ALDVersionProviderFactory;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.exceptions.*;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.OperatorExceptionType;


import com.thoughtworks.xstream.*;
import com.thoughtworks.xstream.io.xml.DomDriver;

import org.apache.xmlbeans.*;
import org.graphdrawing.graphml.xmlns.GraphmlDocument;

import java.io.*;
import java.util.*;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.annotation.Annotation;

/** This is the abstract super class for all Alida operators.
  * Alida opertators facilitate automatic logging of the processing historys
  * and automatic generation of user interfaces.
  *<p>
  * The interface consisting of all parameters
  * needs to be define by annotating the corresponding fields (i.e. member variables) 
  * using the {@link Parameter}
  * annotation.
  * The direction of a parameter may be input (IN), output (OUT), or both input and output (INOUT).
  * A Parameter may be declared to be supplemental. Supplemental parapeters, e.g., control 
  * output of debugging information
  * or return intermediate results. The outputs of the operator are expected to
  * be independent of the value of these supplemental values which are not stored in the processing history.
  * A parameter with direction IN or INOUT which is not supplemental may be declared to be required.
  * For OUT parameter and supplemental parameters the required field is ignored.
  * If a parameter of an operator is expected to be documented in the data flow of
  * the processing history, it
  * may be of any Java class where the instances are
  * by references. This excludes only primitive data types, interned strings,
  * and cached numerical objects.
  * If the parameter is not be part of the data flow all classes are acceptable.
  * <p>
  * Values of parameters and inputs have to be set prior to invoking the operator via <code>runOp()</code>
  * and the resulting outputs may be retrieved
  * from the operator after return from <code>runOp()</code>.
  * <p>
  * A default constructor without arguments has to be implemented for all
  * features of code generation and generic execution to be available.
  * For generic execution the operator has to be annotated with {@link de.unihalle.informatik.Alida.annotations.ALDAOperator}
  * with <code>allowGenericExecution</code> set to true.
  * <p>
  * The method <code>operate()</code> implements the processing of this operator
  * and has to be overridden when implementing non-abstract operators. All information
  * passed into and back from the operator are passed via the member variables of the operator
  * which are properly annotated.
  * This method should by no means be used to invoke processing of the operator directly,
  * rather  <code>runOp()</code> is to be used.
  * <p> 
  * The method <code>runOp()</code> is called by the user of an operator to invoke processing.
  * <code>runOp()</code> first checks if all required parameters and inputs are set to non-null values
  * and subsequently check the validity of parameters and inputs
  * as defined by the method <code>validateCustom</code> which has to be overridden when implementing an operator.
  * <p>
  * Refer to the Alida manual for more details and examples.
  * 
  * @author posch
 */


abstract public class ALDOperator {

	// ====================================================
	// MEMBER VARIABLES

	/** Name of the operator
	 */
	public final String name;

	/** Verbose flag to be inherited by all operators.
	 */
	@Parameter( label= "Verbose", supplemental = true, direction = Parameter.Direction.IN, 
                description = "Verbose flag")
	protected Boolean verbose = new Boolean(false);

	/** Does this operator prefer a complete DAG or a data dependency DAG.
	 * This field may be changed when implementing an operator.
	 */
	protected boolean completeDAG = true;

	/**
	 * Access to the central port hash.
	 */
	protected static ALDPortHashAccess portHashAccess = new ALDPortHashAccess();
	
	/** A weak hashmap which contains for each thread a stack of opNodes reflecting
	 * the current state of the method stack.
	 *  Whenever <code>runOp()</code> is called, an opNode is created which is associated with
	 *  this call. It is pushed onto the stack and popped at the end of <code>runOp()</code>.
	 * As the first element a opNode for the dummy operator ALDToplevelOperator
	 * is created which represents outer or top most level of operator calls.
	 */
	static WeakHashMap<Thread,Stack<ALDOpNode>> opNodeStackHash= new WeakHashMap<Thread,Stack<ALDOpNode>>();

	/** Hash representing the interface of this operator. 
      * It contains descriptors for all annotated parameters collected from this operator 
      * and all its super classes.
      */
	private Hashtable<String,ALDOpParameterDescriptor> parameterDescriptorsAll;

	/**
	 *  Instance of a class providing version information.
	 */
	protected ALDVersionProvider versionProvider = 
		ALDVersionProviderFactory.getProviderInstance();
	
	// really only debugging
	private boolean debug = false;

	// ----------------------------------------------------
	// This is currently experimental for debugging purposes

	/** Mode of implicit construction of the processing graph.
	 * Still experimental at this point!!
	 * 0 = regular construction
	 * 1 = do not create opNodes for hidden operator invocations
	 * 2 = prevent implicit construction completely
	 */
	private int constructionMode = 0;


	// END of This is currently experimental for debugging purposes
	// ----------------------------------------------------

	// ====================================================
	// METHODS

	/** This constructor initializes an operator.
	 *  The name is retrieved from the classname and the parameter hash is initialized.
	 */
	protected ALDOperator()  throws ALDOperatorException {
		this.name = this.getClass().getSimpleName();

		this.parameterDescriptorsAll = new Hashtable<String,ALDOpParameterDescriptor>();

		// loop for this class and all super classes over all declared fields to find Annotations
		Class<?> myclass = this.getClass();
		Class currentClass = myclass;
		do {
        
			for ( Field field : myclass.getDeclaredFields() ) {
				String name = field.getName();

				Parameter pAnnotation = field.getAnnotation( Parameter.class);

				// if field is annotated as parameter create and add appropriate descriptor
				// and not already decleared in a class downstream of the inheritance hierachy
				// (as we add the first field for each name we find when walking up the inheritance hierachy
				if ( pAnnotation != null && !fieldContained( name) ) {
					Object defaultValue = null;

					String explanation;
					explanation = pAnnotation.description();

					Parameter.Type	myType;

					String label;
					if ( pAnnotation.label().trim().equals("") ) {
						label = name;
					} else {
						label = pAnnotation.label();
					}


					if ( pAnnotation.direction() != Parameter.Direction.UNKNOWN ) {
						// this is the current version of parameter annotation
						ALDOpParameterDescriptor parameterDescriptor =
							new ALDOpParameterDescriptor( name, 
										pAnnotation.direction(), pAnnotation.supplemental(),
										field.getType(), explanation, label, pAnnotation.required(), 
										defaultValue, field, pAnnotation.dataIOOrder(),
										pAnnotation.mode());

						if ( this.debug ) {
							System.out.println( "   ALDOperator():: adding output parameter <" + name + ">");
						}
						this.parameterDescriptorsAll.put( name, parameterDescriptor);

					} else {
						// this is the previous version of parameter annotation still
						// handled for compatibility
						myType = pAnnotation.type();
	
						if ( myType == Parameter.Type.OUTPUT ) {
							// output
							ALDOpParameterDescriptor parameterDescriptor =
								new ALDOpParameterDescriptor( name, 
										Parameter.Direction.OUT, false,
										field.getType(), explanation, label,
										pAnnotation.required(), defaultValue, field,
										pAnnotation.dataIOOrder(), pAnnotation.mode()); 
							if ( ! allowedClassForIO( field.getType()) ) {
								throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.INVALID_CLASS, 
											"<" + field.getName() + "> of class <" + field.getType().getName() + ">"+
	                                    " in operator " + this.name);
							}
							if ( this.debug ) {
								System.out.println( "   ALDOperator():: adding output parameter <" + name + ">");
							}
							this.parameterDescriptorsAll.put( name, parameterDescriptor);
	
	
						} else if ( myType == Parameter.Type.INPUT ) {
							// input
							ALDOpParameterDescriptor parameterDescriptor =
								new ALDOpParameterDescriptor( name, 
										Parameter.Direction.IN, false,
										field.getType(), explanation, label,
										pAnnotation.required(), defaultValue, field,
										pAnnotation.dataIOOrder(), pAnnotation.mode()); 
							if ( ! allowedClassForIO( field.getType()) ) {
								throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.INVALID_CLASS, 
										"<" + field.getName() + "> of class <" + field.getType().getName() + ">" +
										" in operator " + this.name);
							}
							if ( this.debug ) {
								System.out.println( "   ALDOperator():: adding input parameter <" + name + ">" +
									" of type <" + field.getType() + ">");
							}
							this.parameterDescriptorsAll.put( name, parameterDescriptor);
	
						} else if ( myType == Parameter.Type.PARAMETER ) {
							// parameter
							ALDOpParameterDescriptor parameterDescriptor =
								new ALDOpParameterDescriptor( name, 
										Parameter.Direction.IN, false,
										field.getType(), explanation, label,
										pAnnotation.required(), defaultValue, field,
										pAnnotation.dataIOOrder(), pAnnotation.mode()); 
							if ( this.debug ) {
								System.out.println( "   ALDOperator():: adding parameter parameter <" + name + ">");
							}
							this.parameterDescriptorsAll.put( name, parameterDescriptor);
						} else if ( myType == Parameter.Type.SUPPLEMENTAL ) {
							// supplemental
							ALDOpParameterDescriptor parameterDescriptor =
								new ALDOpParameterDescriptor( name, 
										Parameter.Direction.IN, true,
										field.getType(), explanation, label,
										false, defaultValue, field, pAnnotation.dataIOOrder(),
										pAnnotation.mode()); 
							if ( this.debug ) {
								System.out.println( "   ALDOperator():: adding supplemental parameter <" + name + ">");
							}
							this.parameterDescriptorsAll.put( name, parameterDescriptor);
						}
					}
				}
			} 

			myclass = myclass.getSuperclass();
        } while ( myclass != null );
	}

	/** This method does the actual work and needs to be implemented by every subclass.
	 */

	protected abstract void operate() throws ALDOperatorException,ALDProcessingDAGException;

	/** Get the name of this operator 
	 * @return name of the operator
	 */
	public final String getName() {
		return this.name;
	}

	/** Get the version of this operator 
	 * @return version of the operator
	 */
	public final String getVersion() {
		if (this.versionProvider != null) {
			return this.versionProvider.getVersion();
		}
		return "unknown";
	}

	/**
	 * Get a reference to the port hash access object.
     * @return reference to the port hash access object
	 */
	public static ALDPortHashAccess getALDPortHashAccessKey() {
		return portHashAccess;
	}
	
	// -------------------------------------------------------
	// functions to parameters, their descriptors, and generic getter and setter methods

	/** Get the number of parameters
	 * @return number of parameters
	 */
	public final int getNumParameters() {
		return this.parameterDescriptorsAll.size();
	}

	/** Get the names of in or inout parameters.
     *
	 * @return names of in or inout parameters
	 */
	public final Collection<String> getInInoutNames() {
		return getInInoutNames( null);
	}

	/** Get the names of in or inout parameters.
     *
	 * @param useRequired If true return only required in or inout parameters,
     *		if false return only not required in or inout parameters.
     *      If <code>null</code> then return all in or inout parameters.
	 * @return names of in or inout parameters
	 */
	// TODO could be made more efficent with member variables
	public final Collection<String> getInInoutNames( Boolean useRequired) {
		LinkedList<String> names = new LinkedList<String>();
		for( ALDOpParameterDescriptor descr : parameterDescriptorsAll.values() ){
			if ( (useRequired == null || useRequired == descr.required) &&
                 (! descr.supplemental) &&
			     ( descr.direction == Parameter.Direction.IN ||
				   descr.direction == Parameter.Direction.INOUT ) ) {
				names.add( descr.name);
        	} 
        } 

		return names;
	}

	/** Get the names of out or inout parameters 
	 * @return names of out or inout parameters
	 */
	// TODO could be made more efficent with member variables
	public final Collection<String> getOutInoutNames() {
		LinkedList<String> names = new LinkedList<String>();
		for( ALDOpParameterDescriptor descr : parameterDescriptorsAll.values() ){
			if (( descr.direction == Parameter.Direction.OUT ||
						descr.direction == Parameter.Direction.INOUT ) ) {
				names.add( descr.name);
        	} 
        } 

		return names;
	}

	/** Get the names of supplemental parameters 
	 * @return names of supplemental parameters
	 */
	// TODO could be made more efficent with member variables
	public final Collection<String> getSupplementalNames() {
		LinkedList<String> names = new LinkedList<String>();
		for( ALDOpParameterDescriptor descr : parameterDescriptorsAll.values() ){
			if ( descr.supplemental ) {
				names.add( descr.name);
        	} 
        } 

		return names;
	}

	/** Get the names of all parameters
	 * @return collection of all parameter names
	 */
	public final Collection<String> getParameterNames() {
		return this.parameterDescriptorsAll.keySet();
	}

	/** Get the parameter descriptor for given name.
	 * @param	name	Name of the parameter to get the new value for
	 * @return	descriptor value 
	 */
	public final ALDOpParameterDescriptor getParameterDescriptor( String name) 
		throws ALDOperatorException {
		if ( ! this.parameterDescriptorsAll.containsKey( name) )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);
		return this.parameterDescriptorsAll.get( name).clone();
	}

	/** Get the value of a parameter specified by name. 
	 *
	 * @param	name	Name of the parameter to get the new value for
	 * @return	value of the parameter
	 */
	final public Object getParameter(String name) 
		throws ALDOperatorException {
		if ( ! this.parameterDescriptorsAll.containsKey( name)  )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);

		try {
			Field field = getParameterDescriptor( name).field;
			field.setAccessible(true);
			return field.get( this);
		} catch ( IllegalAccessException e ) {
            e.printStackTrace();
        }
		return null;
	}

	/** Set the value of a parameter specified by name. 
	 *
	 * @param	name	Name of the parameter to set a new value for
	 * @param	value	new value
	 */
	final public void setParameter(String name, Object value) 
		throws ALDOperatorException {
		if ( ! this.parameterDescriptorsAll.containsKey( name)  )
			throw new ALDOperatorException(ALDOperatorException.OperatorExceptionType.UNKOWN_NAME,name);

		try {
			Field field = getParameterDescriptor( name).field;
			field.setAccessible(true);
			field.set( this, value);
		} catch ( IllegalAccessException e ) {
			e.printStackTrace();
		}
	}
	
	//==========================
	// runOp METHOD

	/** A legal method to invoke the operator and handles everything 
	 * necessary to protocol the processing history.
	 * This call has the same effect as <code>runOp( false)</code>, i.e.
	 * runs the operator as not hidden.
	 * 
	 * @param callingOperator  Parent operator which invokes this operator
	 * @see ALDOperator#runOp(boolean)
	 */
	@Deprecated
	public final void runOp( ALDOperator callingOperator) throws ALDOperatorException,ALDProcessingDAGException {
		runOp( false);
	}

	/** A legal method to invoke the operator and handles everything 
     * necessary to protocol the processing history.
     * This call has the same effect as <code>runOp( false)</code>, i.e.
     * runs the operator as not hidden.
	 *
	 * @see ALDOperator#runOp(boolean)
     */
	public final void runOp() throws ALDOperatorException,ALDProcessingDAGException {
		runOp( false);
	}

	/** A legal method to invoke the operator and handles everything 
	 * necessary to protocol the processing history.
	 * Prior to calling <code>runOp</code>, all IN and INOUT parameters have to be set.
	 * When called <code>runOp</code> first validates the parameters.
	 * Validity requires for all operators, that all required IN and INOUT parameters
	 * to have non null values.
	 * In addition the implementation of an operator may impose further constrains
	 * defined by the method <code>validateCustom</code>
	 * which e.g. may restrict the interval of numerical parameters.
	 * <p>
	 * After successful validation the <code>operate</code> method is called to
	 * do the actuall processing of this operator.
	 * Upon return from <code>runOp</code>, resulting output data
	 * may be retrieved from the operator object.
	 * <br>
	 * If <code>hidden</code> is true, the activation of this operator will
	 * not be visible in the processing history.
	 *
	 * @param   hidden    The activation of this operator is hidden in the processing history..
	 */

	public final void runOp( boolean hidden) throws ALDOperatorException,ALDProcessingDAGException {

        if ( getVerbose() || this.debug ) {
            System.out.println( "");
            System.out.println( "Start of runOp for operator " + this.name);
		}

        if ( this.debug ) {
            this.print();
        }
 
		validate();

		// find parent opNode via opNode stack of this thread
		Stack<ALDOpNode>  currentThreadsOpNodeStack = findCurrentThreadsOpNodeStack();
		ALDOpNode parentViaThread = currentThreadsOpNodeStack.peek();

		ALDOpNode opNode = null;
		ALDPort origPorts[] = null;
		Collection<String> inInoutNames = null;

		// documentation in the processing history
		if ( constructionMode == 0  || 
			 ( constructionMode == 1 && 
               ! hidden &&
			   ! parentViaThread.getHidden() ) )  {

			// create an opNode for this operation in the DAG
			opNode = new ALDOpNode( this, hidden);


			// register this opNode to its parent opNode via thread parent
			parentViaThread.addDirectChild( opNode);
			opNode.setParent( parentViaThread);
			if ( this.debug ) {
				System.out.println( "ALDOperator::runOp adding direct registered child " + this + "(" + opNode + ")\n" +
						"                   to opNode " + parentViaThread + "(" + parentViaThread.getName() + ")");
			}

			if ( parentViaThread.getHidden() ) 
				opNode.setHidden( true);

			currentThreadsOpNodeStack.push( opNode);

			// handle inputs
			inInoutNames = getInInoutNames();
			origPorts = new ALDPort[ inInoutNames.size()];

			// first we store the original dataport of inputs
			int i = -1;
			for ( String inputName : inInoutNames ) {
				i++;
				Object input = this.getParameter( inputName);
	
				if ( input != null ) {
	
					// register the input object
					if (!portHashAccess.isRegistered(input)) 
						portHashAccess.register(input);				
	
					// connect each input port of this opNode to the port
					// of the input data the operator was supplied with
					opNode.setInOrigin( i, input, getParameterDescriptor( inputName).explanation);
	
					origPorts[i] = portHashAccess.getHistoryLink(input);
				} else {
					opNode.setInOrigin( i, null, getParameterDescriptor( inputName).explanation);
					origPorts[i] = null;
				}
			}

			// and now we can set the dataports of the data to the input ports
			// (we have to do this in a second for-loop to handle multiple inputs with
			// indentical objects correctly)
			i = -1;
			for ( String inputName : inInoutNames ) {
				i++;
				Object input = this.getParameter( inputName);

				if ( input != null ) {
					// set the output port of the input to the input port of this opNode
					portHashAccess.setHistoryLink(input, opNode.getInputPort( i));
				}
			}

			if ( this.debug ) {
				System.out.println( "");
				System.out.println( "  before operate");
				i = -1;
				for ( String inputName : inInoutNames ) {
					i++;
					System.out.println( "    port of " + i + "-th input " + opNode.getInputPort( i) +
							" with origin " + opNode.getInputPort( i).getOrigin() );
				}
			}
		}

		if ( this.debug )
			System.out.println("Running operate...");

		// do the work
		operate();

		if (this.debug )
			System.out.println("Running operate...done.");

		// documentation in the processing history
		if ( constructionMode == 0  || 
			 ( constructionMode == 1 && 
               ! hidden &&
			   ! parentViaThread.getHidden() ) )  {

			// handle output ports: connect output ports of opNode to the passing output parameters
			Collection<String> outInoutNames = getOutInoutNames();
			int i = -1;
			for ( String outputName : outInoutNames ) {
				i++;
				Object output = this.getParameter( outputName);

				// check if datum has already a port
				if (output != null && !portHashAccess.isRegistered(output)) 
					portHashAccess.register(output);			

				if ( output != null ) {
				
					// register the output object to the history DB
					if (!portHashAccess.isRegistered(output)) 
						portHashAccess.register(output);

					// connect each output port of this opNode to the output port
					// its results (i.e. oututs) originate from
					opNode.setOutOrigin( i, output, getParameterDescriptor( outputName).explanation);

					// store properties of output at the output port
					if (output instanceof ALDData) {
						Enumeration<String> keys = ((ALDData)output).getPropertyKeys();
						while ( keys.hasMoreElements() ) {
							String key = keys.nextElement();
							opNode.getOutputPort( i).setProperty( key, ((ALDData)output).getProperty( key));
						}
					}
				} else {
					opNode.setOutOrigin( i, null, getParameterDescriptor( outputName).explanation);
				}
			}
			// restore port of inputs
			i = -1;
			for ( String inputName : inInoutNames ) {
					i++;
				if ( origPorts[i] != null ) 
					portHashAccess.setHistoryLink(this.getParameter( inputName), origPorts[i]);
			}

			// handle outputs: connected leaving outputs parameters to output ports of this opNode
			i = -1;
			for ( String outputName : outInoutNames ) {
				i++;
				Object output = this.getParameter( outputName);
	
				if ( output != null ) {
					// set the output port of the output (i.e. result) of the operator
					// to the output port of the opNode
					portHashAccess.setHistoryLink(output, opNode.getOutputPort(i));
				}
			}

			if ( getVerbose() || this.debug ) {
				System.out.println( "  End of runOp");
			}

			if ( this.debug ) {
				this.print();
				i = -1;
				for ( String outputName : outInoutNames ) {
					i++;
					System.out.println( "    port for " + i + "-th result " + opNode.getOutputPort( i) +
						" with origin " + opNode.getOutputPort( i).getOrigin() );
					if ( opNode.getOutputPort( i).getOrigin() != null )
						opNode.getOutputPort( i).getOrigin().print();
				}
			}

			currentThreadsOpNodeStack.pop();
		}
	}

	// ===================================================================
	// Helper for runOp()

	/** Find the opNode stack for the current thread.
	 * Create a new one, if not existing
	 * @return opNode stack for the current thread
	 */
	private Stack<ALDOpNode> findCurrentThreadsOpNodeStack() {
        Thread currentThread = Thread.currentThread();
        Stack<ALDOpNode> currentThreadsOpNodeStack = opNodeStackHash.get(currentThread);
        if ( currentThreadsOpNodeStack == null ) {
            try {
                if ( this.debug )
                    System.out.println( "   ALDOperator::runOp create opNodeStack for thread " +
                        currentThread);
                currentThreadsOpNodeStack = new Stack<ALDOpNode>();
                opNodeStackHash.put( currentThread, currentThreadsOpNodeStack);
                currentThreadsOpNodeStack.push( new ALDOpNode( new ALDToplevelOperator(), false));
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

		if ( this.debug ) 
			System.out.println( "   ALDOperator::runOp opNodeStack for thread " + currentThread + " found: " + currentThreadsOpNodeStack);
		return currentThreadsOpNodeStack;
    }

	// ===================================================================

	/** Validates the parameters and inputs of this operator.
	 *  The generic part of this validation checks that each required parameter
	 *  and input is set to a non-null value.
	 *  The custom part of validation is operator depended and implemented
	 * via the method <code>validateCustom</code>.
	 *  An ALDOperatorException is thrown if either validation fails.
 	 */
	public final void validate() throws ALDOperatorException {
		validateGeneric();
		validateCustom();
	}

	/** Generic validation of the in and inout parameters of this operator.
	 *  Validation succeeds if each required parameter
	 *  and input is set to a non-null value or supplies a default value.
	 *  An ALDOperatorException is thrown if validation fails.
 	 */
	public final void validateGeneric() throws ALDOperatorException {
		
		Collection<String> inInoutNames = getInInoutNames();
		int i = -1;
		for ( String inputName : inInoutNames ) {
			i++;
			if ( this.parameterDescriptorsAll.get( inputName).required &&
					getParameter( inputName) == null &&
					getParameterDescriptor( inputName).defaultValue  == null ) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED, 
												"ALDOperator.validateGeneric(): Required input parameter\"" + inputName + "\""+
                                    " is null in operator " + this.name);  
			}            
		}
	}

	/**  Operator specific validation of parameters and inputs.
	 *   Should be overridden when extending an operator to impose specific constraints.
 	 */
	public void validateCustom() throws ALDOperatorException {
		return;
	}

	// ===================================================================
	// GETTER/SETTER for verbose parameter
	/** Set the verbose state of this opertor.
	 * @param	verbose 	New verbose state
	 */
	public void setVerbose( Boolean verbose) throws ALDOperatorException {
		this.verbose = verbose;
	}

	/** Return the verbose state of this opertor.
	 * @return  Current verbose state
	 */
	public Boolean getVerbose() throws ALDOperatorException {
		return this.verbose;
	}

	// ===================================================================
	// METHODS TO HANDLE HISTORY

	/**
	 * Reads and set the history graph of the given object from file.
	 * 
	 * @param obj			Object for which the history is to be read.
	 * @param filename	File from where the history is to be read.
	 * @see ALDPortHashAccess#readHistory(Object,String)
	 */
	public static void readHistory(Object obj, String filename) {
		portHashAccess.readHistory(obj, filename);
	}
	
	/** 
	 * Write the processing history if any to a graphml file.
	 * Equivalent to <code>writeHistory(obj, filename, ALDProcessingDAG.HistoryType.COMPLETE, false)</code>
     *
	 * @param	obj	Object for which the history to write for.
	 * @param	filename	Filename to write the processing history into. 
	 */
	public static void writeHistory(Object obj, String filename) 
		throws ALDProcessingDAGException,ALDOperatorException {
		writeHistory(obj, filename, ALDProcessingDAG.HistoryType.COMPLETE, false);
	}

	/** 
	 * Write the processing history if any to a graphml file.
	 * Equivalent to <code>writeHistory(obj, filename, historyType, false)</code>
     *
	 * @param	obj	Object for which the history to write for.
	 * @param	filename	Filename to write the processing history into. 
	 * @param	historyType	Type/mode of the history.
	 */
	public static void writeHistory(Object obj, String filename, ALDProcessingDAG.HistoryType historyType) 
	throws ALDProcessingDAGException,ALDOperatorException {

		writeHistory(obj, filename, historyType, false);
	}

	/** 
	 * Write the processing history if any to a graphml file.
	 *
	 * @param	obj	Object for which the history to write for.
	 * @param	filename	Filename to write the processing history into. 
	 * 									(as graphml/XML file)
	 * @param	historyType	Type/mode of the history.
	 * @param	ignoreHiding	If true, hiding of opNodes is ignored.
     *
	 * @see ALDPortHashAccess#writeHistory(Object,String,ALDProcessingDAG.HistoryType,boolean)
	 */
	public static void writeHistory(Object obj, String filename,
			ALDProcessingDAG.HistoryType historyType,	boolean ignoreHiding) 
	throws ALDProcessingDAGException,ALDOperatorException {
		portHashAccess.writeHistory(obj, filename, historyType, ignoreHiding);
	}

	// ===================================================================

	/** Write the parameter values to an xml file.
	 * Uses xstream for serialization.
	 *
	 * @param	filename	filename of the xml file to write to
	 */

	public void writeParametersToXml( String filename ) {
		try {
			PrintStream out = new PrintStream( filename);

			XStream xstream = new XStream(new DomDriver());
			String xml = xstream.toXML( new ALDParameterWrapper(this));
	
			out.println( xml);
		} catch ( Exception e) {
			System.err.println(e);
			e.printStackTrace();
		}

	}

	/** Parse the parameter values to an XmlObject.
	 * Uses xstream for serialization.
	 *
	 * @return serialization of the paraeter of this operator
	 */

	public XmlObject parametersToXmlObject() 
		throws XmlException {

		XStream xstream = new XStream(new DomDriver());
		String xml = xstream.toXML( new ALDParameterWrapper(this));
		XmlObject res = XmlObject.Factory.parse( xml);
	
		return res;
	}

	/** Set the parameter values as read from an Xml file.
	 * The class and package name found in the Xml are compared to
	 * the operator to set the parameters to and are required to match.
	 *
	 * @param	filename	filename of the xml file to read from
	 * @return	version string found in Xml file or null if Xml file could not be properly read
	 *				or class or package name check failed
     * @see ALDOperator#setParametersFromXml(String,String,String)
	 */

	public String setParametersFromXml( String filename ) {
		return setParametersFromXml( filename, getClass().getSimpleName(), getClass().getPackage().getName());
	}
		
	/** Set the parameter values as read from an Xml file.
	 * The class and package name found in the Xml are compared to
	 * the <code>className</code> and <code>packageName</code> respectively.
	 * If one of these names is <code>null</code> any name in the Xml file
	 * is accepted.
	 *
	 * @param	filename	filename of the xml file to read from
	 * @param	className	class name expected in the Xml file or null
	 * @param	packageName	package name expected in the Xml file or null
	 * @return	version string found in Xml file of null if Xml file could not be properly read
	 *				or class or package name check failed
	 */

	public String setParametersFromXml( String filename, String className, String packageName ) {
		
		try {
			FileInputStream in = new FileInputStream( filename);

			XStream xstream = new XStream(new DomDriver());

			ALDParameterWrapper pw = (ALDParameterWrapper)(xstream.fromXML( in));

			if ( className != null && ! className.equals( pw.getClassName()) ) {
				System.err.println( "ALDOperator::setParametersFromXml class name does not match, got " + 
					pw.getClassName() + " but expected " + className);
				return null;
			}

			if ( packageName != null && ! packageName.equals( pw.getPackageName()) ) {
				System.err.println( "ALDOperator::setParametersFromXml package name does not match, got " + 
					pw.getPackageName() + " but expected " + packageName);
				return null;
			}

			// we have to set the values directly
            for ( String pName : this.getParameterNames() ) {
				try {
					setParameter( pName, pw.getParameteres().get( pName));
				} catch ( ALDOperatorException e ) {
                       System.err.println( "ALDOperator::setParametersFromXml Error: cannot get value for parameter " +
                                       pName);
                       e.printStackTrace();
                }
			}

			return pw.getVersion();
			
		} catch ( FileNotFoundException e) {
			System.err.println( "ALDOperator::setParametersFromXml cannot open " + filename);
			return null;
		}

	}

	/*
	 * Object serialization.
	 */
	
	/**
	 * Serializes the operator to an XML file.
	 * <p>
	 * Note that all member variables, i.e. the complete state of the object
	 * at the time of serialization is saved. Only transient members are 
	 * ignored.
	 * 
	 * @param filename	File where to save the data.
	 */
	public void serializeToXmlFile( String filename ) {
		try {
			PrintStream out = new PrintStream( filename);
			XStream xstream = new XStream(new DomDriver());
			String xml = xstream.toXML(this);
			out.println( xml);
		} catch ( Exception e) {
			System.err.println(e);
			e.printStackTrace();
		}
	}
	
	/**
	 * Deserializes an operator state from the given file.
	 * <p>
	 * Only transient members are ignored.
	 * 
	 * @param filename	File from where to read the data.
	 */
	public ALDOperator deserializeFromXmlFile( String filename ) {
		try {
			FileInputStream in = new FileInputStream( filename);
			XStream xstream = new XStream(new DomDriver());
			return (ALDOperator)(xstream.fromXML( in));
		} catch ( FileNotFoundException e) {
			System.err.println( "ALDOperator::setParametersFromXml cannot open " + filename);
		}
		return null;
	}

	// ===================================================================

	/** Print some information of the current state this operator to System.out.
	 */

	public void print() {
		print( System.out);
	}
		
	/** Print some information the current state of this operator to <code>outfile</code>.
	 * @param	outfile	Stream to print on
	 */

	public void print( PrintStream outfile) {
		print( outfile, true);
	}

	/** Print information of the interface of this operator to System.out.
	 */

	public void printInterface() {
		print( System.out, false);
	}
		
	/** Print information of the interface of this operator to <code>outfile</code>.
	 * @param	outfile	Stream to print on
	 */

	public void printInterface( PrintStream outfile) {
		print( outfile, false);
	}

	/** Print information of the interface and values of this operator to <code>outfile</code>.
	 * If printValue is true, additional the current value is printed.
	 * This method note intended for public use.
	 * @param	outfile	Stream to print on
	 * @param	printValue
	 */

	protected void print( PrintStream outfile, boolean printValue) {
		outfile.println( "Interface of ALDOperator: " + this.name);

		//for ( String pName : getParameterNames() ) 
		if ( getInInoutNames( true).size() > 0 )
			System.out.println( "Required input parameters");
		for ( String pName : getInInoutNames( true) )
			parameterDescriptorsAll.get( pName).print( outfile, printValue ? this : null);

		if ( getInInoutNames( false).size() > 0 )
			System.out.println( "Optional input parameters");
		for ( String pName : getInInoutNames( false) )
			parameterDescriptorsAll.get( pName).print( outfile, printValue ? this : null);

		if ( getOutInoutNames().size() > 0 )
		System.out.println( "Output parameters");
		for ( String pName : getOutInoutNames() )
			parameterDescriptorsAll.get( pName).print( outfile, printValue ? this : null);

		if ( getSupplementalNames().size() > 0 )
		System.out.println( "Supplemental parameters");
		for ( String pName : getSupplementalNames() )
			parameterDescriptorsAll.get( pName).print( outfile, printValue ? this : null);

	}

	/** Returns a string containing printale information about this operator.
     */
	public String toString() {
		StringBuffer buf = new StringBuffer();

		for ( String pName : getInInoutNames(true) ) {
			try {
				buf.append( pName + "=" + getParameter( pName).toString() + "\n");
			} catch (Exception e) {
			}
		}

		return new String( buf);
	}
	// ================== helper function

	/** Is this class allowed as Input and/or Output?
	 */
	boolean allowedClassForIO( Class currentClass) {
		if ( this.debug ) {
			System.out.println( "       ALDOperator::allowedClassForIO for class " +
				currentClass.getName());
		}

		return ! ( currentClass == int.class || 
				   currentClass == byte.class || 
				   currentClass == short.class || 
				   currentClass == long.class || 
				   currentClass == float.class || 
				   currentClass == double.class || 
				   currentClass == boolean.class || 
				   currentClass == char.class) ;
	}

	/** Does any hashtable contain the key, i.e. this field?
	 */
	private boolean fieldContained( String key) {
		return this.parameterDescriptorsAll.containsKey( key) ;
	}
	

}

