/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5195 $
 * $Date: 2012-03-11 09:24:51 +0100 (So, 11. Mär 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.Alida.operator;

import java.lang.reflect.Field;
import java.io.PrintStream;

import de.unihalle.informatik.Alida.annotations.Parameter;


/** 
 * This class describes generic parameters in Alida used, e.g., in paramterized classes and operators.
 * <p>
 *  The descriptions contains a name, the java class (<code>mylass</code>, and an explanatory string.
 *  The <code>label</code> may be used, e.g. in a GUI, to identify the parameter
 *  instead of using it <code>name</code>.
 *  The <code>required</code> flag may not be interpreted for all parameters.
 *  <p>
 *  The field <code>guiOrder</code> may be use to order of appearance of parameters in user interfaces,
 *  where smaller order indicates early appearance.
 *  Likewise <code>handlingMode</code> may be used to influence GUIs.
 *  <p>
 *  Note: default values are valid only for the (very old) implementation 
 *  without annotations an will disapear in th enear future.
 *  <p>
 *  The field is used for internal means.
 *  <p>
 *  Implementational note: a reference to the Parameter-annotation is NOT included to allow for
 *  wrapping of non-Alida operators as Alida operators.

 * 
 * @author moeller
 * @see ALDOperator
 */

public class ALDParameterDescriptor {

	protected final String name;
	
	protected final Class<?> myclass;
	
	protected final String explanation;
	
	protected final String label;
	
	protected Object defaultValue;

	protected final Field field;

	protected final boolean required;
	
	protected final int dataIOOrder;
	
	protected final Parameter.ExpertMode handlingMode;
	
	/** 
	 * Constructor. 
	 *
	 * @param	name					Name of parameter.
	 * @param	cl				Java class. 
	 * @param	explanation 	Explanatory string.
	 * @param label				Label for parameter.
	 * @param	required			Is this argument required for the operator?
	 * @param	defaultValue 			Default value.
	 * @param	field 				The field for this member.
	 */
	public ALDParameterDescriptor( String name, Class<?> cl, 
			String explanation, String label, boolean required, Object defaultValue, 
			Field field, int guiOrder, Parameter.ExpertMode mode) {
		this.name = name;
		this.myclass = cl;
		this.explanation = explanation;
		this.label = label;
		this.required = required;
		this.defaultValue = defaultValue;
		this.field = field;
		this.dataIOOrder = guiOrder;
		this.handlingMode = mode;
	}

	/**
	 * Returns name of parameter.
	 */
	public String getName() {
  	return this.name;
  }

	/**
	 * Returns class of parameter.
	 */
	public Class<?> getMyclass() {
  	return this.myclass;
  }

	/**
	 * Returns explanation for parameter.
	 */
	public String getExplanation() {
  	return this.explanation;
  }

	/**
	 * Returns label for parameter.
	 */
	public String getLabel() {
  	return this.label;
  }
	
	/**
	 * Returns data I/O order of parameter.
	 */
	public int getDataIOOrder() {
		return this.dataIOOrder;
	}

	/**
	 * Returns mode of parameter handling.
	 */
	public Parameter.ExpertMode getHandlingMode() {
		return this.handlingMode;
	}
	
	/**
	 * Returns is parameter is required.
	 */
	public boolean isRequired() {
  	return this.required;
  }

	/**
	 * Returns default value of parameter.
	 */
	public Object getDefaultValue() {
  	return this.defaultValue;
  }

	/**
	 * Specify default value of parameter.
	 */
	public void setDefaultValue(Object def) {
  	this.defaultValue = def;
  }

	/**
	 * Returns a reference to the field of the parameter.
	 */
	public Field getField() {
  	return this.field;
  }

	/** 
	 * Clones an instance.
   */
	@Override
  public ALDParameterDescriptor clone() {
		ALDParameterDescriptor obj = 
			new ALDParameterDescriptor( this.name, this.myclass, this.explanation, 
 			 this.label, this.required, this.defaultValue, this.field, this.dataIOOrder,
 			 this.handlingMode);
		return obj;
	}

	/** 
	 * Print this descriptor to standard out.
	 */
	public void print( ) {
		print( System.out);
	}

	/** 
	 * Print this descriptor to outfile.
	 */
	public void print( PrintStream outfile) {
		outfile.println( "  Parameter <" + this.name +  "> " +
				"(" + ", type: " + this.myclass.getSimpleName() +
				", label: " + this.label +", (" + this.explanation + ")");
	}
}
