/*
 * This file is part of Alida, a Java library for 
 * Advanced Library for Integrated Development of Data Analysis Applications.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on Alida, visit
 *
 *    http://www.informatik.uni-halle.de/alida/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5472 $
 * $Date: 2012-04-18 15:37:19 +0200 (Mi, 18. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.Alida.operator.gui;

import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.io.File;

import de.unihalle.informatik.Alida.operator.*;
import de.unihalle.informatik.Alida.operator.events.ALDControlEvent;
import de.unihalle.informatik.Alida.operator.events.ALDControlEvent.ALDControlEventType;
import de.unihalle.informatik.Alida.operator.events.ALDConfigurationEvent;


import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDOperatorParameterPanel;

/**	
 * Frame to configure and run an operator.
 * <p>
 * If the operator is properly configured, i.e. its input parameters are set, 
 * it is executed using an <code>ALDExecuteOperatorFrame</code>. 
 * This control frame is free to decide, what executing precisily means, 
 * usually it is to start the operator via its <code>runOp()</code> method
 * and display or store the results appropriately.
 * <p>
 * Controllable operators trigger the display of additional buttons in GUI 
 * for pausing, resuming and interrupting calculations.
 * 
 * @author Stefan Posch
 */

public class ALDOperatorControlFrame extends JFrame 
	implements ActionListener, ItemListener {

	/**
	 * Local flag for debug output.
	 */
	@SuppressWarnings("unused")
  private boolean debug = false;

	/**
	 * Status of the associated thread to execute operator.
	 */
	protected static enum OpThreadStatus {
		/**
		 * Thread is not running.
		 */
		THREAD_STOPPED,
		/**
		 * Thread is running.
		 */
		THREAD_RUNNING,
		/**
		 * Thread is paused.
		 */
		THREAD_PAUSED
	}

	/** 
	 * The operator associated with this frame.
	 */
	protected ALDOperator op = null;
	
	/**
	 * Current status of the operator execution thread.
	 */
	private OpThreadStatus threadStatus = OpThreadStatus.THREAD_STOPPED;
	
	/**
	 * Flag to indicate if selected operator is controllable.
	 */
	protected boolean isOpControllable = false;
	
	/** 
	 * Frame to execute a configured operator.
	 * <p>
	 * Configured means that the input parameters have been set properly.
	 */
	protected ALDExecuteOperator executeOperator;

	/** 
	 * The top level panel of this frame.
	 */
	protected JPanel inputPanel;
	
	/**
	 * Title string of window.
	 */
	protected String titleString = null;
	
	/**
	 * Mode for displaying parameters.
	 * <p>
	 * In standard mode only parameters annotated as standard are shown,
	 * in advanced mode all parameters annotated are displayed.
	 */
	protected Parameter.ExpertMode displayMode = Parameter.ExpertMode.STANDARD;
	
	/**
	 * Last directory visited, initially it's user's home.
	 */
	protected String lastDirectory = System.getProperty("user.home");
	
	/**
	 * Last selected file.
	 */
	protected File lastFile = new File("operatorParams.xml");

	/*
	 * Buttons and other components for interactions with running operators.
	 */
	
	/**
	 * Run button.
	 */
	protected JButton runButton;
	/**
	 * Pause button, changes its label.
	 */
	protected JButton pauseButton;
	/**
	 * Stop button to stop snake optimization (after end of iteration).
	 */
	protected JButton stopButton;
	/**
	 * Button for stepping through optimization process.
	 */
	protected JButton stepButton;
	/**
	 * Box for activating step-through mode.
	 */
	protected JCheckBox stepThroughBox;
	/**
	 * Field for specifying number of iterations per step-through step.
	 */
	protected JTextField stepThroughStepSize;

	/**
	 * Label of status bar, changes dynamically.
	 */
	protected JLabel status;
	
	/**
	 * Panel where operator parameters are embedded.
	 */
	protected ALDOperatorParameterPanel operatorParameterPanel;

	/**
	 * Default constructor.
	 * @param opName	Name of associated operator.
	 */
	public ALDOperatorControlFrame(String opName) {

		// figure out if selected operator is 'controllable'
		try {
			this.op = (ALDOperator)(Class.forName( opName).newInstance());
		} catch (Exception ex) { 
			ex.printStackTrace();
			Object[] options = { "OK" };
			JOptionPane.showOptionDialog(null, 
					"Problems instantiating chosen operator...\n" +
					"Please check class implementation or contact programmer.",
					"Warning", JOptionPane.DEFAULT_OPTION,
					JOptionPane.WARNING_MESSAGE,
					null, options, options[0]);
			return;
		}
		
		// figure out if selected operator is 'controllable'
		if (this.op instanceof ALDOperatorControllable) {
			this.isOpControllable = true;
		}
		else {
			this.isOpControllable = false;
		}

		if (this.titleString == null)
			this.titleString = "ALDOperatorControlFrame: " + this.op.getName();
		// init the operator executer
		this.executeOperator = new ALDExecuteOperator();
		
		// init the window
		this.setupWindow();
	}
	
	/**
	 * Construct the frame to configure an operator.
	 */
	protected void setupWindow() {

		// set up the main panal
    this.inputPanel = new JPanel();
    this.add(this.inputPanel);
    
    BoxLayout ylayout = new BoxLayout( this.inputPanel, BoxLayout.Y_AXIS );
    this.inputPanel.setLayout( ylayout);

    // add operator parameter configuration panel
    this.addParamConfigPanel();

    // add panel with control buttons
    this.addButtonPanel();

		// add status bar
		this.status = new JLabel("Status: Ready");
		JPanel statusBar = new JPanel();
		statusBar.setBorder(BorderFactory.createLoweredBevelBorder());
		statusBar.add(this.status);

		// add the panels to the main panel
		this.inputPanel.add(statusBar);

		// add a nice menubar
		JMenuBar mainWindowMenu = new JMenuBar();
		JMenu fileM = new JMenu("File");
		JMenuItem itemSave = new JMenuItem("Save Settings");
		itemSave.setActionCommand("fileM_save");
		itemSave.addActionListener(this);
		JMenuItem itemLoad = new JMenuItem("Load Settings");
		itemLoad.setActionCommand("fileM_load");
		itemLoad.addActionListener(this);
		JMenuItem itemQuit = new JMenuItem("Quit");
		itemQuit.setActionCommand("fileM_quit");
		itemQuit.addActionListener(this);
		fileM.add(itemSave);
		fileM.add(itemLoad);
		fileM.add(new JSeparator());
		fileM.add(itemQuit);
		
		JMenu actionsM = new JMenu("Actions");
		JMenuItem itemRun = new JMenuItem("Run");
		itemRun.setActionCommand("actionsM_run");
		itemRun.addActionListener(this);
		JMenuItem itemValidate = new JMenuItem("Validate");
		itemValidate.setActionCommand("actionsM_validate");
		itemValidate.addActionListener(this);
		actionsM.add(itemRun);
		actionsM.add(itemValidate);
		
		JMenu viewM = new JMenu("View");
		ButtonGroup operatorLevelGroup = new ButtonGroup();
		// default
		JRadioButtonMenuItem radioItemStandard = 
			new JRadioButtonMenuItem("Standard");
		if (this.displayMode.equals(Parameter.ExpertMode.STANDARD))
			radioItemStandard.setSelected(true);
		radioItemStandard.setActionCommand("viewM_standard");
		radioItemStandard.addActionListener(this);
		JRadioButtonMenuItem radioItemAdvanced =
			new JRadioButtonMenuItem("Advanced");
		if (this.displayMode.equals(Parameter.ExpertMode.ADVANCED))
			radioItemAdvanced.setSelected(true);
		radioItemAdvanced.setActionCommand("viewM_advanced");
		radioItemAdvanced.addActionListener(this);
		operatorLevelGroup.add(radioItemStandard);
		operatorLevelGroup.add(radioItemAdvanced);
		viewM.add(radioItemStandard);
		viewM.add(radioItemAdvanced);

		// generate help menu
		JMenu helpM = this.generateHelpMenu();

		mainWindowMenu.add(fileM);
		mainWindowMenu.add(actionsM);
		mainWindowMenu.add(viewM);
		mainWindowMenu.add(Box.createHorizontalGlue());
		mainWindowMenu.add(helpM);

		// and go ..
		this.setTitle(this.titleString);
		
		// estimate size according to number of parameters in panel...
		int paramNum = this.op.getParameterNames().size();
		this.setSize( 600, paramNum*50 < 700 ? 700 : paramNum*50);
		this.setJMenuBar(mainWindowMenu);
		this.pack();
	}

	/**
	 * Sets up the help menu.
	 * @return Generated help menu.
	 */
	protected JMenu generateHelpMenu() {
		JMenu helpM = new JMenu("Help");
		JMenuItem itemHelp = new JMenuItem("Online Help");
		itemHelp.addActionListener(OnlineHelpDisplayer.getHelpActionListener(
				itemHelp,this.op.getClass().getName()));
		JMenuItem itemAbout = new JMenuItem("About Alida");
		itemAbout.setActionCommand("helpM_about");
		itemAbout.addActionListener(this);
		helpM.add(itemHelp);
		helpM.add(itemAbout);
		return helpM;
	}

	/**
	 * Adds operator parameter configuration panel to input panel.
	 */
	protected void addParamConfigPanel() {
		this.operatorParameterPanel = 
			new ALDOperatorParameterPanel(this.op, this.displayMode, true);
		JScrollPane scrollPane = new JScrollPane(this.operatorParameterPanel);
		this.inputPanel.add(scrollPane);
	}
	
	/**
	 * Adds set of control buttons to the input panel.
	 */
	protected void addButtonPanel() {
		
		JPanel runPanel = new JPanel();
		runPanel.setLayout(new GridLayout(2,1));
//		runPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));

		// controllable operators support step-wise execution...
		if (this.isOpControllable) {
			this.stepThroughBox = new JCheckBox("Step-wise execution");
			this.stepThroughBox.setActionCommand("stepFlagToggled");
			this.stepThroughBox.addItemListener(this);
			JLabel label = new JLabel("                 Step size = ");
			this.stepThroughStepSize = new JTextField("1",5);
			this.stepThroughStepSize.setEnabled(false);
			this.stepButton = new JButton("Next Step");
			this.stepButton.setActionCommand( "step");
			this.stepButton.addActionListener( this);
			this.stepButton.setBounds(50, 60, 80, 30);
			this.stepButton.setEnabled(false);
			
			JPanel stepPanel = new JPanel();
			stepPanel.setLayout(new BoxLayout(stepPanel, BoxLayout.LINE_AXIS));
			stepPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
			stepPanel.add(this.stepThroughBox);
			stepPanel.add(label);
			stepPanel.add(this.stepThroughStepSize);
			stepPanel.add(Box.createRigidArea(new Dimension(70, 0)));
			stepPanel.add(this.stepButton);
			runPanel.add(stepPanel);
		}
		
		// quit button
		JButton quitButton = new JButton("Quit");
		quitButton.setActionCommand( "quit");
		quitButton.addActionListener( this);
		quitButton.setBounds(50, 60, 80, 30);

		// validate button
//		JButton validateButton = new JButton("Validate");
//		validateButton.setActionCommand( "validate");
//		validateButton.addActionListener( this);
//		validateButton.setBounds(50, 60, 80, 30);

		// run button
		this.runButton = new JButton("Run");
		this.runButton.setActionCommand( "run");
		this.runButton.addActionListener( this);
		this.runButton.setBounds(50, 60, 80, 30);

		// validatedRun button
//		JButton vrunButton = new JButton("validatedRun");
//		vrunButton.setActionCommand( "validatedRun");
//		vrunButton.addActionListener( this);
//		vrunButton.setBounds(50, 60, 80, 30);

		// pause button for controllable plugins
		this.pauseButton = new JButton("Pause");
		this.pauseButton.setActionCommand("pause");
		this.pauseButton.addActionListener(this);
		this.pauseButton.setBounds(50, 60, 80, 30);
		this.pauseButton.setEnabled(false);

		// stop button for controllable plugins
		this.stopButton = new JButton("Stop");
		this.stopButton.setActionCommand("stop");
		this.stopButton.addActionListener(this);
		this.stopButton.setBounds(50, 60, 80, 30);
		this.stopButton.setEnabled(false);

		// now set up a panel to hold all buttons
		JPanel controlPanel = new JPanel();
		controlPanel.setLayout(new BoxLayout(controlPanel, BoxLayout.LINE_AXIS));
		controlPanel.setBorder(BorderFactory.createEmptyBorder(0, 10, 10, 10));
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(Box.createHorizontalGlue());
//		controlPanel.add(validateButton);
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(this.runButton);
		// if operator is not controllable, disable control buttons
		if (this.isOpControllable) {
			controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
			controlPanel.add(this.pauseButton);
			controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
			controlPanel.add(this.stopButton);
		}
		controlPanel.add(Box.createRigidArea(new Dimension(10, 0)));
		controlPanel.add(quitButton);

		runPanel.add(controlPanel);
		this.inputPanel.add(runPanel);
	}

	/**
	 * Changes the status message to the given string.
	 * 
	 * @param msg	Message to be shown in statusbar.
	 */
	protected void setStatus(String msg) {
		this.status.setText("Status: " + msg);
	}

	@Override
  public void actionPerformed(ActionEvent e) {

		// local variables
		String command = e.getActionCommand();

		// run the configured operator
		if (command.equals("run") || command.equals("actionsM_run")) {
			if (!this.operatorParameterPanel.validateOperatorParameters())
				return;
			this.runButton.setEnabled(false);
			if (this.isOpControllable) {
				this.pauseButton.setEnabled(true);
				this.stopButton.setEnabled(true);
			}
			this.threadStatus = OpThreadStatus.THREAD_RUNNING;
			this.setStatus("Running...");
			try {
				this.operatorParameterPanel.setOperatorParameters();
				if (this.op instanceof ALDOperatorControllable) {
					// tell the operator the requested step size via event...
					ALDConfigurationEvent confev = new ALDConfigurationEvent(this);
					confev.disableStepwiseExecution();
					// ... send it to operator
					((ALDOperatorControllable)this.op).handleALDConfigurationEvent(confev);
				}
				this.executeOperator.init(this.op, this);
				new Thread(this.executeOperator).start();
				//this.setVisible( false);
			} catch (Exception ex) {
				System.out.println(ex.getMessage());
				ex.printStackTrace();
			}
		} 
		// validate 
		else if (command.equals("actionsM_validate")) {
			this.operatorParameterPanel.validateOperatorParameters();
		}
		// quit, i.e. close window
		else if (command.equals("quit") || command.equals("fileM_quit")) {
			if (   this.threadStatus == OpThreadStatus.THREAD_RUNNING 
					|| this.threadStatus == OpThreadStatus.THREAD_PAUSED) {
				Object[] options = { "YES", "NO" };
				int selection = JOptionPane.showOptionDialog(null, 
						"The operator is running or paused!\n" +
						"If you quit now, proper termination is impossible.\n" +
						"Do you reallly want to quit?", "Warning",
						JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
						null, options, options[0]);
				switch(selection)
				{
				case JOptionPane.YES_OPTION:
					this.threadStatus = OpThreadStatus.THREAD_STOPPED;
					this.setVisible( false);
					break;
				case JOptionPane.NO_OPTION:
				case JOptionPane.CANCEL_OPTION:
					return;
				}
			}
			else {
				this.threadStatus = OpThreadStatus.THREAD_STOPPED;
				this.setVisible( false);				
			}
		} 
		// stop operator execution
		else if (command.equals("stop")) {
			((ALDOperatorControllable)this.op).handleALDControlEvent(
					new ALDControlEvent(this,ALDControlEventType.STOP_EVENT));
			this.threadStatus = OpThreadStatus.THREAD_STOPPED;
			this.stopButton.setEnabled(false);
			this.pauseButton.setEnabled(false);
			this.setStatus("Stopped.");
		} 
		// pause operator execution
		else if (command.equals("pause")) {
			((ALDOperatorControllable)this.op).handleALDControlEvent(
					new ALDControlEvent(this,ALDControlEventType.PAUSE_EVENT));
			this.pauseButton.setActionCommand("continue");
			this.pauseButton.setText("Continue");
			this.stopButton.setEnabled(false);
			this.threadStatus = OpThreadStatus.THREAD_PAUSED;
			this.setStatus("Paused...");
		} 
		// resume operator execution
		else if (command.equals("continue")) {
			((ALDOperatorControllable)this.op).handleALDControlEvent(
					new ALDControlEvent(this,ALDControlEventType.RESUME_EVENT));
			this.pauseButton.setActionCommand("pause");
			this.pauseButton.setText("Pause");
			this.stopButton.setEnabled(true);
			this.threadStatus = OpThreadStatus.THREAD_RUNNING;
			this.setStatus("Running...");
		}
		// next step in step-wise execution
		else if (command.equals("step")) {
			if (this.threadStatus == OpThreadStatus.THREAD_RUNNING) {
				((ALDOperatorControllable)this.op).handleALDControlEvent(
						new ALDControlEvent(this,ALDControlEventType.STEP_EVENT));
				this.stopButton.setEnabled(true);
			}
			// start the thread
			else {
				this.threadStatus = OpThreadStatus.THREAD_RUNNING;
				this.setStatus("Running...");
				try {
					this.operatorParameterPanel.setOperatorParameters();
					String stepSizeText = this.stepThroughStepSize.getText();
					int size = Integer.valueOf(stepSizeText).intValue();
					// tell the operator the requested step size via event...
					ALDConfigurationEvent confev = new ALDConfigurationEvent(this);
					confev.setStepsize(size);
					confev.enableStepwiseExecution();
					// ... send it to operator
					((ALDOperatorControllable)this.op).handleALDConfigurationEvent(confev);
					this.executeOperator.init(this.op, this);
					new Thread(this.executeOperator).start();
					//this.setVisible( false);
				} catch (Exception ex) {
					System.out.println(ex.getMessage());
					ex.printStackTrace();
				}
			}
		} 
		// thread terminated (attention: event is not induced by a button,
		// but just from internally...)
		else if (command.equals("threadstop")) {
			this.setStatus("Ready.");
			this.runButton.setEnabled(true);
			this.pauseButton.setActionCommand("pause");
			this.pauseButton.setText("Pause");
			this.pauseButton.setEnabled(false);
			this.stopButton.setEnabled(false);
			this.threadStatus = OpThreadStatus.THREAD_STOPPED;
		}
		
		// handle menu item commands

		else if (command.equals("fileM_save")) {
			// open file chooser
			JFileChooser getFileDialog= new JFileChooser();
			getFileDialog.setApproveButtonText("Save");
			getFileDialog.setCurrentDirectory(new File(this.lastDirectory));
			getFileDialog.setSelectedFile(this.lastFile);
			getFileDialog.setFileSelectionMode(JFileChooser.FILES_ONLY);
			int returnVal = getFileDialog.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				this.operatorParameterPanel.setOperatorParameters();
				File file = getFileDialog.getSelectedFile();
				this.lastFile = file;
				this.op.serializeToXmlFile(file.getAbsolutePath());
				this.lastDirectory = file.getAbsolutePath();
			}
		}
		else if (command.equals("fileM_load")) {
			// open file chooser
			JFileChooser getFileDialog= new JFileChooser();
			getFileDialog.setCurrentDirectory(new File(this.lastDirectory));
			getFileDialog.setSelectedFile(this.lastFile);
			getFileDialog.setFileSelectionMode(JFileChooser.FILES_ONLY);
			int returnVal = getFileDialog.showOpenDialog(null);
			if (returnVal == JFileChooser.APPROVE_OPTION) {
				File file = getFileDialog.getSelectedFile();
				this.lastDirectory = file.getAbsolutePath();
				this.lastFile = file;
				this.op = this.op.deserializeFromXmlFile(file.getAbsolutePath());
				// update the GUI with new parameter values
				this.operatorParameterPanel.updateConfigPanel(this.op);
			}
		}
		else if (command.equals("viewM_standard")) {
			this.operatorParameterPanel.changeViewMode(Parameter.ExpertMode.STANDARD);
			this.repaint();
		}
		else if (command.equals("viewM_advanced")) {
			this.operatorParameterPanel.changeViewMode(Parameter.ExpertMode.ADVANCED);
			this.repaint();
		}
		else if (command.equals("helpM_about")) {
			Object[] options = { "OK" };
			JOptionPane.showOptionDialog(null, 
					"Alida / MiToBo Graphical Operator Runner - \n" +
					"Chosen Operator: " + this.op.getName() + "\n" +
					" @2012 Martin Luther University Halle-Wittenberg", 
					"About Alida / MiToBo",
					JOptionPane.DEFAULT_OPTION, JOptionPane.PLAIN_MESSAGE,
					null, options, options[0]);
		}
	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	@Override
	public void itemStateChanged(ItemEvent e) {

		String idString = null;

		// get the object affected by the event
		Object source = e.getItemSelectable();

		if (source instanceof JCheckBox) {
			JCheckBox box = (JCheckBox) source;
			idString = box.getActionCommand();
		}
		if (source instanceof JComboBox) {
			JComboBox box = (JComboBox) source;
			idString = box.getActionCommand();
		}

		// step through optimization
		if (idString == "stepFlagToggled") {
			if (this.stepThroughBox.isSelected()) {
				// step-wise execution
				this.stepThroughStepSize.setEnabled(true);
				this.stepButton.setEnabled(true);
				this.runButton.setEnabled(false);
				this.pauseButton.setEnabled(false);
			}
			else {
				// no step-wise execution
				this.stepThroughStepSize.setEnabled(false);
				this.stepButton.setEnabled(false);
				this.runButton.setEnabled(true);
				this.pauseButton.setEnabled(false);
			}
		}
		// update GUI
		this.repaint();
	}
}
