/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev$
 * $Date$
 * $Author$
 * 
 */

package de.unihalle.informatik.MiToBo.apps.scratchAssay;

import ij.IJ;

import java.io.File;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.Vector;

import loci.common.services.DependencyException;
import loci.common.services.ServiceException;
import loci.formats.FormatException;

import de.unihalle.informatik.Alida.dataio.provider.helpers.ALDTableWindow;
import de.unihalle.informatik.Alida.datatypes.ALDFileString;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.annotations.ALDAOperator.Level;
import de.unihalle.informatik.Alida.annotations.Parameter.ExpertMode;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.operator.MTBOperator;
import de.unihalle.informatik.MiToBo.gui.MTBTableModel;
import de.unihalle.informatik.MiToBo.io.files.ImageReaderMTB;
import de.unihalle.informatik.MiToBo.io.files.ImageWriterMTB;

/**
 * class for analyzing the scratch area in several scratch assay images
 * 
 * @author glass
 *
 */
@ALDAOperator(genericExecutionMode=ALDAOperator.ExecutionMode.ALL, level=Level.APPLICATION)

public class ScratchAssayAnalyzer extends MTBOperator
{
	@Parameter(label = "first image file", required = true, direction = Parameter.Direction.IN, supplemental = false, description = "name of the first scratch assay image", mode=ExpertMode.STANDARD, dataIOOrder = 0)
	private ALDFileString fileName = new ALDFileString(IJ.getDirectory("current"));
	
	@Parameter(label = "scratch orientation", required = true, direction = Parameter.Direction.IN, supplemental = false, description = "is scratch horizontally or vertically oriented", mode = ExpertMode.STANDARD, dataIOOrder = 1)
	private ScratchOrientation orientation = ScratchOrientation.HORIZONTALLY;
	
	@Parameter(label = "\u03C3", required = true, direction = Parameter.Direction.IN, supplemental = false, description = "standard deviation of gauss filter", mode=ExpertMode.STANDARD, dataIOOrder = 2)
	private Integer sigma = null;
	
	@Parameter(label = "entropy filter size", required = true, direction = Parameter.Direction.IN, supplemental = false, description = "size of entropy filter mask", mode=ExpertMode.STANDARD, dataIOOrder = 3)
	private Integer entropyFilterSize = null;
	
	@Parameter(label = "maximum iterations", required = false, direction = Parameter.Direction.IN, supplemental = false, description = "maximum number of iterations for level set segmentation", mode=ExpertMode.ADVANCED, dataIOOrder = 0)
	private Integer maxIter = 2000;
	
	@Parameter(label = "don't check for scratch presence", required = false, direction = Parameter.Direction.IN, supplemental = false, description = "don't check for scratch presence prior to segmentation", mode=ExpertMode.STANDARD, dataIOOrder = 1)
	private Boolean noCheck = false;
	
	@Parameter(label = "batch mode", required = false, direction = Parameter.Direction.IN, supplemental = false, description = "should all image files in the given directory be analyzed", mode=ExpertMode.STANDARD, dataIOOrder = 2)
	private Boolean doBatch = true;
	
	@Parameter(label = "use external svm file", required = false, direction = Parameter.Direction.IN, supplemental = false, description = "should an external svm file be used for classification", mode=ExpertMode.ADVANCED, dataIOOrder = 3)
	private Boolean useExternalSVM = false;
	
	@Parameter(label = "external svm file", required = false, direction = Parameter.Direction.IN, supplemental = false, description = "absolute path to external svm model file", mode=ExpertMode.ADVANCED, dataIOOrder = 4)
	private ALDFileString svmFile = new ALDFileString(IJ.getDirectory("current"));
	//private String svmFile = null;

	@Parameter(label = "results table", required = true, direction = Parameter.Direction.OUT, supplemental = false, description = "table containing the resulting values")
	private MTBTableModel resultsTable = null;
	
	@Parameter(label = "silent", required = false, direction = Parameter.Direction.IN, supplemental = false, description = "should results be stored automatically", mode=ExpertMode.STANDARD)
	private Boolean silent = false;
	
	
	public enum ScratchOrientation
	{
		HORIZONTALLY,
		VERTICALLY
	}
	
	private Boolean isHorizontal = true;
	
	
	public ScratchAssayAnalyzer() throws ALDOperatorException
	{
		super();
	}
	
	
	/**
	 * 
	 * @param fileName	name of the first scratch assay image
	 * @param sigma	standard deviation of gauss filter
	 * @param entropyFilterSize	size of entropy filter mask
	 * @param isHorizontal	is scratch horizontally oriented (assumed to be vertically oriented else)
	 * @param noCheck	should scratch present not be checked
	 * @param doBatch	should all image files in the given directory be analyzed
	 * @param maxIter	maximum number of iterations for level set segmentation
	 * @throws ALDOperatorException
	 */
	public ScratchAssayAnalyzer(String fileName, int sigma, int entropyFilterSize, boolean isHorizontal, boolean noCheck, boolean doBatch, int maxIter) throws ALDOperatorException
	{
		this.fileName = new ALDFileString(fileName);
		
		this.sigma = sigma;
		this.entropyFilterSize = entropyFilterSize;
		this.isHorizontal = isHorizontal;
		this.noCheck = noCheck;
		this.doBatch = doBatch;
		this.maxIter = maxIter;
	}
	

	@Override
	public void operate() throws ALDOperatorException, ALDProcessingDAGException
	{	
		switch(orientation)
		{
			case HORIZONTALLY:
				isHorizontal = true;
				break;
			case VERTICALLY:
				isHorizontal = false;
				break;
		}
		
		String directory = new File(fileName.getFileName()).getParent();
		
		File dir = new File(directory);
		
		File[] files = dir.listFiles();
		
		Vector<String> scratchFiles = new Vector<String>();
		Vector<Double> scratchAreas = new Vector<Double>();
		Vector<Double> totalAreas = new Vector<Double>();		
		double refArea = Double.MIN_VALUE;
		
		Vector<Integer> numIterations = new Vector<Integer>();	// vector for storing the number of iterations used for segmentation
		Vector<Long> runtimes = new Vector<Long>();	// vector for storing the runtimes of each analysis
		
		//int fileNum = 0;
		
		if(doBatch && files != null)	// consider all image files in the given directory
		{
			IJ.showProgress(0);
			
			String saveDir = directory + "/analysis/";
			
			if(noCheck)
			{
				saveDir = directory + "/analysis_noCheck/";
			}
			
			//String entropyDir = directory.getString() + "/entropy/";
			File saveDirectory = new File(saveDir);
			//File entropyDirectory = new File(entropyDir);
			boolean dirCreated = false;
			
			if(silent)
			{
				dirCreated = saveDirectory.mkdir();
				//dirCreated = saveDirectory.mkdir() && entropyDirectory.mkdir();
			}
			
			if(!dirCreated)
			{
				silent = false;
			}
			
			int n = files.length;
			
			for(int i = 0; i < n; i++)
			{
				File currFile = files[i];
				
				if(currFile.isFile())
				{
					String path = currFile.getAbsolutePath();
					
					if(verbose)
					{
						System.out.println("processing " + path);
					}
					
					//MTBImage currImg = readImage(path);
					
					try
					{
						MTBImage currImg = readImage(path);
						
						if(currImg != null)
						{
							ScratchAssaySegmenter segmenter = new ScratchAssaySegmenter(currImg, sigma, entropyFilterSize, 
																						isHorizontal, noCheck, maxIter);
							
							//if(svmFile != null)
							if(useExternalSVM)
							{
								segmenter.setSVMFile(svmFile.getFileName());
							}
							
							segmenter.setVerbose(verbose);
							segmenter.runOp();
							
							double area = segmenter.getScratchArea();	// get determined scratch area
							
							MTBImage currResultImg = segmenter.getResultImage();	// get segmented scratch image
							currResultImg.setTitle(path + "_segmentation");
							
							String currSavePath = saveDir + currFile.getName() + "_segmentation.tif";
							//String currEntropyPath = entropyDir + currFile.getName() + "_entropy.tif";
							
							if(silent)
							{
								writeImage(currResultImg, currSavePath);
							
								//writeImage(segmenter.getEntropyImage().convertType(MTBImageType.MTB_FLOAT, false), currEntropyPath);
							}
							else
							{
								currResultImg.show();
							}
							
							if(path.equalsIgnoreCase(fileName.getFileName()))
							{
								refArea = area;
							}
							
							scratchFiles.add(path);
							scratchAreas.add(area);
							totalAreas.add((double)(currImg.getSizeX() * currImg.getSizeY()));
							numIterations.add(segmenter.getNumIterations());
							runtimes.add(segmenter.getRuntime());
						}
						
					}
					catch(NullPointerException e)
					{
						e.printStackTrace();
					}
					catch(ALDOperatorException e)
					{
						e.printStackTrace();
					} 
					catch(ALDProcessingDAGException e)
					{
						e.printStackTrace();
					}
					
				}
				
				IJ.showProgress(i, n);
			}
			
			resultsTable = makeTable(scratchFiles, scratchAreas, totalAreas, refArea, numIterations, runtimes);
			
			if(silent)
			{
				File tableFile = new File(saveDir + "ResultTab.txt");
				
				resultsTable.saveTable(tableFile);
			}
			else
			{
				ALDTableWindow tw = new ALDTableWindow(resultsTable);
				tw.setTitle("results");
				tw.openWindow();
			}
		}
		else	// only one image has to be analyzed
		{
			String path = fileName.getFileName();
			
			if(verbose)
			{
				System.out.println("processing " + path);
			}
			
			MTBImage currImg = readImage(path);
			
			try
			{
				ScratchAssaySegmenter segmenter = new ScratchAssaySegmenter(currImg, sigma, entropyFilterSize, 
																			isHorizontal, noCheck, maxIter);
				
				if(useExternalSVM)
				{
					segmenter.setSVMFile(svmFile.getFileName());
				}
				
				segmenter.setVerbose(verbose);
				segmenter.runOp();
				
				double area = segmenter.getScratchArea();	// get determined scratch area
				
				MTBImage currResultImg = segmenter.getResultImage();	// get segmented scratch image
				currResultImg.setTitle(path + "_segmentation");
				currResultImg.show();
				
				scratchFiles.add(path);
				scratchAreas.add(area);
				totalAreas.add((double)(currImg.getSizeX() * currImg.getSizeY()));
				numIterations.add(segmenter.getNumIterations());
				runtimes.add(segmenter.getRuntime());
				
				refArea = area;
				
				resultsTable = makeTable(scratchFiles, scratchAreas, totalAreas, refArea, numIterations, runtimes);
				
				ALDTableWindow tw = new ALDTableWindow(resultsTable);
				tw.setTitle("results");
				tw.openWindow();
			}
			catch(ALDOperatorException e)
			{
				e.printStackTrace();
			} 
			catch(ALDProcessingDAGException e)
			{
				e.printStackTrace();
			}
		}
		
	}
	
	
	/**
	 * create results table
	 * 
	 * @param scratchFiles
	 * @param scratchAreas
	 * @param totalAreas
	 * @param refArea
	 * @return table containing the results
	 */
	private MTBTableModel makeTable(Vector<String> scratchFiles, Vector<Double> scratchAreas, Vector<Double> totalAreas, double refArea, Vector<Integer> numIterations, Vector<Long> runtimes)
	{
		int n = scratchFiles.size();
		
		// display options
		NumberFormat nf = NumberFormat.getInstance();
		nf.setMaximumFractionDigits(2);
		
		// initialize table
		Vector<String> header = new Vector<String>();
		header.add("file");
		header.add("area (pixel)");
		header.add("fraction of total area(%)");
		header.add("area normalized to first image (%)");
		header.add("area decrease in reference to first image (pixel)");
		
		if(verbose)
		{
			header.add("number of iterations required");
			header.add("runtime required (sec)");
		}
		
		MTBTableModel table = new MTBTableModel(n, header.size(), header);
		
		for(int i = 0; i < n; i++)
		{
			double area = scratchAreas.elementAt(i);
			
			// insert values into results table
			table.setValueAt(scratchFiles.elementAt(i), i, 0);
			table.setValueAt(nf.format(area), i, 1);
			table.setValueAt(nf.format((area / totalAreas.elementAt(i)) * 100), i, 2);
			
			if(refArea != 0)
			{
				table.setValueAt(nf.format((area / refArea) * 100), i, 3);
			}
			else
			{
				table.setValueAt(nf.format(0), i, 3);
			}
			table.setValueAt(nf.format(refArea - area), i, 4);
			
			if(verbose)
			{
				table.setValueAt(numIterations.elementAt(i), i, 5);
				table.setValueAt(runtimes.elementAt(i), i, 6);
			}
			
		}
		
		return table;
	}
	
	
	/**
	 * 
	 * @param path
	 * @return image read from given path (if it points to a valid image)
	 */
	private MTBImage readImage(String path)
	{		
		MTBImage img = null;
		
		try
		{
			ImageReaderMTB reader = new ImageReaderMTB(path);
			reader.runOp();
			
			img = reader.getResultMTBImage();
		} 
		catch(ALDOperatorException e1)
		{
			e1.printStackTrace();
		} 
		catch(FormatException e1)
		{
			e1.printStackTrace();
		} 
		catch(IOException e1)
		{
			e1.printStackTrace();
		} 
		catch(DependencyException e1)
		{
			e1.printStackTrace();
		} 
		catch(ServiceException e1)
		{
			e1.printStackTrace();
		} 
		catch(ALDProcessingDAGException e)
		{
			e.printStackTrace();
		}
		
		return img;
	}
	
	
	/**
	 * writes the given image to a file with the specified path
	 * 
	 * @param img MTBImage to save
	 * @param path absolute path for storing
	 */
	private void writeImage(MTBImage img, String path)
	{
		try
		{
			ImageWriterMTB writer = new ImageWriterMTB(img, path);
			
			writer.runOp();
		} 
		catch(ALDOperatorException e)
		{
			e.printStackTrace();
		} 
		catch(ALDProcessingDAGException e)
		{
			e.printStackTrace();
		}
	}
	
	
	/**
	 * should results be saved without user interaction
	 * 
	 * @param silent
	 */
	public void isSilent(boolean silent)
	{
		this.silent = silent;
	}
	
	
	/**
	 * sets the path to a svm model file used for classification
	 * 
	 * @param path
	 */
	public void setSVMFile(String path)
	{
		useExternalSVM = true;
		
		this.svmFile = new ALDFileString(path);
	}

}
