/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5463 $
 * $Date: 2012-04-17 17:17:15 +0200 (Tue, 17 Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.core.dataio;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.BufferedImage;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.LinkedList;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JComponent;



import de.unihalle.informatik.Alida.annotations.ALDDataIOProvider;
import de.unihalle.informatik.Alida.dataio.provider.swing.ALDDataIOSwing;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException;
import de.unihalle.informatik.Alida.exceptions.ALDDataIOProviderException.ALDDataIOProviderExceptionType;

/**
 * Data I/O provider for GUI-OpRunner for <code>java.awt.Color</code> objects.
 * 
 * @author Oliver Gress
 *
 */
@ALDDataIOProvider

public class AwtColorDataIOSwing implements ALDDataIOSwing {

	@Override
	public Collection<Class<?>> providedClasses() {
		LinkedList<Class<?>> l = new LinkedList<Class<?>>();
		l.add(Color.class);
		return l;
	}

	@Override
	public JComponent createGUIElement(Field field, Class<?> cl, Object obj)
			throws ALDDataIOProviderException {
		
		if (obj != null && !(obj instanceof Color))	{
			throw new ALDDataIOProviderException(ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR, 
					"Object is instance of " + obj.getClass() + " instead of " + Color.class);
		}

		return new ColorChooserButton((Color)obj);
	}

	@Override
	public void setValue(Field field, Class<?> cl, JComponent guiElement,
			Object value) throws ALDDataIOProviderException {
		if (value != null && !(value instanceof Color))
			throw new ALDDataIOProviderException(ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR, 
					"value is instance of " + value.getClass() + " instead of " + Color.class);
		if (!(guiElement instanceof ColorChooserButton))
			throw new ALDDataIOProviderException(ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT,
					"guiElement is instance of " + guiElement.getClass() + " instead of " + ColorChooserButton.class);
		
		((ColorChooserButton)guiElement).setChosenColor((Color)value);
	}

	@Override
	public Object readData(Field field, Class<?> cl, JComponent guiElement)
			throws ALDDataIOProviderException {
		if (!(guiElement instanceof ColorChooserButton))
			throw new ALDDataIOProviderException(ALDDataIOProviderExceptionType.INVALID_GUI_ELEMENT,
					"guiElement is instance of " + guiElement.getClass() + " instead of " + ColorChooserButton.class);
		
		return ((ColorChooserButton)guiElement).getChosenColor();
	}

	@Override
	public JComponent writeData(Object obj) throws ALDDataIOProviderException {
		if (obj != null && !(obj instanceof Color))	{
			throw new ALDDataIOProviderException(ALDDataIOProviderExceptionType.OBJECT_TYPE_ERROR, 
					"Object is instance of " + obj.getClass() + " instead of " + Color.class);
		}

		return new ColorChooserButton((Color)obj);
	}
	

	private class ColorChooserButton extends JButton implements ActionListener {
		
		private Color color;
		
		/**
		 * 
		 */
		private static final long serialVersionUID = -7605258820929390792L;

		public ColorChooserButton(Color initialColor) {
			
			this.color = initialColor;
			
			this.updateButtonColor(this.color);
			
			this.setActionCommand("openColorChooser");
			
			this.addActionListener(this);
			
			this.setToolTipText("Choose a color to paint regions with. Press 'Cancel' in the ColorChooser to unset the color.");
		}

		@Override
		public void actionPerformed(ActionEvent e) {
			if (e.getActionCommand().equals("openColorChooser")) {
				
				this.color = JColorChooser.showDialog(null, "ColorChooser", this.color);
				this.updateButtonColor(this.color);
			}
			
		}
		
		public Color getChosenColor() {
			return this.color;
		}
		
		public void setChosenColor(Color c) {
			this.color = c;
			this.updateButtonColor(this.color);
		}
		
		private void updateButtonColor(Color c) {
			JButton jcbutton = new JButton("dummy");
			
			int icsize = 15;
			int linecolor = 0;
			
			BufferedImage im = new BufferedImage(icsize, icsize, BufferedImage.TYPE_INT_RGB);
			if (c == null) {
				int bg = jcbutton.getBackground().getRGB();
				
				for (int y = 0; y < icsize; y++) {
					for (int x = 0; x < icsize; x++) {
						if (x == y || x == icsize-y-1 || x == 0 || y == 0 || x == icsize-1 || y == icsize-1)
							im.setRGB(x, y, linecolor);
						else
							im.setRGB(x, y, bg);
					}
				}
			}
			else {
				for (int y = 0; y < icsize; y++) {
					for (int x = 0; x < icsize; x++) {
						if (x == 0 || y == 0 || x == icsize-1 || y == icsize-1)
							im.setRGB(x, y, linecolor);
						else
							im.setRGB(x, y, c.getRGB());
					}
				}
			}

			ImageIcon ic = new ImageIcon(im);
			this.setIcon(ic);
			this.updateUI();
		}
		
	}

}
