/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5010 $
 * $Date: 2012-02-22 10:56:56 +0100 (Mi, 22. Feb 2012) $
 * $Author: misiak $
 * 
 */

package de.unihalle.informatik.MiToBo.core.datatypes;

import java.awt.geom.Point2D;
import java.util.Vector;



import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.io.files.ImageWriterMTB;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.ContourOnLabeledComponents.ContourType;
import de.unihalle.informatik.MiToBo_xml.MTBXMLPoint2DDoubleType;
import de.unihalle.informatik.MiToBo_xml.MTBXMLPointVectorType;
import de.unihalle.informatik.MiToBo_xml.MTBXMLRegion2DType;

/**
 * Class to implement a 2D region object. the region is stored in a vector of 2D
 * double points with its x- and y-coordinate. The area, center of mass,
 * moments, and other properties of the region can be calculated. For further
 * details watch the implemented methods.
 * 
 * 
 * @author Danny Misiak
 */

public class MTBRegion2D {

		/**
		 * Id of the region (uniqueness is not guaranteed nor checked!).
		 */
		int id;

		/**
		 * Pixels belonging to region.
		 */
		Vector<Point2D.Double> points;

		/**
		 * Center of mass in x (just sum, not normalized by area!).
		 */
		float com_x;

		/**
		 * Center of mass in y (just sum, not normalized by area!).
		 */
		float com_y;

		/**
		 * Size of region in pixels.
		 */
		int area;

		/**
		 * Construct a new empty MTBRegion2D object.
		 */
		public MTBRegion2D() {
				this.points = new Vector<Point2D.Double>();
				this.com_x = 0.0f;
				this.com_y = 0.0f;
				this.area = 0;
		}

		/**
		 * Construct a new MTBRegion2D from the given points.
		 * 
		 * @param inPoints
		 *          vector of region points
		 */
		public MTBRegion2D(Vector<Point2D.Double> inPoints) {
				this.points = new Vector<Point2D.Double>();
				this.com_x = 0.0f;
				this.com_y = 0.0f;
				this.area = 0;
				for (int i = 0; i < inPoints.size(); i++) {
						this.addPixel(inPoints.elementAt(i));
				}
		}

		/**
		 * Construct a new MTBRegion2D from a region object that was constructed from
		 * a xml-representation. Only used for reading region sets (
		 * {@link MTBRegion2DSet}) from file.
		 * 
		 * @param xmlregion2D
		 *          object that represents the region read from xml
		 */
		MTBRegion2D(MTBXMLRegion2DType xmlregion2D) {

				this.id = xmlregion2D.getId();
				this.area = xmlregion2D.getArea();
				this.com_x = xmlregion2D.getComX();
				this.com_y = xmlregion2D.getComY();

				MTBXMLPoint2DDoubleType[] ptTypes = xmlregion2D.getPoints().getPointArray();
				this.points = new Vector<Point2D.Double>(ptTypes.length);

				for (int i = 0; i < ptTypes.length; i++) {
						this.points.add(new Point2D.Double(ptTypes[i].getX(), ptTypes[i].getY()));
				}
		}

		/**
		 * Construct an object that represents this region by xml. Only used for
		 * writing region sets ({@link MTBRegion2DSet}) to file.
		 */
		MTBXMLRegion2DType toXMLType() {
				MTBXMLRegion2DType xmlregion2D = MTBXMLRegion2DType.Factory.newInstance();

				xmlregion2D.setArea(this.area);
				xmlregion2D.setId(this.id);
				xmlregion2D.setComX(this.com_x);
				xmlregion2D.setComY(this.com_y);

				MTBXMLPoint2DDoubleType[] ptTypes = new MTBXMLPoint2DDoubleType[this.points
				    .size()];

				for (int i = 0; i < this.points.size(); i++) {
						ptTypes[i] = MTBXMLPoint2DDoubleType.Factory.newInstance();
						ptTypes[i].setX(this.points.get(i).x);
						ptTypes[i].setY(this.points.get(i).y);
				}

				MTBXMLPointVectorType pts = MTBXMLPointVectorType.Factory.newInstance();
				pts.setPointArray(ptTypes);

				xmlregion2D.setPoints(pts);

				return xmlregion2D;
		}

		/**
		 * Override java.lang.Object.clone() to create and return a copy of this
		 * object.
		 */
		@Override
		public MTBRegion2D clone() {
				MTBRegion2D tmpRegion = new MTBRegion2D(this.points);
				return tmpRegion;
		}

		/**
		 * Get a Region2D copy of this object.
		 * 
		 * @return Copy of this Region2D object.
		 */
		public MTBRegion2D getRegion() {
				return this.clone();
		}

		/**
		 * Return all points of the region.
		 * 
		 * @return Region points.
		 */
		public Vector<Point2D.Double> getPoints() {
				return this.points;
		}

		/**
		 * Return the area of the region in pixels.
		 * 
		 * @return Pixel area of the region.
		 */
		public int getArea() {
				return this.area;
		}

		/**
		 * Return x-coordinate of the center of mass.
		 * 
		 * @return Center of mass coordinate in x-dimension.
		 */
		public float getCenterOfMass_X() {
				return this.com_x / this.area;
		}

		/**
		 * Return y-coordinate of the center of mass.
		 * 
		 * @return Center of mass coordinate in y-dimension.
		 */
		public float getCenterOfMass_Y() {
				return this.com_y / this.area;
		}

		/**
		 * Extracts the extreme coordinates of the region in each dimension.
		 * 
		 * @return Array with extreme values: [xmin, ymin, xmax, ymax]
		 */
		public double[] getMinMaxCoordinates() {
				double extremes[] = new double[4];
				extremes[0] = Double.MAX_VALUE;
				extremes[1] = Double.MAX_VALUE;
				extremes[2] = Double.MIN_VALUE;
				extremes[3] = Double.MIN_VALUE;
				for (Point2D.Double p : this.points) {
						if (p.x < extremes[0])
								extremes[0] = p.x;
						if (p.x > extremes[2])
								extremes[2] = p.x;
						if (p.y < extremes[1])
								extremes[1] = p.y;
						if (p.y > extremes[3])
								extremes[3] = p.y;
				}
				return extremes;
		}

		/**
		 * Return the ID of the region. Uniqueness is not guaranteed nor checked!
		 * 
		 * @return Region ID.
		 */
		public int getID() {
				return this.id;
		}

		/**
		 * Append a point to the region from the given coordinates. The point is added
		 * at the end of the point vector.
		 * 
		 * @param x
		 *          coordinate in x-dimension
		 * @param y
		 *          coordinate in y-dimension
		 */
		public void addPixel(int x, int y) {
				this.area++;
				this.com_x = this.com_x + x;
				this.com_y = this.com_y + y;
				this.points.addElement(new Point2D.Double(x, y));
		}

		/**
		 * Append a point to the region from the given point object. The point is
		 * added at the end of the point vector.
		 * 
		 * @param p
		 *          point to add
		 */
		public void addPixel(Point2D.Double p) {
				this.area++;
				this.com_x += p.getX();
				this.com_y += p.getY();
				this.points.addElement(p);
		}

		/**
		 * Set the ID of the region. Uniqueness is not guaranteed nor checked!
		 * 
		 * @param id
		 *          Region ID
		 */
		public void setID(int id) {
				this.id = id;
		}

		/**
		 * Join a region with this region. New point objects are created for the new
		 * region.
		 * 
		 * @param reg
		 *          region to join
		 * @return New region (union of the old region and the join region).
		 */
		public MTBRegion2D join(MTBRegion2D reg) {
				int x, y;
				boolean exists;
				MTBRegion2D r = new MTBRegion2D();
				r.id = this.id;
				// add pixels of this region
				for (int i = 0; i < this.points.size(); i++) {
						r.addPixel((int) this.points.get(i).x, (int) this.points.get(i).y);
				}
				// add pixels of region 'reg' if they do not already exist in this region
				for (int j = 0; j < reg.points.size(); j++) {
						exists = false;
						x = (int) reg.points.get(j).x;
						y = (int) reg.points.get(j).y;
						for (int i = 0; i < this.points.size(); i++) {
								if ((int) this.points.get(i).x == x && (int) this.points.get(i).y == y) {
										exists = true;
										break;
								}
						}
						if (!exists) {
								r.addPixel(x, y);
						}
				}
				return r;
		}

		/**
		 * Test if a point is inside the region or not.
		 * 
		 * @param p
		 *          point that maybe lies in the region
		 * @return True if point p is inside, false if not.
		 */
		public boolean contains(Point2D.Double p) {
				for (int i = 0; i < this.points.size(); i++) {
						Point2D.Double r = this.points.elementAt(i);
						if (p.equals(r)) {
								return true;
						}
				}
				return false;
		}

		/**
		 * Method to save the region on a binary image with specific width and height.
		 * 
		 * @param file
		 *          path where the file should be saved, can be null if image should
		 *          not be stored at disk
		 * @param width
		 *          width of the binary image
		 * @param height
		 *          height of the binary image
		 * @return 2D region in a binary image of type MTBImageByte.
		 * 
		 * @throws ALDOperatorException
		 * @throws ALDProcessingDAGException
		 */
		public MTBImageByte toMTBImageByte(String file, int width, int height)
		    throws ALDOperatorException, ALDProcessingDAGException {
				MTBImageByte imageToDraw = (MTBImageByte) MTBImage.createMTBImage(width,
				    height, 1, 1, 1, MTBImageType.MTB_BYTE);
				for (int i = 0; i < this.points.size(); i++) {
						Point2D.Double p = this.points.elementAt(i);
						imageToDraw.putValueDouble((int) Math.round(p.getX()), (int) Math.round(p
						    .getY()), 255);
				}
				// save the region in a binary image
				if (file != null) {
						ImageWriterMTB writer = new ImageWriterMTB(imageToDraw, file);
						writer.setOverwrite(false);
						writer.runOp(null);
				}
				return imageToDraw;
		}

		/**
		 * Method to save a region with in a given MTBImage image.
		 * 
		 * @param file
		 *          path where the file should be saved, can be null if image should
		 *          not be stored at disk
		 * @param image
		 *          image where the region should be drawn in
		 * @return 2D inner- and outer-contours in a binary image of type
		 *         MTBImageByte.
		 * 
		 * @throws ALDProcessingDAGException
		 * @throws ALDOperatorException
		 */
		public MTBImage toMTBImage(String file, MTBImage image)
		    throws ALDOperatorException, ALDProcessingDAGException {
				MTBImage imageToDraw = image.duplicate();
				for (int i = 0; i < this.points.size(); i++) {
						Point2D.Double p = this.points.elementAt(i);
						imageToDraw.putValueDouble((int) Math.round(p.getX()), (int) Math.round(p
						    .getY()), image.getTypeMax());
				}
				// save the region in the given image on the disk
				if (file != null) {
						ImageWriterMTB writer = new ImageWriterMTB(imageToDraw, file);
						writer.setOverwrite(false);
						writer.runOp(null);
				}
				return imageToDraw;
		}

		/**
		 * Method to get the contour object from the current 2D region. Outer and
		 * inner contours will be returned from the current region object. The image
		 * size, including the region, is calculated by the minimum and maximum x- and
		 * y-coordinate of the bounding box of the region.
		 * 
		 * @return Inner and outer contours of the given region.
		 * 
		 * @throws ALDOperatorException
		 */
		public MTBContour2D getContour() throws ALDOperatorException,
		    ALDProcessingDAGException {
				/*
				 * Calculate the bounding box to get the maximum x and y value of the
				 * region.
				 */
				double[] boundingBox = this.getBoundingBox();
				int imageSizeX = (int) Math.round(boundingBox[2] + 1);
				int imageSizeY = (int) Math.round(boundingBox[3] + 1);
				/*
				 * Calculate outer and inner contours of the region.
				 */
				MTBImageByte tmpImage = (MTBImageByte) MTBImage.createMTBImage(imageSizeX,
				    imageSizeY, 1, 1, 1, MTBImageType.MTB_BYTE);
				for (int i = 0; i < this.points.size(); i++) {
						Point2D.Double p = this.points.elementAt(i);
						tmpImage.putValueInt((int) Math.round(p.x), (int) Math.round(p.y), 255);
				}
				MTBRegion2DSet regions = new MTBRegion2DSet(0, 0, imageSizeX, imageSizeY);
				regions.add(this);
				ContourOnLabeledComponents clc = new ContourOnLabeledComponents(tmpImage,
				    regions, ContourType.OUT_IN_CONTOUR, 1);
				clc.runOp(null);
				return clc.getResultContours().elementAt(0);
		}

		// /**
		// * TODO UNTESTED!!!
		// *
		// * @param regions
		// * @param imgW
		// * @param imgH
		// * @return
		// */
		// public static MTBImageRGB makeLabelImageRGB(Vector<MTBRegion2D> regions,
		// int imgW, int imgH) {
		// MTBImageRGB img = (MTBImageRGB) MTBImage.createMTBImage(imgW, imgH, 1, 1,
		// 1, MTBImageType.MTB_RGB);
		// // TreeSet<Integer> usedColors = new TreeSet<Integer>();
		//
		// Random randGen = new Random();
		// for (MTBRegion2D r : regions) {
		// // // get random color
		// // do {
		// // gray= randGen.nextInt(256*256);
		// // if (gray==0) gray= gray+1;
		// // } while (usedColors.contains(new Integer(gray)));
		// // usedColors.add(new Integer(gray));
		// // get random color
		// int red = randGen.nextInt(256);
		// int green = randGen.nextInt(256);
		// int blue = randGen.nextInt(256);
		// for (Point2D.Double p : r.getPoints())
		// img.putValue((int) p.x, (int) p.y, red, green, blue);
		// }
		// return img;
		// }

		/*
		 * >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>
		 * 
		 * Calculation of moments p,q for the MTBREgion2D.
		 * 
		 * W. Burger, M. J. Burge: "Digitale Bildverarbeitung" Springer-Verlag, 2005
		 * www.imagingbook.com
		 */

		/**
		 * Calculate moment of the region of order p,q.
		 * 
		 * @param p
		 *          order of x-component
		 * @param q
		 *          oder of y-component
		 * @return Moment as double value.
		 */
		public double getMoment(int p, int q) {
				double Mpq = 0.0;
				for (int i = 0; i < this.area; i++) {
						Point2D.Double pt = this.points.elementAt(i);
						Mpq += Math.pow(pt.x, p) * Math.pow(pt.y, q);
				}
				return Mpq;
		}

		/**
		 * Calculate central moment of the region of order p,q.
		 * 
		 * @param p
		 *          order of x-component
		 * @param q
		 *          oder of y-component
		 * @return Central moment as double value.
		 */
		public double getCentralMoment(int p, int q) {
				double m00 = this.getMoment(0, 0);
				double xCtr = this.getMoment(1, 0) / m00;
				double yCtr = this.getMoment(0, 1) / m00;
				double cMpq = 0.0;
				for (int i = 0; i < this.area; i++) {
						Point2D.Double pt = this.points.elementAt(i);
						cMpq += Math.pow(pt.x - xCtr, p) * Math.pow(pt.y - yCtr, q);
				}
				return cMpq;
		}

		/**
		 * Calculate normalized central moment of the region of order p,q.
		 * 
		 * @param p
		 *          order of x-component
		 * @param q
		 *          oder of y-component
		 * @return Normalized central moment as double value.
		 */
		public double getNormalCentralMoment(int p, int q) {
				double m00 = this.getMoment(0, 0);
				double norm = Math.pow(m00, (double) (p + q + 2) / 2);
				return this.getCentralMoment(p, q) / norm;
		}

		/**
		 * Calculate orientation of principal axis of the MTBRegion2D, using the atan2
		 * function. The orientation (angle) is given in radian in a range [-pi, +pi]
		 * 
		 * @return Orientation in radian.
		 */
		public double getOrientation() {
				// calculate different moments
				double m11 = this.getCentralMoment(1, 1);
				double m20 = this.getCentralMoment(2, 0);
				double m02 = this.getCentralMoment(0, 2);
				// calculate orientation of principal axis
				double ori = 0.5 * Math.atan2((2.0 * m11), (m20 - m02));
				return ori;
		}

		/**
		 * Calculate eccentricity of the region in range [0,1]. 1 means a sustained
		 * region.
		 * 
		 * @return Eccentricity as double value.
		 */
		public double getEccentricity() {
				// calculate different moments
				double m11 = this.getCentralMoment(1, 1);
				double m20 = this.getCentralMoment(2, 0);
				double m02 = this.getCentralMoment(0, 2);
				// calculate eccentricity
				double ecc = (Math.pow((m20 - m02), 2) + 4 * Math.pow(m11, 2))
				    / (Math.pow((m20 + m02), 2));
				return ecc;
		}

		/*
		 * <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
		 */

		/**
		 * Calculates circularity of given region (1 for perfect circle, else
		 * smaller).
		 * 
		 * @param imageSizeX
		 *          width of image that includes the region
		 * @param imageSizeY
		 *          height of image that includes the region
		 * @return Roundness of the region.
		 * @throws ALDProcessingDAGException
		 * @throws ALDOperatorException
		 */
		public double getCircularity() throws ALDOperatorException,
		    ALDProcessingDAGException {
				MTBContour2D contour = getContour();
				double perimeter = contour.getContourLength();
				if (perimeter != 0) {
						return (4 * Math.PI * this.area) / (perimeter * perimeter);
				} else {
						return 0;
				}
		}

		/**
		 * Calculates the axes-parallel bounding box of the region.
		 * <p>
		 * The function extracts the coordinates of the upper left and lower right
		 * corner of the bounding box of the region. Note that the there is at least
		 * one point of the region lying on each side of the bounding box, i.e. the
		 * region not just touches the box, but lies on it.
		 * <p>
		 * The result array contains the corner coordinates in the following order:
		 * [xmin, ymin, xmax, ymax]
		 * 
		 * @return Coordinates of upper left and lower right corners.
		 */
		public double[] getBoundingBox() {
				double minX = Double.MAX_VALUE;
				double maxX = 0;
				double minY = Double.MAX_VALUE;
				double maxY = 0;

				for (Point2D.Double p : this.points) {
						if (p.x > maxX)
								maxX = p.x;
						if (p.x < minX)
								minX = p.x;
						if (p.y > maxY)
								maxY = p.y;
						if (p.y < minY)
								minY = p.y;
				}
				double[] bbox = new double[4];
				bbox[0] = minX;
				bbox[1] = minY;
				bbox[2] = maxX;
				bbox[3] = maxY;
				return bbox;
		}
}
