/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5517 $
 * $Date: 2012-04-24 13:08:33 +0200 (Di, 24. Apr 2012) $
 * $Author: gress $
 * 
 */

package de.unihalle.informatik.MiToBo.drawing;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.Random;
import java.util.Vector;

import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.OperatorExceptionType;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.annotations.Parameter.Direction;
import de.unihalle.informatik.Alida.annotations.Parameter.ExpertMode;
import de.unihalle.informatik.MiToBo.core.datatypes.MTBRegion2DSet;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage.MTBImageType;
import de.unihalle.informatik.MiToBo.core.operator.MTBOperator;

/**
 * A class to visualize 2D regions. 
 * Background is always 0, regions can be drawn in different ways.
 * Five kinds of region images can be created or drawn to an existing image:
 * <p>
 * LABEL_IMAGE:  draw gray value image where regions are labeled by their index in the set + 1
 * ID_IMAGE:     draw gray value image where regions are labeled by their region id (see MTBRegion2D.id)
 * MASK_IMAGE:   draw a mask image with equal gray values for all regions
 * COLOR_IMAGE:  draw a mask with colored regions
 * TRANSPARENT_IMAGE: draw a mask with colored regions, but the regions are transparent to show the underlying intensity structure of the given target image
 * <p>
 * Constructors take the kind of image and the regions to be drawn as well as an eventual target image.
 * Default configurations are set by the constructors. Use the different get/set methods to specify
 * nondefault parameters like color, grayvalue etc.
 * <p>
 * Be sure to set the xMin, xMax, yMin, yMax values of the input MTBRegion2DSet, because these values are
 * used to determine the image size!!
 * 
 * @author Oliver Gress
 *
 */

@ALDAOperator(genericExecutionMode=ALDAOperator.ExecutionMode.ALL)
public class DrawRegion2DSet extends MTBOperator {
	
	@Parameter( label= "Gray value", required = false, direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 3,
	        description = "A gray value used to paint all regions")
	private Double grayValue = null;

	@Parameter( label= "Draw type", required = true, direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 2,
	        description = "The type of image to be drawn")
	private DrawType drawType = null;

	@Parameter( label= "Random generator", required = false, direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 5,
			description = "a random generator generate a random color for each region")
	private Random random = null;

	@Parameter( label= "Color", required = false,  direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 4,
			description = "A color used to paint all regions")
	private Color color = null;

	@Parameter( label= "Output-image type", required = true,  direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 6,
			description = "The datatype of the output image")
	private MTBImageType imageType = null;

	@Parameter( label= "Input regions", required = true, direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 1,
			description = "Regions to draw to image")
	private MTBRegion2DSet inputRegions = null;

	@Parameter( label= "Target image", required = false, direction = Direction.IN, 
			mode = ExpertMode.STANDARD, dataIOOrder = 7,
			description = "Draw regions to this target image")
	private MTBImage targetImage = null;

	@Parameter( label= "Result image", required = true, direction = Direction.OUT, 
			mode = ExpertMode.STANDARD, dataIOOrder = 1,
			description = "Result image")
	private MTBImage resultImage = null;
	
	/**
	 * Type of image to be drawn:
	 * LABEL_IMAGE:  draw gray value image where regions are labeled by their index in the set + 1
	 * ID_IMAGE:     draw gray value image where regions are labeled by their region id (see MTBRegion2D.id)
	 * MASK_IMAGE:   draw a mask image with equal gray values for all regions
	 * COLOR_IMAGE:  draw a mask with colored regions
	 * TRANSPARENT_IMAGE: draw a mask with colored regions, but the regions are transparent to show the underlying intensity structure of the given target image
	 * 
	 * @author Oliver Gress
	 *
	 */
	public enum DrawType {
		LABEL_IMAGE, ID_IMAGE, MASK_IMAGE, COLOR_IMAGE, TRANSPARENT_IMAGE
	}
	
	/**
	 * Default constructor where NO parameters are set. Don't use this for region image creation.
	 */
	public DrawRegion2DSet() throws ALDOperatorException {	
	}
	
	/**
	 * Simple constructor to create the most common region image types:
	 * see DrawType 
	 * dtype = COLOR_IMAGE: Mask image with randomly colored regions
	 * dtype = MASK_IMAGE:  Mask image of type MTB_BYTE, where regions have value 255
	 * dtype = LABEL_IMAGE: Label image of type MTB_INT
	 * dtype = ID_IMAGE:    ID image of type MTB_INT
	 * dtype = TRANSPARENT_IMAGE: Image with transparent randomly colored regions, needs a RGB target image to work
	 * 
	 * @param dtype drawing type
	 * @param regions a set of regions (set xMin,xMax,yMin,yMax which are used to determine image size)
	 */
	public DrawRegion2DSet(DrawType dtype, MTBRegion2DSet regions) throws ALDOperatorException {
		this.setInputRegions(regions);
		this.setDrawType(dtype);
	
		
		if (dtype == DrawType.COLOR_IMAGE || dtype == DrawType.TRANSPARENT_IMAGE) {
			this.setImageType(MTBImageType.MTB_RGB);
			this.setRandom(new Random(0));
		}
		else if (dtype == DrawType.MASK_IMAGE) {
			this.setImageType(MTBImageType.MTB_BYTE);
			this.setGrayValue(255.0);
		}
		else {
			this.setImageType(MTBImageType.MTB_INT);
		}
	}
	
	/**
	 * Simple constructor to draw the most common region image types to a given image: 
	 * dtype = COLOR_IMAGE: Mask image with randomly colored regions
	 * dtype = MASK_IMAGE:  Mask image of type MTB_BYTE, where regions have value 255
	 * dtype = LABEL_IMAGE: Label image of type MTB_INT
	 * dtype = ID_IMAGE:    ID image of type MTB_INT
	 * dtype = TRANSPARENT_IMAGE: Image with transparent randomly colored regions, needs a RGB target image to work
	 * 
	 * @param dtype drawing type
	 * @param regions a set of regions
	 * @param targetimage target image to draw the regions to
	 */
	public DrawRegion2DSet(DrawType dtype, MTBRegion2DSet regions, MTBImage targetimage) throws ALDOperatorException {
		this(dtype, regions);
		this.setTargetImage(targetimage);
	}


	@Override
	public void validateCustom() throws ALDOperatorException {
		
		if (this.getDrawType() == DrawType.LABEL_IMAGE) {
			if (this.getImageType() == MTBImageType.MTB_RGB) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion2DSet.validateCustom(): Label image cannot be drawn to a RGB image.");
			}
			
		}
		else if (this.getDrawType() == DrawType.ID_IMAGE) {
			if (this.getImageType() == MTBImageType.MTB_RGB) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion2DSet.validateCustom(): ID image cannot be drawn to a RGB image.");
			}
			else if (this.getImageType() == MTBImageType.MTB_BYTE
					|| this.getImageType() == MTBImageType.MTB_SHORT) {
				System.out.println("Warning: DrawRegion2DSet.validateCustom(): ID image of regions is drawn to " +
						" a MTB_BYTE or MTB_SHORT image. IDs are integer values, possible loss of information.");
			}	
		}
		else if (this.getDrawType() == DrawType.COLOR_IMAGE) {
			if (this.getImageType() == MTBImageType.MTB_RGB) {

				if (!(this.getColor() == null ^ this.getRandom() == null))
					throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion2DSet.validateCustom(): For color region images, a color or a" +
								" a random number generator (random colors) must be specified (exclusively).");
			}
			else {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
						"DrawRegion2DSet.validateCustom(): For color region image, the output type (probably the target) image must be MTB_RGB.");
			}	
		}
		else if (this.getDrawType() == DrawType.TRANSPARENT_IMAGE) {
			if (this.getTargetImage() == null) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
					"DrawRegion2DSet.validateCustom(): For transparent color region image, you must specify a RGB target image");	
			}
			
			if (this.getImageType() == MTBImageType.MTB_RGB) {

				if (!(this.getColor() == null ^ this.getRandom() == null))
					throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
								"DrawRegion2DSet.validateCustom(): For transparent color region images, a color or a" +
								" a random number generator (random colors) must be specified (exclusively).");
			}
			else {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
						"DrawRegion2DSet.validateCustom(): For transparent color region image, the output type (probably the target) image must be MTB_RGB.");
			}	
		}
		else if (this.getDrawType() == DrawType.MASK_IMAGE) {
			if (this.getGrayValue() == null) {
				throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
						"DrawRegion2DSet.validateCustom(): A gray value must be specified to draw a mask image.");
			}
		}
		else {
			throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
					"DrawRegion2DSet.validateCustom(): Unknown DrawType: " + this.getDrawType());
		}
		
	}
	
	/**
	 * Set the input regions to be drawn
	 */
	public void setInputRegions(MTBRegion2DSet regions) {
		this.inputRegions = regions;
	}
	
	/**
	 * Get the regions that have to be drawn
	 * @return set of regions
	 */
	public MTBRegion2DSet getInputRegions() {
		return this.inputRegions;
	}	
	
	/**
	 * Specify a target image to which the regions are drawn, specify 'null' to create a new image
	 */
	public void setTargetImage(MTBImage targetimage) {
		this.targetImage = targetimage;
		this.imageType = targetimage.getType();
	}
	
	/**
	 * Get the target image to which the regions are drawn, if one was specified. 
	 * @return target image or null, if none was specified and the operator creates a new image
	 */
	public MTBImage getTargetImage() {
		return this.targetImage;
	}
	
	/**
	 * Set the kind of region image to be drawn
	 * @param dtype see DrawType
	 */
	public void setDrawType(DrawType dtype) {
		this.drawType = dtype;
	}
	
	/**
	 * Get the kind of region image drawn by the operator
	 */
	public DrawType getDrawType() {
		return this.drawType;
	}	
	
	/**
	 * Set the resulting image's datatype. If a target image was specified and the
	 * given type differs from the target image's type, the target image is set to null.
	 * In this case a new image is created.
	 * @param type result image type
	 */
	public void setImageType(MTBImageType type) {
		if (this.getTargetImage() != null && this.getTargetImage().getType() != type) {
			this.setTargetImage(null);
		}
		this.imageType = type;
	}
	
	/**
	 * Get the datatype of the resulting image
	 */
	public MTBImageType getImageType() {
		return this.imageType;
	}	
	
	/**
	 * Set the gray value of the regions in a mask image (MASK_IMAGE only). Be aware of the
	 * resulting image's datatype.
	 */
	public void setGrayValue(Double value) {
		this.grayValue = value;
	}
	
	/**
	 * Get the gray value of the regions in a mask image (MASK_IMAGE only).
	 * @return Gray value or null, if none was specified
	 */
	public Double getGrayValue() {
		return this.grayValue;
	}
	
	/**
	 * Set the uniform color of the regions in a color image (COLOR_IMAGE only).
	 */
	public void setColor(Color c) {
		this.color = c;
		this.random = null;
	}
	
	/**
	 * Get the uniform color of the regions in a color image (COLOR_IMAGE only).
	 * @return Color or null, if none was specified.
	 */
	public Color getColor() {
		return this.color;
	}
	
	/**
	 * Set a random number generator to draw each region in a random color (COLOR_IMAGE only).
	 */
	public void setRandom(Random r) {
		this.random = r;
		this.color = null;
	}
	
	/**
	 * Get the random number generator which is responsable to draw each region in a random color (COLOR_IMAGE only).
	 * @return Random number generator or null, if none was specified
	 */
	public Random getRandom() {
		return this.random;
	}
	
	/**
	 * Set the resulting image.
	 */
	protected void setResultImage(MTBImage image) {
		this.resultImage = image;
	}
	
	/**
	 * Get the resulting region image. If a target image (InputImage) was specified, this is the same object.
	 */
	public MTBImage getResultImage() {
		return this.resultImage;
	}
	
	@Override
	protected void operate() throws ALDOperatorException,
			ALDProcessingDAGException {

		MTBRegion2DSet regs = this.getInputRegions();
		
		DrawType dtype = this.getDrawType();
		MTBImage outImg = null;
		int x,y;
		int sizeX, sizeY;
		
		if (this.getTargetImage() != null) {
			sizeX = this.getTargetImage().getSizeX();
			sizeY = this.getTargetImage().getSizeY();
			outImg = this.getTargetImage();
		}
		else {
			sizeX = (int)Math.ceil(regs.getXmax() - regs.getXmin()) + 1;
			sizeY = (int)Math.ceil(regs.getYmax() - regs.getYmin()) + 1;
			outImg = MTBImage.createMTBImage(sizeX, 
												sizeY, 
												1, 
												1, 
												1,
												this.getImageType());
		}
		
		
		
		if (dtype == DrawType.COLOR_IMAGE) {
		
			Color color = this.getColor();
			Random rand = this.getRandom();

			int c = 255;
			// if output has one color for all regions
			if (color != null)
				c = this.color2int(color);
				
			for (int i = 0; i < regs.size(); i++) {
				Vector<Point2D.Double> pts = regs.get(i).getPoints();
				
				// if output has random colors for the different regions
				if (rand != null) {
					c = this.randomColor(rand);
				}
				
				for (int j = 0; j < pts.size(); j++) {
					x = (int)pts.get(j).getX();
					y = (int)pts.get(j).getY();
					
					if (x >= 0 && x < sizeX && y >= 0 && y < sizeY) {
						
						outImg.putValueInt(x, y, c);
					}
					else {
						System.err.println("Warning: Cannot draw region pixel outside of image. Coordinate (" + x + "," + y + "), image size " + sizeX + "x" + sizeY + ".");
					}
				}	
			}
		}
		else if (dtype == DrawType.TRANSPARENT_IMAGE) {
			
			Color color = this.getColor();
			Random rand = this.getRandom();
			Color color2;
			
			int c = 255;
			double val;
			// if output has one color for all regions
				
			for (int i = 0; i < regs.size(); i++) {
				Vector<Point2D.Double> pts = regs.get(i).getPoints();
				
				// if output has random colors for the different regions
				if (rand != null) {
					color = this.int2Color(this.randomColor(rand));
				}
				
				for (int j = 0; j < pts.size(); j++) {
					x = (int)pts.get(j).getX();
					y = (int)pts.get(j).getY();
					
					if (x >= 0 && x < sizeX && y >= 0 && y < sizeY) {
						color2 = this.int2Color(outImg.getValueInt(x, y));
						val = (color2.getRed() + color2.getGreen() + color2.getBlue())/(3.0*255.0);
						c = this.color2int(new Color((int)Math.round(color.getRed()*val),
													(int)Math.round(color.getGreen()*val),
													(int)Math.round(color.getBlue()*val)));
						
						outImg.putValueInt(x, y, c);
					}
					else {
						System.err.println("Warning: Cannot draw region pixel outside of image. Coordinate (" + x + "," + y + "), image size " + sizeX + "x" + sizeY + ".");
					}
				}	
			}
		}
		else {

			double v = 0.0;
			// if output has one color for all regions
			if (this.getGrayValue() != null)
				v = this.getGrayValue();
			
			for (int i = 0; i < regs.size(); i++) {
				Vector<Point2D.Double> pts = regs.get(i).getPoints();
				
				// if output has random colors for the different regions
				if (dtype == DrawType.LABEL_IMAGE) {
					v += 1.0;
					if (v > outImg.getTypeMax()) {
						v = 1.0;
						System.err.println("Warning: Region label exceeds maximum supported pixel value. Going on resetting label to 1.");
					}
				}
				else if (dtype == DrawType.ID_IMAGE) {
					v = regs.get(i).getID();
					if (v == 0.0) {
						System.err.println("Warning: Drawing region with ID 0 as ID_IMAGE. Same pixel value as default background.");
					}
				}
				
				for (int j = 0; j < pts.size(); j++) {
					x = (int)pts.get(j).getX();
					y = (int)pts.get(j).getY();
					
					if (x >= 0 && x < sizeX && y >= 0 && y < sizeY) {
						outImg.putValueDouble(x, y, v);
					}
					else {
						System.err.println("Warning: Cannot draw region pixel outside of image. Coordinate (" + x + "," + y + "), image size " + sizeX + "x" + sizeY + ".");
					}
				}	
			}
		}
		
		this.setResultImage(outImg);
	}

	protected int randomColor(Random r) {
		return (((((int)(r.nextDouble()*200.0 + 55)) & 0xff) << 16) 
			  + ((((int)(r.nextDouble()*200.0 + 55)) & 0xff) << 8) 
			  + (((int)(r.nextDouble()*200.0 + 55)) & 0xff));
	}
	
	protected int color2int(Color c) {
		return (((c.getRed() & 0xff) << 16) + ((c.getGreen() & 0xff) << 8) + (c.getBlue() & 0xff));
	}

	protected Color int2Color(int color) {
		return new Color(((color & 0xff0000)>>16), ((color & 0xff00)>>8), (color & 0xff));
	}
}
