/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5245 $
 * $Date: 2012-03-22 17:28:25 +0100 (Do, 22. Mär 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.activecontours.energies;



import de.unihalle.informatik.MiToBo.segmentation.basics.MTBSegmentationInterface;

/**
 * Interface for active contour energies to be used with PDE techniques.
 * <p>
 * This class defines a common interface for segmentation energies to be used
 * with active contour techniques that rely on PDE approaches. In particular,
 * all energies implementing this interface have to provide methods for 
 * calculating derivatives at given image positions, updating internal 
 * parameters and allow to query for minimal and maximal energy values for 
 * normalization purposes.
 * <p>
 * Note that concrete implementations may assume that {@link updateParameters} 
 * will have been called before derivatives are queried, i.e., the method
 * {@link getDerivative} does not need to take care for parameter updates.
 * 
 * @author moeller
 */

public interface MTBGenericEnergyPDE {

  /**
   * Update internal parameters according to given segmentation.
   */
  public abstract void updateParameters(MTBSegmentationInterface seg);

  /**
   * Get the energy derivative value at position (x,y,z).
   */
  public abstract double getDerivative(MTBSegmentationInterface seg,
  																			int x, int y, int z);

  /**
   * Get the maximal possible value of the energy.
   * <p>
   * If it is not possible to specify maximal value, return value is NaN.
   */
  public abstract Double getEnergyMaxVal();
  
  /**
   * Get the minimal possible value of the energy.
   * <p>
   * If it is not possible to specify maximal value, return value is NaN.
   */
  public abstract Double getEnergyMinVal();
  
  /**
   * Every subclass has to implement the method for self-documentation.
   *
   * @return String representing the energy.
   */
  @Override
  public abstract String toString ();
}
