/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2011
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5387 $
 * $Date: 2012-04-10 15:46:32 +0200 (Di, 10. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.activecontours.energies;

import java.util.Vector;



import de.unihalle.informatik.Alida.annotations.ALDClassParameter;
import de.unihalle.informatik.Alida.annotations.Parameter.ExpertMode;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.segmentation.basics.MTBSegmentationInterface;
import de.unihalle.informatik.MiToBo.segmentation.basics.CalcSegmentationStatistics;
import de.unihalle.informatik.MiToBo.segmentation.basics.CalcSegmentationStatistics.calcTargets;

/**
 * Active contour energy based on Chan-Vese region fitting for single contour.
 * <p>
 * This energy models the inner and outer regions of a contour by a Gaussian 
 * model, i.e. each region is specified by a mean intensity value and 
 * deviations from this value are penalized.
 * <p>
 * Note that class 0 is interpreted as background while class 1 is 
 * taken as foreground or contour interior. If the provided membership 
 * function represents more than two classes, only the first two of them
 * are used and all others ignored.
 * 
 * @author moeller
 */

public class MTBGenericEnergyPDE_CVRegionFit implements MTBGenericEnergyPDE {

	/**
	 * Image to work on.
	 */
	@ALDClassParameter(label="Input Image",mode=ExpertMode.STANDARD,
			dataIOOrder=-10)
	protected transient MTBImage inImg;
	
	/**
	 * Weighting factor for inner region fit.
	 */
	@ALDClassParameter(label="Interior Lambda",mode=ExpertMode.STANDARD,
			dataIOOrder=-5)
	protected double lambda_in = 0.5;

	/**
	 * Weighting factor for outer region fit.
	 */
	@ALDClassParameter(label="Exterior Lambda",mode=ExpertMode.STANDARD,
			dataIOOrder=-4)
	protected double lambda_out = 0.5;
	
	/**
	 * Dynamic range of the image.
	 */
	private transient double inRange;
	
	/**
	 * Operator to calculate energy parameters, i.e., some statistical 
	 * numbers on the regions of the segmentation.
	 */
	private transient CalcSegmentationStatistics imgStatsOp = null;
	
	/**
   * Region sizes.
   */
  protected transient int [] areas = null;

  /**
   * Average intensities.
   */
  protected transient double [] means = null;
  
	/**
   * Average intensities of non-regions.
   * <p>
   * A non-region is the complement of a region with regarding the segmentation,
   * i.e., subsumes all pixels belonging to any other region.
   */
  protected transient double [] means_compl = null;

  /**
   * Default constructor.
   */
  public MTBGenericEnergyPDE_CVRegionFit() {
  	// nothing to do here
  }
  
  /**
   * Default constructor with arguments.
   * 
   * @param im	Image to work on.
   * @param l_in	Weighting factor for inner region fit.
   * @param l_out	Weighting factor for outer region fit.
   * @throws ALDOperatorException 
   */
  public MTBGenericEnergyPDE_CVRegionFit(MTBImage im,double lin,double lout) {
    this.inImg = im;
  	this.lambda_in= lin;
  	this.lambda_out= lout;
  }

  /**
   * Method to properly initialize instances of this class.
   * @throws ALDOperatorException
   */
  protected void init() throws ALDOperatorException {
  	this.inRange = 
  		this.inImg.getMinMaxDouble()[1] - this.inImg.getMinMaxDouble()[0];
  	// init the statistics operator
  	this.imgStatsOp = new CalcSegmentationStatistics(this.inImg);
  	Vector<calcTargets> toCalc= 
  		new Vector<CalcSegmentationStatistics.calcTargets>();
  	toCalc.add(calcTargets.classMean);
  	this.imgStatsOp.setTargets(toCalc);
  }
  
  /**
   * Updates parameters according to given membership.
   * @param mem		Current segmentation object.
   */
  @Override
  public void updateParameters(MTBSegmentationInterface mem) {
  	
  	// update the energy parameter
  	try {
  		this.imgStatsOp.setSegmentation(mem);
	    this.imgStatsOp.runOp(true);
	    this.areas = this.imgStatsOp.getRegionSizes();
	  	this.means = this.imgStatsOp.getRegionMeans();
	  	this.means_compl = new double[this.means.length];
	  	
	  	double totalArea = 0;
	  	for (int i=0;i<this.areas.length;++i) 
	  		totalArea += this.areas[i];
	  	
	  	// update the complemental average values
	  	for (int i=0;i<this.means.length;++i) {
	  		double sum = 0;
	  		for (int j=0;j<this.means.length;++j) {
	  			if (j == i)
	  				continue;
	  			sum += this.means[j] * this.areas[j];
	  		}
	  		sum /= totalArea;
	  		this.means_compl[i] = sum;
	  	}
	  	
    } catch (ALDOperatorException e) {
    	System.out.println("MTBEnergy_ChanVese: " + 
    		"update failed, keeping old params...");
    } catch (ALDProcessingDAGException e) {
    	System.out.println("MTBEnergy_ChanVese: " + 
    		"update failed, keeping old params...");
    }
  }

  /**
   * Returns the energy derivative value at position (x,y,z).
   */
  @Override
  public double getDerivative(MTBSegmentationInterface seg,int x,int y,int z) {

  	// calculate statistical numbers of interior and exterior class
  	double mean_in = this.means[1];
  	double mean_out = this.means[0];
  	double pixInt = this.inImg.getValueDouble(x, y, z);
    return    this.lambda_in * (pixInt - mean_in) * (pixInt - mean_in)
    				- this.lambda_out * (pixInt - mean_out) * (pixInt - mean_out);
  }
  
  @Override
  public Double getEnergyMaxVal() {
		return new Double(this.lambda_in * this.inRange * this.inRange);  	
  }
  
  @Override
  public Double getEnergyMinVal() {
		return new Double(- this.lambda_out * this.inRange * this.inRange);
  }
  
  /**
   * Returns the average region intensities last calculated.
   */
  public double [] getMeans() {
  	return this.means.clone();
  }
}
