/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5243 $
 * $Date: 2012-03-22 16:12:00 +0100 (Do, 22. Mär 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.basics;

import java.util.Vector;

import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.ALDAOperator.ExecutionMode;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.annotations.Parameter.Direction;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.operator.MTBOperator;

/**
 * Helper class for doing statistical calculations on 2D/3D grayscale images. 
 * 
 * @author moeller
 */
@ALDAOperator(genericExecutionMode=ExecutionMode.ALL)

public class CalcSegmentationStatistics extends MTBOperator {

	/**
	 * Statistical numbers provided by this class.
	 */
	public static enum calcTargets {
		/**
		 * Size of individual regions.
		 */
		classSize,
		/**
		 * Average intensity values of all regions.
		 */
		classMean,
		/**
		 * Intensity variance within each region.
		 */
		classVar		
	}
	
  /**
   * Input image.
   */
  @Parameter(label="image", direction = Direction.IN, required = true,
 		description = "Input image to work on.")
  private MTBImage image = null;

	/**
	 * Segmentation object.
	 */
	@Parameter(label="segmentation", direction = Direction.IN,
		required = true, description = "Segmentation of image.")
  private MTBSegmentationInterface segmentation = null;

  /**
   * List of targets to calculate.
   */
  @Parameter(label="targets", direction = Direction.IN,
  	required = true, description="List of targets to calculate.")
  private Vector<calcTargets>  targetList = null;

  /**
   * List of calculated region sizes.
   */
  @Parameter(label="regionsizes", direction = Direction.OUT,
  	description = "Calculated region sizes.")
  private int [] regionSizes = null;

  /**
   * List of calculated region average values.
   */
  @Parameter(label="regionmeans", direction = Direction.OUT,
  	description = "Calculated region averages.")
  private double [] regionMeans = null;

  /**
   * List of calculated region variance values.
   */
  @Parameter(label="regionvars", direction = Direction.OUT,
  	description = "Calculated region variances.")
  private double [] regionVars = null;

  /**
	 * Width of working image.
	 */
	private int iWidth;
	
	/**
	 * Height of working image. 
	 */
	private int iHeight;
	
	/**
	 * Depth of working image. 
	 */
	private int iDepth;

	/**
   * Default constructor without arguments.
   * <p>
   * Note, there are parameters that need to be set prior to executing the
   * operator. Don't call runOp() before doing that!
   */
  public CalcSegmentationStatistics()
  	throws ALDOperatorException {
  	// nothing to do here...
  }

  /**
   * Default constructor.
   * 
   * @param im	Image to work on.
   */
  public CalcSegmentationStatistics(MTBImage im)
  	throws ALDOperatorException {
  	this.image= im;
  	this.iDepth= im.getSizeZ();
  	this.iHeight= im.getSizeY();
  	this.iWidth= im.getSizeX();
  }

  /**
   * Default constructor.
   * 
   * @param im	Image to work on.
   * @param seg Segmentation.
   */
  public CalcSegmentationStatistics(MTBImage im, MTBSegmentationInterface seg)
  	throws ALDOperatorException {
  	this.image= im;
  	this.iDepth= im.getSizeZ();
  	this.iHeight= im.getSizeY();
  	this.iWidth= im.getSizeX();
  	this.segmentation= seg;
  }
  
  /**
   * Default constructor.
   * 
   * @param im	Image to work on.
   * @param seg Image segmentation.
   * @param targets List of statistical numbers to calculate per region.
   */
  public CalcSegmentationStatistics(MTBImage im, MTBSegmentationInterface seg,
  																 Vector<calcTargets> targets) 
  	throws ALDOperatorException {
  	this.image= im;
  	this.iDepth= im.getSizeZ();
  	this.iHeight= im.getSizeY();
  	this.iWidth= im.getSizeX();
  	this.segmentation= seg;
  	this.targetList= targets;
  }
  
  /**
   * Specifiy image segmentation.
   */
  public void setSegmentation(MTBSegmentationInterface seg) {
  	this.segmentation = seg;
  }

  /**
   * Specify targets to calculate.
   */
  public void setTargets(Vector<calcTargets> targets) {
  	this.targetList = targets;
  }

  /**
   * Get array of region areas indexed with class labels.
   */
  public int [] getRegionSizes() {
  	return this.regionSizes;
  }
  
  /**
   * Get region average intensity array indexed with class labels.
   */
  public double [] getRegionMeans() {
  	return this.regionMeans;
  }
  
  /**
   * Get region intensity variance array indexed with class labels.
   */
  public double [] getRegionVars() {
  	return this.regionVars;
  }
  
	@Override
  protected void operate() {
		
		if (this.targetList.isEmpty())
			return;
		
		// update size of internal variables
		int regionNum = this.segmentation.getNumberOfClasses();
		int maxLabel = this.segmentation.getMaxLabel();
		int arraySize = (regionNum == maxLabel + 1) ? regionNum : maxLabel + 1;
		
		this.regionSizes = new int [arraySize];
		this.regionMeans = new double [arraySize];
		this.regionVars = new double [arraySize];
		double [] weightSum = new double [arraySize];
		
		// calculate average intensities and sizes based on visible pixels
		if (   this.targetList.contains(calcTargets.classSize) 
				|| this.targetList.contains(calcTargets.classMean)
				|| this.targetList.contains(calcTargets.classVar)) {
			
			for (int z = 0; z < this.iDepth; ++z) {
				for (int y = 0; y < this.iHeight; ++y) {
					for (int x = 0; x < this.iWidth; ++x) {

						if (!this.segmentation.isVisible(x, y, z))
							continue;

						int classLabel = this.segmentation.getClass(x, y, z);
						weightSum[classLabel] += this.segmentation.getWeight(x, y, z);
						this.regionMeans[classLabel] += this.image.getValueInt(x, y, z)  
						                         * this.segmentation.getWeight(x, y, z);
						this.regionSizes[classLabel]++;
					}
				}
			}
			// compute final parameters
			for (int c = 0; c<arraySize; ++c) {
//				if (this.regionSizes[c] > 0)
				if (weightSum[c] > 0)
					this.regionMeans[c] = this.regionMeans[c]/weightSum[c];
			}
		}

		// calculate intensity variance of visible pixels
		weightSum = new double[arraySize];
		if (this.targetList.contains(calcTargets.classVar)) {

			for (int z = 0; z < this.iDepth; ++z) {
				for (int y = 0; y < this.iHeight; ++y) {
					for (int x = 0; x < this.iWidth; ++x) {

						if (!this.segmentation.isVisible(x, y))
							continue;

						int classLabel = this.segmentation.getClass(x, y);
						weightSum[classLabel] += this.segmentation.getWeight(x, y, z);
						this.regionVars[classLabel] += 
                this.segmentation.getWeight(x, y, z) *
	 						 (this.image.getValueInt(x, y) - this.regionMeans[classLabel])
							*(this.image.getValueInt(x, y) - this.regionMeans[classLabel]);
					}
				}
			}
			// compute final parameters
			for (int c = 0; c<arraySize; ++c) {
//			if (this.regionSizes[c] > 0)
				if (weightSum[c] > 0)
					this.regionVars[c] = this.regionVars[c]/weightSum[c];
			}
    }
  }	  
}
