/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5495 $
 * $Date: 2012-04-19 16:00:48 +0200 (Do, 19. Apr 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE;


import java.io.PrintStream;

import de.unihalle.informatik.MiToBo.core.datatypes.images.*;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;



/**
 * A class to implement the Chan-Vese energy for nonPDE levelsets.
 *
 * @author Stefan Posch
 *
 */

@ALDDerivedClass

public class MTBChanVeseEnergyNonPDE extends MTBGenericEnergyNonPDE
{
	private MTBCVFittingEnergyNonPDE fittingEnergy;
	private MTBLengthEnergyNonPDE    lengthEnergy;
	private MTBSizeEnergyNonPDE      sizeEnergy;

	/**
	 * Construct an enery object for the level set function <code>LSM</code> realizing the Chan-Vese energy. 
	 * The level set function may be two or multi phase.
	 * The lambda for the background is lambdaBg and for all for all object phases an identical lambdaFg. 
	 * Mu and nu for length and size term.
	 * 
	 * @param img
	 * @param LSM
	 * @param lambdaBg
	 * @param lambdaFg
	 * @param mu
	 * @param nu
	 */
	public MTBChanVeseEnergyNonPDE(MTBImage img, MTBLevelsetMembership LSM, double lambdaBg, double lambdaFg, double mu, double nu)
	{
		this.name = new String( "Chan-Vese energy");
		fittingEnergy = new MTBCVFittingEnergyNonPDE( img, LSM, lambdaBg, lambdaFg);
		lengthEnergy = MTBLengthEnergyNonPDE.createEnergy( LSM, mu);
		//lengthEnergy = MTBLengthEnergyNonPDE.createEnergy( mu,
		//							LSM.getNumPhases() != 2,
		//							LSM.getSizeZ() != 1);
		sizeEnergy = new MTBSizeEnergyNonPDE( nu);
	}

	/**
	 * Construct an enery object for the level set function <code>LSM</code> realizing the Chan-Vese energy. 
	 */
	public MTBChanVeseEnergyNonPDE()
	{
		this.name = new String( "Chan-Vese energy");
	}


    // TODO: maybe we just shold call deltaE with newPhase given to avoid redundance,
    //        or leave it for efficience ??
	@Override
	public double deltaE(int x , int y , int z , MTBLevelsetMembership LSM) {
		return fittingEnergy.deltaE( x, y, z, LSM) +
			    lengthEnergy.deltaE( x, y, z, LSM) +
			    sizeEnergy.deltaE( x, y, z, LSM);
	}
	
	@Override
	public double deltaE(int x , int y , int z , byte newPhase, MTBLevelsetMembership LSM)
	{
		return fittingEnergy.deltaE( x, y, z, newPhase, LSM) +
				 lengthEnergy.deltaE( x, y, z, newPhase, LSM) +
				 sizeEnergy.deltaE( x, y, z, newPhase, LSM);
	}
	
	@Override
	public double E(MTBLevelsetMembership LSM) {
		return fittingEnergy.E( LSM) +
				 lengthEnergy.E( LSM) +
				 sizeEnergy.E( LSM);
	}
	
	@Override
	public void updateParams(int x, int y, int z, MTBLevelsetMembership LSM) {
            fittingEnergy.updateParams(x, y, z, LSM);
            lengthEnergy.updateParams(x, y, z, LSM);
            sizeEnergy.updateParams(x, y, z, LSM);
	}

	@Override
	public void updateParams(int x, int y, int z, byte newPhase, MTBLevelsetMembership LSM) {
        fittingEnergy.updateParams(x, y, z, newPhase, LSM);
        lengthEnergy.updateParams(x, y, z, newPhase, LSM);
        sizeEnergy.updateParams(x, y, z, newPhase, LSM);
	}

	@Override
	public void estimateParams(MTBLevelsetMembership LSM) {
		fittingEnergy.estimateParams(LSM);
		lengthEnergy.estimateParams(LSM);
		sizeEnergy.estimateParams(LSM);
	}
	
	@Override
	public void setDebug( int debug) {
		this.debug = debug;
		fittingEnergy.setDebug( debug);
		lengthEnergy.setDebug( debug);
		sizeEnergy.setDebug( debug);
	}

	@Override
	public  String 	toString() {
        return name + " composed of\n" +
			   fittingEnergy.toString() + "\n" +
               lengthEnergy.toString() + "\n" +
               sizeEnergy.toString();

    }

	@Override
	public void print( MTBLevelsetMembership LSM, PrintStream out, String indent) {
		out.println( "Chan-Vese energy");
			
		String newIndent = getNewIndent( indent);
        fittingEnergy.print( LSM, out, newIndent);
        lengthEnergy.print( LSM, out, newIndent);
        sizeEnergy.print( LSM, out, newIndent);
			
	}

}
