/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5495 $
 * $Date: 2012-04-19 16:00:48 +0200 (Do, 19. Apr 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE;


import java.lang.Math;
import java.io.PrintStream;

import de.unihalle.informatik.MiToBo.core.datatypes.images.*;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;



/**
  * This class models a length energy where for each non background phase
  * the length of this phase has a Gaussian prior. The parameters of the Gaussian
  * are the mean alpha and the standdard deviation sigma.
  * The loglikelihood of
  * this prior - weight with the constant mu - defines the energy.
  */

@ALDDerivedClass

public class MTBGaussianLengthEnergyNonPDE extends MTBLengthEnergyNonPDE
{
	private final double SQRT_2 = Math.sqrt(2.0);

	private double alpha; // mean of Gaussian
	private double sigma; // stddev of Gaussian
	private double sigma_2; // sigma^2

	private double l[];   // current length of each phase

	private int numPhases; // number of phases including background

	/**
	 * constructor
	 * 
	 * @param img	Image to segment
	 * @param LSM	Levelset function
	 * @param mu
	 * @param alpha
	 * @param sigma
	 */
	public MTBGaussianLengthEnergyNonPDE(MTBImage img, MTBLevelsetMembership LSM, 
								double mu, double alpha, double sigma) {
		super( mu);
		this.name = new String( "MTBGaussianLengthEnergyNonPDE");
		this.numPhases = LSM.getNumPhases();

		this.alpha = alpha;
		this.sigma = sigma;
		this.sigma_2 = sigma*sigma;
		this.l = new double[numPhases+1];

		this.estimateParams(LSM);
	}

	/**
	 * constructor
	 */
	public MTBGaussianLengthEnergyNonPDE() {
		this.name = new String( "MTBGaussianLengthEnergyNonPDE");
	}

	/**
	 */
	public double deltaE(int x, int y, int z, byte newPhase, MTBLevelsetMembership LSM)
	{
		int sizeX = (LSM).getSizeX();
		int sizeY = (LSM).getSizeY();
		int sizeZ = (LSM).getSizeZ();

		// 3D not yet implemented
		if ( sizeZ != 1 ) {
			System.err.println( "MTBGaussianLengthEnergyNonPDE: deltaE(int x, int y, int z, MTBLevelsetMembership LSM) not implemented yet");
			System.exit(1);
		}
	
        double lenDelta[] = new double[numPhases+1]; // accumulate delta for each phase
		computeDeltaLength( x, y, newPhase, LSM, lenDelta);

		double delta = 0.0;
		delta -=  getEnergyForLength( l[LSM.getPhase(x,y)]);
		delta -=  getEnergyForLength( l[newPhase]);
		delta +=  getEnergyForLength( l[LSM.getPhase(x,y)]+lenDelta[LSM.getPhase(x,y)]);
		delta +=  getEnergyForLength( l[newPhase]+lenDelta[newPhase]);

		return mu * delta;
    }

	/**
	*/
	public double E(MTBLevelsetMembership LSM)
	{
		double sum = 0.0;
			
		if ( LSM.getSizeZ() != 1 ) {
			System.err.println( "MTBGaussianLengthEnergyNonPDE: E for 3D not implemented yet");
			System.exit(1);
		} else {
			//TODO: as soon as update is checked, we can compute directly the enery from l[]
			estimateParams( LSM);

        	for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				sum += getEnergyForLength( l[p]);

		}
		return this.mu*sum;
	}
	
	/**
	 */
	public void updateParams(int x, int y, int z, byte newPhase, MTBLevelsetMembership LSM)
	{
		if ( LSM.getSizeZ() != 1 ) {
			System.err.println( "MTBGaussianLengthEnergyNonPDE: updateParams for flipping not implemented yet");
			System.exit(1);
		}

		if ( (debug & FLAG_LENGTH) != 0 ) 
			System.out.println( "updateParams @(" + x + "," + y + ")" + 
								LSM.getPhase(x,y) + " --> " + newPhase);
		//String str = new String(  "");
		//for ( byte p = 1 ; p <= numPhases ; p++ )
            //str = str.concat( "\t" + p + ": " + l[p]);
		//System.out.println( "updateParams - before: " + str);

        double lenDelta[] = new double[numPhases+1]; // accumulate delta for each phase
		computeDeltaLength( x, y, newPhase, LSM, lenDelta);
		// TODO: only old and new phase may have changed OR call with this.l instead of lenDelta
		for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) {
			l[p] += lenDelta[p];
		}
	}

	/**
	 */
	protected void estimateParams(MTBLevelsetMembership LSM) {
		//System.out.println( "MTBGaussianLengthEnergyNonPDE: estimateParams");
		//String str = new String();
		//for ( byte p = 0 ; p <= numPhases ; p++ )
		//    str = str.concat( "\t" + p + ": " + l[p] + "\n");
		//System.out.println( str);

        for ( byte p = 1 ; p <= numPhases ; p++ )
			this.l[p] = 0.0;

		if ( LSM.getSizeZ() != 1 ) {
			System.err.println( "MTBGaussianLengthEnergyNonPDE: E for 3D not implemented yet");
			System.exit(1);
		} else {
			for(int y = 0; y < LSM.getSizeY(); y++) {
				for(int x = 0; x < LSM.getSizeX(); x++) {
					byte phaseCurrent = LSM.getPhase( x,y); // 
					byte phaseRight = getPhaseCheckX( x+1,y,LSM);
					byte phaseBelow = getPhaseCheckY( x,y+1,LSM);

					gradMagHeaviside( phaseCurrent, phaseRight, phaseBelow, this.l, false);
				}
			}
			
		}
		//str = new String();
		//for ( byte p = 0 ; p <= numPhases ; p++ )
		//   str = str.concat( "\t" + p + ": " + l[p] + "\n");
		//System.out.println( str);
	}
	
	/**
	 */
	public  String 	toString() {
        String str = new String(  name + "    mu = " + mu + 
								"\talpha " + alpha + "\tsigma " + sigma);
		return str;
    }

	/**
	 */
	public void print( MTBLevelsetMembership LSM, PrintStream out, String indent) {
		out.println( this + " energy: " + E( LSM));
		out.print( "    ");
		for ( byte p = 2 ; p <= numPhases ; p++ )
			out.print( l[p] + "\t");
		out.println();
    }

	/**
	 * Return phase of pixel x,y in levelset function LSM. 
	 * Check if x or y is in bounds. If not, return background phase
	 *
	 * @param x
	 * @param y
	 * @param LSM	levelset function
	 */
	// TODO should become inline
	static private byte getPhaseCheckXY( int x, int y, MTBLevelsetMembership LSM) {
		return ( x >= 0 && x < LSM.getSizeX() && y >= 0 && y < LSM.getSizeY() ) ? LSM.getPhase( x,y) : MTBLevelsetMembership.BG_PHASE;
    }

	/**
	 * Return phase of pixel x,y in levelset function LSM. 
	 * Check if x is in bounds. If not, return background phase
	 *
	 * @param x
	 * @param y
	 * @param LSM	levelset function
	 */
	// TODO should become inline
	static private byte getPhaseCheckX( int x, int y, MTBLevelsetMembership LSM) {
		return ( x >= 0 && x < LSM.getSizeX() ) ? LSM.getPhase( x,y) : MTBLevelsetMembership.BG_PHASE;
    }

	/**
	 * Return phase of pixel x,y in levelset function LSM. 
	 * Check if y is in bounds. If not, return background phase
	 *
	 * @param x
	 * @param y
	 * @param LSM	levelset function
	 */
	// TODO should become inline
	static private byte getPhaseCheckY( int x, int y, MTBLevelsetMembership LSM) {
		return ( y >= 0 && y < LSM.getSizeY() ) ? LSM.getPhase( x,y) : MTBLevelsetMembership.BG_PHASE;
    }

	/**
	 * Return energy for length
	 *
	 * @param l
	 */
	// TODO should become inline
	private double getEnergyForLength( double l) {
		return (l - alpha)*(l - alpha) / sigma_2;
    }

	/**
	 * @param value
	 */
	public double heaviside(double value)
	{
			if(value > 1) return 1.0;
			else return 0.0;
	}

	/**
	 * Compute of change in length for each phase if we would change the phase of (x,y) to newPhase
	 *
	 * @param x
	 * @param y
	 * @param newPhase
	 * @param LSM
	 */

	// TODO: use getPhaseCheck?
	private void computeDeltaLength(int x, int y, byte newPhase, MTBLevelsetMembership LSM, double lenDelta[])
	{
		if ( (debug & FLAG_LENGTH) != 0 ) 
			System.out.println( "computeDeltaLength @(" + x + "," + y + "): " + 
								LSM.getPhase( x, y) + " --> " + newPhase);

		//for ( int yy = y-1 ; yy <= y+1 ; yy++ ) {
    		//System.out.print( "    ");
			//for ( int xx = x-1 ; xx <= x+1 ; xx++ ) {
    			//System.out.print( LSM.getPhase( xx,yy) + " ");
			//}
			//System.out.println( );
		//}

		int sizeX = LSM.getSizeX();
		int sizeY = LSM.getSizeY();


		byte oldPhase = LSM.getPhase( x,y); // old phase at x,y
		byte phaseCurrent; // for each pixel considered: its phase 
		byte phaseRight; // for each pixel considered: phase of its right neighbor
		byte phaseBelow; // for each pixel considered: phase of its below neighbor

		// consider current pixel (x,y)
		phaseRight = getPhaseCheckX( x+1,y,LSM);
		phaseBelow = getPhaseCheckY( x,y+1,LSM);

		gradMagHeaviside( oldPhase, phaseRight, phaseBelow, lenDelta, true);
		if ( (debug & FLAG_LENGTH) != 0 ) {
		    for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				System.out.print( " " + lenDelta[p]);
			System.out.println();
		}

		gradMagHeaviside( newPhase, phaseRight, phaseBelow, lenDelta, false);
		if ( (debug & FLAG_LENGTH) != 0 ) {
		    for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				System.out.print( " " + lenDelta[p]);
			System.out.println();
		}

		// now its above neighbor
		phaseCurrent = getPhaseCheckY( x,y-1,LSM); 
		phaseRight = getPhaseCheckXY( x+1,y-1, LSM); 

		gradMagHeaviside( phaseCurrent, phaseRight, oldPhase, lenDelta, true);
		if ( (debug & FLAG_LENGTH) != 0 ) {
		    for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				System.out.print( " " + lenDelta[p]);
			System.out.println();
		}

		gradMagHeaviside( phaseCurrent, phaseRight, newPhase, lenDelta, false);
		if ( (debug & FLAG_LENGTH) != 0 ) {
		    for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				System.out.print( " " + lenDelta[p]);
			System.out.println();
		}


		// finally its left neighbor
		phaseCurrent = getPhaseCheckX( x-1,y, LSM); 
		phaseBelow = getPhaseCheckXY( x-1,y+1, LSM);

		gradMagHeaviside( phaseCurrent, oldPhase, phaseBelow, lenDelta, true);
		if ( (debug & FLAG_LENGTH) != 0 ) {
		    for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				System.out.print( " " + lenDelta[p]);
			System.out.println();
		}

		gradMagHeaviside( phaseCurrent, newPhase, phaseBelow, lenDelta, false);
		if ( (debug & FLAG_LENGTH) != 0 ) {
		    for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) 
				System.out.print( " " + lenDelta[p]);
			System.out.println();
		}


        // TODO: for testing, should be remove if ok
        for ( byte p = MTBLevelsetMembership.BG_PHASE ; p <= numPhases ; p++ ) {
            if ( p != oldPhase && p != newPhase && lenDelta[p] != 0.0 ) {
                System.err.println( "MTBGaussianLengthEnergyNonPDE, computeDeltaLength: PANIC");
				System.exit(1);
        	}
        }
	}

	/**
	 * Calculate the gradient magnitude of the foreward difference of the heavyside function
	 * corresponding to the level set for one pixel. Add the contribution
     * for each phase to the values in the array l.
     * If negate is true, the return negated contributions, i.e. subtracted.
	 * 
	 * @param p	heaviside functtion at (x,y,z)
	 * @param q heaviside functtion at (x+1,y,z)
	 * @param r heaviside functtion at (x,y+1,z)
	 * @return
	 */

    // we consider also the BG_PHASE which is not required, but make the code cleaner
	private void gradMagHeaviside(byte p, byte q, byte r, double l[], boolean negate) {
		if ( q == r ) {
			if ( p != q ) {
				if ( negate ) {
					l[p] -= SQRT_2;
					l[q] -= SQRT_2;
				} else {
					l[p] += SQRT_2;
					l[q] += SQRT_2;
				} 
			} 
		} else {
			if ( p == q ) {
				// r != p
				if ( negate ) {
					l[p] -= 1;
					l[r] -= 1;
				} else {
					l[p] += 1;
					l[r] += 1;
				} 

			} else if ( p == r ) {
				// q != p
				if ( negate ) {
					l[p] -= 1;
					l[q] -= 1;
				} else {
					l[p] += 1;
					l[q] += 1;
				} 
			} else {
				// q != r
				if ( negate ) {
					l[p] -= SQRT_2;
					l[q] -= 1;
					l[r] -= 1;
				} else {
					l[p] += SQRT_2;
					l[q] += 1;
					l[r] += 1;

				}
			}
		}
	}

//		/** Get energy term a pixel (x,y) for LSM. 2D only
//		 * @param x
//		 * @param y
//		 * @param LSM
//		*/
//
//		// TODO: not really checked
//		public double E( int x, int y, MTBLevelsetMembership LSM) {
//			int sizeX = (LSM).getSizeX();
//			int sizeY = (LSM).getSizeY();
//
//			byte phaseCurrent = LSM.getPhase( x,y); // old phase at x,y
//			byte phaseRight = ( (x+1) < sizeX) ? LSM.getPhase( x+1,y) : MTBLevelsetMembership.BG_PHASE;
//			byte phaseBelow = ( (y+1) < sizeY) ? LSM.getPhase( x,y+1) : MTBLevelsetMembership.BG_PHASE;
//
//			return mu * gradMagHeaviside( phaseCurrent, phaseRight, phaseBelow);
//        }
}
