/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5250 $
 * $Date: 2012-03-23 10:20:18 +0100 (Fr, 23. Mär 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE;



import java.io.PrintStream;

/** An abstract class for energies intended for a nonPDE levelset approach.
 * Typically each instance of this class is associated
 * with an instance of <code>MTBLevelsetMembership</code>.
 * As these two separate objects (of class <code>MTBGenericEnergyNonPDE</code> and <code>MTBLevelsetMembership</code>)
 * interact some caution has to be taken into account.
 * This is the case if the energy object relies on parameters
 * depending on the current state of the levelset function associated with the energy.
 * Examples for such parameters are the mean intensities of phases/regions
 * as used in the fitting term of the Chan-Vese-Energy.
 * In this case changing of the levelset function and setting of the parameters of the energy object
 * need to be synchronized.
 * The concept used e.g. in the solver <code>LevelsetSolveNonPDE</code>
 * is as follows: First the parameters of the energy object are updated where the
 * the leveleset function supplied as argument is still in the old state
 * and the anticipated change of the leveleset function is given via arguments, too.
 * Subequently the energy object assumes that this change of the leveleset function
 * is conducted immediately to again have both energy function and 
 * leveleset function in compatible states.
 * <p>
 * When deriving from this base class it is necessary to override at least the
 * constructor (defining the name of the energy) as well as
 * the abstract methods <code>E</code> and <code>toString</code>.
 * Also the abstract version of the 
 * method <code>deltaE</code> 
 * which explicitly refers to the new phase the pixel should change to has to be implemented.
 * The second version of <code>deltaE</code>
 * makes sense in a strict way only for two phase approaches and an appropriate
 * default implementation is supplied.
 * In case the derived energy class relies on parameters depending on the
 * current state of the levelset function associated with the energy,
 * then also the version of <code>updateParams</code>  has to be implemented
 * which explicitly refers to the new phase the pixel should change to.
 * The second version of <code>updateParams</code> has an appropriate default implementation.
 * In this case it is also highly recommended if not necessary to implement the method
 * <code>estimateParams</code>.
 * <p>
 * If the neighborhood of pixels is used to
 * define the energy then
 * it is recommended to implement a 2D and 3D version of the energy
 * in separate classes.
 * This is the case e.g. if derivatives are used.
 */


public abstract class MTBGenericEnergyNonPDE {

	/** Name of the energy.
	 */
	protected String name; // name of the energy

	/** Bit mask for debug output on <code>System.out</code>.
	 */
	protected int debug; 

	/**
	 * Constructor
	 * 
	 */
	public MTBGenericEnergyNonPDE() {
		this.name = new String( "MTBGenericEnergyNonPDE");
	}

	/**
	 * Calculate differences of energy if the pixel/voxel (x,y,z) changes phase in
	 * the levelset function <code>phi</code> supplied.
	 * 
	 * This version makes sense only for two phase levelset function. In the
	 * multiphase case any object phase is changed to the background, while a
	 * background pixel/voxel is changed to the first object phase.
	 *
	 * @param x
	 * @param y
	 * @param z
	 * @param phi  levelset function to work on
	 * @return		difference of energy
	 */
	public double deltaE(int x, int y, int z, MTBLevelsetMembership phi) {
		if ( phi.getPhase( x, y, z) == MTBLevelsetMembership.BG_PHASE ) {
			return deltaE( x, y, z, (byte)(MTBLevelsetMembership.BG_PHASE+1), phi);
		} else {
			return deltaE( x, y, z, MTBLevelsetMembership.BG_PHASE, phi);
		}
	}

	/**
	 * Calculate differences of energy if the pixel/voxel (x,y,z) is changed
	 * in the levelset function <code>phi</code> supplied
	 * from current phase to newphase.
	 * 
	 * @param x		x-coordinate of voxel
	 * @param y		y-coordinate of voxel
	 * @param z		z-coordinate of voxel
	 * @param newPhase		new phase to change pixel/voxel to
     * @param phi  levelset function to work on
	 * @return		difference of energy
	 */
	public abstract double deltaE(int x, int y, int z, byte newPhase, MTBLevelsetMembership phi);

	/**
	 * Return the complete energy for the levelset function <code>phi</code>.
	 * <br>
	 *  Note: this may be an expensive operation.
	 *
	 * @param phi	levelset function
	 * @return		weight energy
	*/
	public abstract double E(MTBLevelsetMembership phi);
	
	/** Update parameters (if any) of the energy object.
	 * 
     *  Update parameters if we want to change a pixel/voxel to the other state,
     *  assuming two phase levelset function.
	 *  For a multi phase levelset function flipping of phases is realized
	 * 	as in {@link MTBGenericEnergyNonPDE#deltaE}.
	 * <br>
     *  The level set function <code> phi</code> is still in the old state and should be updated subsequently.
	 *
	 * @param x
	 * @param y
	 * @param z
	 * @param phi
	 */
	public void updateParams(int x, int y, int z, MTBLevelsetMembership phi)
	{
		if ( phi.getPhase( x, y, z) == MTBLevelsetMembership.BG_PHASE ) {
			updateParams( x, y, z, (byte)(MTBLevelsetMembership.BG_PHASE+1), phi);
		} else {
			updateParams( x, y, z, MTBLevelsetMembership.BG_PHASE, phi);
		}
	}
	
	/** Update parameters (if any) of the energy object
	 * if we want to change a pixel/voxel to newPhase.
	 * 
     *  The level set function <code> phi</code> is still in the old state and should be updated subsequently.
	 *
	 * @param x
	 * @param y
	 * @param z
	 * @param newPhase		new phase to change pixel/voxel to
	 * @param phi
	 */
	public void updateParams(int x, int y, int z, byte newPhase, MTBLevelsetMembership phi) {
	};
	
	/**
	 * Estimate internal parameters (if any) for the level set function <code>phi</code>.
	 *
	 * @param phi
	 */
	protected void estimateParams(MTBLevelsetMembership phi) {
	}

	/**
	 * Print this energy object - including the value of the energy for 
	 * <code>phi</code>. 
	 * Each line output is prefixed with the string <code>indent</code>.
	 * 
	 *  <b>Note:</b> this may be an expensive operation.
	 *
	 * @param phi	levelset function
	 * @param out	stream to print to
	 * @param indent	indentation string
	 */
	public void print( MTBLevelsetMembership phi, PrintStream out, String indent) {
		out.println( indent + name + " energy = " + E( phi));
	}

	/**
	 * Return an ascii representation identifying the energy and internal parameters,
	 * but not the state.
	 */
	public abstract String toString();

	/**
	 * Set debug bit mask
	 *
	 * @param debug	bit mask
	 */
	public void setDebug( int debug) {
		this.debug = debug;
	}

	/**
	 * Return a new indentation string. If old indent was empty, just leave is,
	 * otherwise add four spaces.
	 *
	 * @param indent	Indentation asked for
	 */
	protected String getNewIndent( String indent) {
		return indent == null  ? new String ("") : indent + "    ";
	}
}
