/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5495 $
 * $Date: 2012-04-19 16:00:48 +0200 (Do, 19. Apr 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE;

import de.unihalle.informatik.MiToBo.core.datatypes.images.*;



/**
 * Implementation of 2D mean separation energy of Yezzi et al.
 * <p>
 * For details, see:
 *	
 * Yezzi, Tsai, Willsky, "A Fully Global Approach to Image Segmentation via 
 * Coupled Curve Evolution Equations", JVCIR, 13, 195-216 (2002) 
 *
 * Note that this implementation is only applicable for two-phase levelsets! 
 *
 * @author moeller
 * 
 * @todo ggf. Intensitätsnormalisierung einbauen
 * @todo ausführlich testen, sobald Topologieerhaltung funktioniert
 */


public class MTBLSMeanSeparationEnergy2DNonPDE extends MTBGenericEnergyNonPDE
{
	
	/**
	 * Image to work on. 
	 */
	private MTBImage inimg;

	/**
	 * Array to hold phase average intensities.
	 */
	private double [] means = new double[2];
	
	/**
	 * Absolute energy value associated with current segmentation.
	 */
	private double energy = 0.0;
	
	/**
	 * Default constructor.
	 * 
	 * @param img		Input image.
	 * @param LSM		Initial segmentation.
	 */
	public MTBLSMeanSeparationEnergy2DNonPDE(MTBImage img, MTBLevelsetMembership LSM)
	{
		if (LSM.getNumPhases() > 2) {
			System.err.println("Levelsets: " +
					"MeanSeparationEnergy is just applicable for 2 phases, returning!");
			return;
		}
		if (LSM.getSizeZ() > 1) {
			System.err.println("Levelsets: " +
					"MeanSeparationEnergy is just applicable for 2D images, returning!");
			return;
		}
		
		this.name = new String( "mean separation energy");
		this.inimg = img;
		
		// estimate initial parameters, i.e. phase average values
		this.estimateMeans(LSM);
	}

	@Override
  public double deltaE(int x, int y, int z, byte newPhase,
  											MTBLevelsetMembership LSM) {
		// valid phase IDs start with 1, 0 is invalid
		byte phase = LSM.getPhase(x, y);
		byte newphase = (phase == 1) ? (byte)2 : (byte)1;
		
		// image intensity value
		double value = this.inimg.getValueDouble(x, y);
		
		// current energy
		double energy_now = this.energy;

		// incremental update of mean in current phase of pixel
		double newMean_oldPhase = this.means[phase-1] + 
							(this.means[phase-1] - value)/(LSM.getSizePhase(phase)-1);
		// incremental update of mean in current phase of pixel
		double newMean_newPhase = this.means[newphase-1] - 
							(this.means[newphase-1] - value)/(LSM.getSizePhase(newphase)+1);

		// new energy
		double energy_new = -0.5 * (newMean_oldPhase - newMean_newPhase) 
														 * (newMean_oldPhase - newMean_newPhase);

		// return the difference
	  return energy_new - energy_now;
  }

	@Override
  public double E(MTBLevelsetMembership LSM) {
		this.estimateMeans(LSM);		
		System.out.println("mean_1 = " + this.means[0]);
		System.out.println("mean_2 = " + this.means[1]);
		return -0.5 * (this.means[0] - this.means[1]) 
								* (this.means[0] - this.means[1]);
  }

	@Override
  public String toString() {
	  return this.name;
  }
	
	/**
	 * Estimate phase average values and current energy.
	 * 
	 * @param LSM Initial levelset segmentation.
	 */
	public void estimateMeans(MTBLevelsetMembership LSM) {

		for ( byte p = 0 ; p < 2 ; ++p )
			this.means[p] = 0.0;

		int phase= 0;
		for(int y = 0; y < LSM.getSizeY(); y++) {
			for(int x = 0; x < LSM.getSizeX(); x++) {
				// check if pixel invalid
				if ((phase = LSM.getPhase( x, y, 0)) != 0)
					this.means[phase-1] += this.inimg.getValueDouble(x, y, 0, 0, 0);
			}
		}
		for ( byte p = 1 ; p <= 2 ; p++ ) 
			this.means[p-1] /= LSM.getSizePhase(p);
		
		// absolute energy value
		this.energy = -0.5 * (this.means[0] - this.means[1]) 
											 * (this.means[0] - this.means[1]);
	}
	
	@Override
  public void updateParams(int x, int y, int z, 
														byte newPhase, MTBLevelsetMembership LSM) {
		byte phase = LSM.getPhase(x, y);
		byte newphase = (phase == 1) ? (byte)2 : (byte)1;
		
		double value = this.inimg.getValueDouble(x, y);
		
		// incremental update of mean in current phase of pixel
		this.means[phase-1] = this.means[phase-1] + 
							(this.means[phase-1] - value)/(LSM.getSizePhase(phase)-1);
		// incremental update of mean in current phase of pixel
		this.means[newphase-1] = this.means[newphase-1] - 
							(this.means[newphase-1] - value)/(LSM.getSizePhase(newphase)+1);
		// new absolute energy value
		this.energy = -0.5 * (this.means[0] - this.means[1]) 
											 * (this.means[0] - this.means[1]);
	}
}
