/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5495 $
 * $Date: 2012-04-19 16:00:48 +0200 (Do, 19. Apr 2012) $
 * $Author: posch $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE;


import java.io.PrintStream;



/** An abstract class for energies for a nonPDF levelset approach
 * based on different length energies of the phases boundaries.
 * NOTE: each create function return a suitable energy instance has to call the
 * method setScaleFactor()!!
 */


public abstract class MTBLengthEnergyNonPDE extends MTBGenericEnergyNonPDE
{
	/** weight of the length term
	 */
	protected double mu = 1.0; 

	/** do normalization?
     * The dimension of the largest direction in the original image (thus level set function)
     * is scale to 1
	 */
	protected boolean doNormalization = false;

	/** Scaling factor for normalization
	 */
	protected double scaleFactor = 1.0;

	/** Bit mask for length energies for debugging purposes.
     * The value has to be coordinated with other bit masks.
     */
	protected final int FLAG_LENGTH = 4; // flag for length term

	/**
	 * Constructor
	 * 
	 */
	public MTBLengthEnergyNonPDE() {
		this( 1.0);
	}

	/**
	 * Constructor
	 * 
	 * @param mu	weight of the length term
	 */
	public MTBLengthEnergyNonPDE(double mu) {
		this.mu = mu;
		this.name = new String( "Length energy");
	}

	/**
	 */
	@Override
	public  String 	toString() {
        return name + ": mu = " + mu + " scaleFactor = " + scaleFactor;
    }

//	/**
//	 * Create a proper lenth energy asked for. Two phase vs multi phase and
//	 * 2D vs 3D.
//	 * Length is approximated using forward differences.
//	 *<br>
//	 * In the future more types, e.g. various methods to apporximate length
//	 * may be available
//	 *
//	 * @param	mu	weight of length term
//	 * @param	multiPhase	
//	 * @param	threeD	
//	 */
//	public static MTBLengthEnergyNonPDE createEnergy( double mu, boolean multiPhase, boolean threeD) {
//		//setScaleFactor( ls);
//
//		if ( multiPhase ) {
//			if ( threeD ) {
//				return new MTBLengthEnergyFW3DMPNonPDE( mu);
//			} else {
//				return new MTBLengthEnergyFW2DMPNonPDE( mu);
//			}
//		} else {
//			if ( threeD ) {
//				return new MTBLengthEnergyFW3D2PNonPDE( mu);
//			} else {
//				return new MTBLengthEnergyFW2D2PNonPDE( mu);
//			}
//		}
//	}

	/**
	 * Create a proper lenth energy asked for by the levelset function.
	 * If sizeZ is one than 2D, 3D otherwise.
	 * If number of phases is 2 than two phase, otherweise multiphase
	 *<br>
	 * In the future more types, e.g. various methods to apporximate length
	 * may be available
	 *
	 * @param	mu weight of length term
	 * @param	ls Levelset function
	 * @return	length energy
	 */
	public static MTBLengthEnergyNonPDE createEnergy(MTBLevelsetMembership ls, double mu) {

		MTBLengthEnergyNonPDE res;
		if ( ls.getNumPhases() > 2 ) {
			if ( ls.getSizeZ() != 1 ) {
				res = new MTBLengthEnergyFW3DMPNonPDE( mu);
			} else {
				res = new MTBLengthEnergyFW2DMPNonPDE( mu);
			}
		} else {
			if ( ls.getSizeZ() != 1 ) {
				res = new MTBLengthEnergyFW3D2PNonPDE( mu);
			} else {
				res = new MTBLengthEnergyFW2D2PNonPDE( mu);
			}
		}

		res.setScaleFactor( ls);
		return res;
	}

    /** set the scale factor if normalization is requested.
     * @param   ls Levelset function
     */
	protected void setScaleFactor( MTBLevelsetMembership ls) {
		if ( doNormalization )  {
			int maxSize = Math.max( Math.max( ls.getSizeX(), ls.getSizeY()), ls.getSizeZ());
			scaleFactor = 1/(double)maxSize;
		} else {
			scaleFactor = 1.0;
		}
	}
    
}
