package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE;

import java.io.PrintStream;

import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;



/**
 * class that implements the energy functional from
 * 
 * O. Dzyubachyk, W. Niessen and E. Meijering,
 * "A VARIATIONAL MODEL FOR LEVEL-SET BASED CELL TRACKING IN TIME-LAPSE FLUORESCENCE MICROSCOPY IMAGES",
 * ISBI, 2007
 * 
 * @author glass
 *
 */

public class MTBMeijeringFittingEnergyNonPDE extends MTBGenericEnergyNonPDE
{
	private final int numPhases; // number of phases including background
	private double[] mean;  // array to hold the mean intensity values of the phases
	private double[] var; // array to hold the variancees of the intensity values of the phases
	private MTBImage img;

	private final int sizeX;
	private final int sizeY;
	private final int sizeZ;
	
	
	/**
	 * constructor
	 * 
	 * @param img
	 * @param ls
	 */
	public MTBMeijeringFittingEnergyNonPDE(MTBImage img, MTBLevelsetMembership ls)
	{
		this.name = "MTBMeijeringFittingEnergyNonPDE";
		this.img = img;
		this.sizeX = img.getSizeX();
		this.sizeY = img.getSizeY();
		this.sizeZ = img.getSizeZ();
		this.numPhases = ls.getNumPhases();
		this.mean = new double[numPhases + 1];
		this.var = new double[numPhases + 1];
		
		estimateParams(ls);
	}
	
	
	/**
	 * 
	 */
	protected void estimateParams(MTBLevelsetMembership ls)
	{
		// determine mean intensity values of all phases
		for(int z = 0; z < sizeZ; z++)
		{
			for(int y = 0; y < sizeY; y++)
			{
				for(int x = 0; x < sizeX; x++)
				{
					mean[ls.getPhase(x, y, z)] += img.getValueDouble(x, y, z, 0, 0);
				}
			}
		}
		
		
		for(byte i = 1; i <= numPhases; i++)
		{
			mean[i] /= ls.getSizePhase(i);
		}
		
		// determine variances of intensity values of all phases
		for(int z = 0; z < sizeZ; z++)
		{
			for(int y = 0; y < sizeY; y++)
			{
				for(int x = 0; x < sizeX; x++)
				{
					var[ls.getPhase(x, y, z)] += ((img.getValueDouble(x, y, z, 0, 0) - mean[ls.getPhase(x, y, z)]) 
											    * (img.getValueDouble(x, y, z, 0, 0) - mean[ls.getPhase(x, y, z)]));
				}
			}
		}
		
		
		for(byte i = 1; i <= numPhases; i++)
		{
			var[i] /= ls.getSizePhase(i);
		}	
	}
	
	
	public void updateParams(int x, int y, int z, byte newPhase, MTBLevelsetMembership ls)
	{
            byte oldPhase = ls.getPhase(x, y, z);

            double val = img.getValueDouble(x, y, z, 0, 0);

            double diffOld = (mean[oldPhase] - val);
            double diffNew = (mean[newPhase] - val);

            double sizeOld =  ls.getSizePhase(oldPhase);
            double sizeNew =  ls.getSizePhase(newPhase);

            // update means and variances adaptively
            mean[oldPhase] = mean[oldPhase] + diffOld / (sizeOld - 1);
            mean[newPhase] = mean[newPhase] - diffNew / (sizeNew + 1);
            
//            var[oldPhase] = (var[oldPhase] - (diffOld * diffOld) / (sizeOld - 1)) * (sizeOld / (sizeOld - 1));
//            var[newPhase] = (var[newPhase] - (diffNew * diffNew) / (sizeNew + 1)) * (sizeNew / (sizeNew + 1));
            
            var[oldPhase] = (var[oldPhase] - (diffOld * diffOld) / (sizeOld - 1)) * (sizeOld / (sizeOld - 1));
            var[newPhase] = (var[newPhase] + (diffNew * diffNew) / (sizeNew + 1)) * (sizeNew / (sizeNew + 1));
	}
	
	
	
	@Override
	public double deltaE(int x, int y, int z, byte newPhase, MTBLevelsetMembership ls)
	{
		byte oldPhase = ls.getPhase(x, y, z);
		double val = img.getValueDouble(x, y, z, 0, 0);
		
		double sizeOld =  ls.getSizePhase(oldPhase);
        double sizeNew =  ls.getSizePhase(newPhase);
		double oldMu = mean[oldPhase];
		double newMu = mean[newPhase];
		double oldSigmaSqr = var[oldPhase];
		double newSigmaSqr = var[newPhase];
		
//		double sOld = (sizeOld / (sizeOld - 1)) * (oldSigmaSqr + (((val - oldMu) * (val - oldMu)) / (sizeNew + 1)));
//		double sNew = (sizeNew / (sizeNew + 1)) * (newSigmaSqr + (((val - newMu) * (val - newMu)) / (sizeOld - 1)));
//		
//		double delta = sizeNew * Math.log(sNew / newSigmaSqr) + sizeOld * Math.log(sOld / oldSigmaSqr) + Math.log(sNew/sOld);
		
		
		double sOld = (sizeOld / (sizeOld - 1)) * (oldSigmaSqr - (((val - oldMu) * (val - oldMu)) / (sizeOld - 1)));
		double sNew = (sizeNew / (sizeNew + 1)) * (newSigmaSqr + (((val - newMu) * (val - newMu)) / (sizeNew + 1)));
		
		double delta = sizeNew * Math.log(sNew / newSigmaSqr) + sizeOld * Math.log(sOld / oldSigmaSqr) + Math.log(sNew/sOld);
		
		return delta;
	}
	

	@Override
	public double E(MTBLevelsetMembership ls)
	{
		double e = 0;
		
		for(int z = 0; z < sizeZ; z++)
		{
			for(int y = 0; y < sizeY; y++)
			{
				for(int x = 0; x < sizeX; x++)
				{
					byte phase = ls.getPhase(x, y, z);
					double val = img.getValueDouble(x, y, z, 0, 0);
					
					e += Math.log(var[phase]) + ((val - mean[phase]) * (val - mean[phase])) / (var[phase]) ;
				}
			}
		}
		
		return e;
	}
	

	@Override
	public String toString()
	{
		StringBuffer sb = new StringBuffer(this.name + " with " + numPhases + " phases\n");
		for(int i = 1; i <= this.numPhases; i++)
		{
			sb.append("Phase " + i + ": mean = " + mean[i] + ", variance = " + var[i] + "\n");
		}
		return sb.toString();
	}

	public void print(MTBLevelsetMembership ls, PrintStream out, String indent)
	{
		out.println(indent + name + " energy = " + E(ls));

		String newIndent = getNewIndent(indent);
		out.print(newIndent + "Means: ");
		for(int i = 1; i <= this.numPhases; i++)
		{
			out.print( indent + mean[i] + "\t");
		}  
		out.println();

        out.println(newIndent + "Variances:");
        for(int i = 1; i <= this.numPhases; i++)
        {
        	out.print(indent + var[i] + "\t");
        }
		out.println();
	}
}
