/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5510 $
 * $Date: 2012-04-23 17:36:57 +0200 (Mo, 23. Apr 2012) $
 * $Author: moeller $
 * 
 */

// TODO:  output of intermediate results for plugin
// TODO: optional: filename base fuer intermediate LS
// TODO: javadoc for all energy types
package de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.ui;

import java.util.Vector;
import java.io.*;

import ij.IJ;
import ij.ImagePlus;
import ij.ImageStack;
import ij.gui.GenericDialog;
import ij.plugin.filter.PlugInFilter;
import ij.process.ImageProcessor;
import ij.io.OpenDialog;


import de.unihalle.informatik.Alida.exceptions.ALDException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;


import de.unihalle.informatik.Alida.annotations.ALDAOperator;
import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.MiToBo.core.datatypes.*;
import de.unihalle.informatik.MiToBo.core.datatypes.images.*;
import de.unihalle.informatik.MiToBo.core.operator.*;
import de.unihalle.informatik.MiToBo.segmentation.regions.labeling.LabelComponentsSequential;
import de.unihalle.informatik.MiToBo.segmentation.contours.extraction.*;
import de.unihalle.informatik.MiToBo.morphology.*;
import de.unihalle.informatik.MiToBo.segmentation.thresholds.ImgThresh;

import de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.*;
import de.unihalle.informatik.MiToBo.segmentation.levelset.nonPDE.LevelsetSolveNonPDE.*;


/**
 * Operator intended as user interface for levelset segmentation
 * using nonPDE optimization and optionally topology preserving.
 * <br>
 * An image containing the initial segmentation may be supplied.
 * Also a mask image indicating invalid pixels/voxels may be given.
 * <p>
 * <b>Details</b>
 * <p>
 * 
 * 
 * The input image is segmented using a levelset technique.
 * Segmentation by be done using two or multiple phases, if <b>multiphase</b> is true.
 * Optionally a topology preserving mode is available (only 2D currently)
 * which assumes initially object phases with exactly one connected component each
 * and no holes. This condition is preserved during optimization.
 * <br>
 * The maximal number of iterations is given as <b>maxiter</b>.
 * <br> 
 * If <b>invalidImg</b> all pixels/voxels in this image with non-zero intensities are
 * interpreted as invalid. This is in principle as if this pixels/voxels would not exist
 * (similar to pixels/voxels outside the image domain):
 * They are neither assigned a phase to, nor are image intensities used, e.g. to compute
 *  gradients.
 * <p>
 * <b>Initialization</b>
 * <br>If no initialization is given, a default initialization,
 * usually a circle/sphere is used.
 * If both <b>initLabelImg</b> and <b>initBinImg</b> are supplied, <b>initLabelImg</b> is used.
 * If a two phase approach is used (i.e. <b>multiphase</b> is false) then both
 * types of initialization are identical: all pixels/voxels with intensity zero are
 * assumed as background (phase), all other as foreground phase.
 * In the multi phase case the labels in <b>initLabelImg</b> are assumed to be the
 * phases to be used for initialization. Again, pixels/voxels with intensity zero are
 * assumed as background (phase).
 * If in the multi phase case <b>initBinImg</b> is given, the image is interpreted
 * as a binary image (zero is again background), a compontent labeling is
 * done and each resulting connected component assign to one phase
 * (in addition to the background phase).
 * <p>
 * <b>Energies</b>
 * <p> Currently the following energies are available (more may be added):
 * <p>
 * The standard Chan-Vese energy 
 * consists of a length, a area, and a fitting term.
 * The corresponding weigths follow the typical notation with <b>mu</b>, <b>nu</b>, and lambdas.
 * <b>lambdafg</b> and <b>lambdabg</b> are used of the foreground phase(s) and the background phase
 * respectively. Note, for multi phase levels set currently all foreground phases
 * share the same <b>lambdafg</b>.
 * <p>
 * The standard Chan-Vese energy
 * where the standard length energy is substituted
 * by a Gaussian length term. This realizes a Gaussian prior on the
 * length of each object phase with mean <b>alpha</b> and standard deviation <b>sigma</b>.
 * <p>
 * some already implemented energies have to be documented here !!!!
 * <p>
 * <b>Result</b>
 * <p>The final segmentation is output as a label image.
 * <p>
 * <b>Additional Output</b>
 * <p>If <b>verbose</b> is set output is written to stdout.
 * The <b>debug</b> argument is a bit mask to control debuggung output.
 * This is intended for developers and not (well) documented, also behaviour may
 * change over time without notice.
 * If <b>spacing</b> is non-zero, intermediate levelset functions after each <b>spacing</b>-th
 * iteration are output to files with prefix <b>ibase</b> (default intermediateLS)
 * and incrementally numbered. As a zero-th levelset function the initialization is
 * prepended and the final result added as a last levelset image (which might happen
 * to show up twice).
 * Additionally the input image is overlayed with the final contours of the object phases
 * as well as with the contours of the intialization. This image is output
 * to the file <b>contourResult</b> (default: contourImg.tif).
 * 
 * <p><b>Limitations</b>
 * Currently not all options are implemented for the plugin version.
 *
 * @author Stefan Posch, Markus Glass
 *
 */


@ALDAOperator

public class LevelSet_Segmentation_NonPDE extends MTBOperator {

    static final Integer DEFAULT_MAXITER = 100;
    static final Double DEFAULT_MU = 1.0;
    static final Double DEFAULT_LAMBDA_FG = 1.0;
    static final Double DEFAULT_LAMBDA_BG = 1.0;
    static final Double DEFAULT_NU = 0.0;
	static final Double DEFAULT_ALPHA = 0.0;
	static final Double DEFAULT_SIGMA = 20.0;
	static final Double DEFAULT_KAPPA = 0.0;

	public enum EnergyType { CHANVESE, CV_FITTING, CHANVESE_GAUSSIAN, CURVATURE, MEIJERING, MEANSEPARATION, CV_FITTING_CURVATURE };

	public LevelSet_Segmentation_NonPDE() throws ALDOperatorException {}

	@Parameter( label= "inImg", required = true, type = Parameter.Type.INPUT,
		description = "Input image to segment")
	MTBImage inImg = null;

	@Parameter( label= "initBinImg", required = false, type = Parameter.Type.INPUT,
		description = "Binary image for initialization")
	MTBImage initBinImg = null;

	@Parameter( label= "initLabelImg", required = false, type = Parameter.Type.INPUT,
		description = "Label image for initialization")
	MTBImage initLabelImg = null;

	@Parameter( label= "invalidImg", required = false, type = Parameter.Type.INPUT,
		description = "Image of invalid pixels (pixels <> 0 are invalid")
	MTBImage invalidImg = null;

	@Parameter( label= "ResultImage", required = false, type = Parameter.Type.OUTPUT,
		description = "Resulting label image")
	MTBImage ResultImage;

	@Parameter( label= "ResultContourImage", required = false, type = Parameter.Type.OUTPUT,
		description = "Resulting contour image")
	MTBImage ResultContourImage;

	@Parameter( label= "multiphase", required = true, type = Parameter.Type.PARAMETER,
		description = "Use multiphase levelsets")
	Boolean multiphase = false;

    @Parameter( label= "preserveTopology", required = true, type = Parameter.Type.PARAMETER,
              description = "Preserve topology during optimization")
    Boolean preserveTopology = false;

	@Parameter( label= "maxIter", required = true, type = Parameter.Type.PARAMETER,
		description = "Maximal number of iterations for optimization")
	Integer maxIter = DEFAULT_MAXITER;

	@Parameter( label= "threshold", required = false, type = Parameter.Type.PARAMETER,
		description = "Maximal number of iterations for optimization")
	Double threshold = null;

	@Parameter( label= "energytype", required = true, type = Parameter.Type.PARAMETER,
		description = "Energy type")
	EnergyType energytype = EnergyType.CHANVESE;

	@Parameter( label= "mu", required = true, type = Parameter.Type.PARAMETER,
		description = "Weight for length term")
	Double mu = 1.0;

	@Parameter( label= "nu", required = true, type = Parameter.Type.PARAMETER,
		description = "Weight for area term")
	Double nu = 1.0;

	@Parameter( label= "lambdafg", required = true, type = Parameter.Type.PARAMETER,
		description = "Weight for foreground homogenity")
	Double lambdafg = 1.0;

	@Parameter( label= "lambdabg", required = true, type = Parameter.Type.PARAMETER,
		description = "Weight for background homogenity")
	Double lambdabg = 1.0;

	@Parameter( label= "alpha", required = true, type = Parameter.Type.PARAMETER,
		description = "Mean intensity for Gaussian length term")
	Double alpha = 0.0;

	@Parameter( label= "sigma", required = true, type = Parameter.Type.PARAMETER,
		description = "Standard deviation for Gaussian length term")
	Double sigma = 1.0;

	@Parameter( label= "kappa", required = true, type = Parameter.Type.PARAMETER,
		description = "Weight for curvature term")
	Double kappa = 1.0;

	@Parameter( label= "debug", required = true, type = Parameter.Type.PARAMETER,
		description = "Bit mask for debugging of levelset solver")
	Integer debug = 0;

    @Parameter( label= "numIterations", type = Parameter.Type.SUPPLEMENTAL,
                description = "Number of iterations performed")
    private Integer numIterations;

	@Parameter( label= "spacing", required = true, type = Parameter.Type.PARAMETER,
		description = "Spacing of intermediate results (none if spacing == 0) solver")
	Integer spacing = 0;

	@Parameter( label= "intermediateLS", required = false, type = Parameter.Type.SUPPLEMENTAL,
		description = "Vector of intermediate results")
	Vector intermediateLS;

	// ===============================================
    // ---- helper functions 

    /**
     * create a paramter object and set/initialize all objects
     */
    private LevelsetSolveNonPDE initSolver() throws ALDOperatorException {

		// initialize levelset function
        MTBLevelsetMembership ls = initLS();

		MTBGenericEnergyNonPDE e = null;
		if ( energytype == EnergyType.CHANVESE ) {
        	e = new MTBChanVeseEnergyNonPDE( inImg, ls, lambdabg, lambdafg, mu, nu);
		} else if ( energytype == EnergyType.CV_FITTING ) {
        	e = new MTBCVFittingEnergyNonPDE( inImg, ls, lambdabg, lambdafg);
		} else if ( energytype == EnergyType.CHANVESE_GAUSSIAN ) {
			MTBGenericEnergyNonPDE fittingEnergy = new MTBCVFittingEnergyNonPDE( inImg, ls, lambdabg, lambdafg);
			MTBGenericEnergyNonPDE lengthEnergy = new MTBGaussianLengthEnergyNonPDE( inImg, ls, mu, alpha,sigma);
			MTBGenericEnergyNonPDE sizeEnergy = new MTBSizeEnergyNonPDE( nu);
			Vector<MTBGenericEnergyNonPDE> energies = new Vector<MTBGenericEnergyNonPDE>(3);
			energies.addElement( fittingEnergy);
			energies.addElement( lengthEnergy);
			energies.addElement( sizeEnergy);
			e = new MTBEnergySumNonPDE( new String("Chan Vese with Gaussian length term"), energies);
		} else if ( energytype == EnergyType.CURVATURE ) {
			e = CurvatureEnergy.createEnergy( ls, kappa);
		} else if ( energytype == EnergyType.MEIJERING ) {
			e = new MTBMeijeringFittingEnergyNonPDE(inImg, ls);
		} else if ( energytype == EnergyType.MEANSEPARATION ) {
			e = new MTBLSMeanSeparationEnergy2DNonPDE(inImg, ls);
		} else if ( energytype == EnergyType.CV_FITTING_CURVATURE ) {
			MTBGenericEnergyNonPDE fittingEnergy = new MTBCVFittingEnergyNonPDE( inImg, ls, lambdabg, lambdafg);
			CurvatureEnergy curvatureEnergy = CurvatureEnergy.createEnergy( ls, kappa);
			Vector<MTBGenericEnergyNonPDE> energies = new Vector<MTBGenericEnergyNonPDE>(2);
			energies.addElement( fittingEnergy);
			energies.addElement( curvatureEnergy);
			e = new MTBEnergySumNonPDE( new String("Chan Vese with curvature"), energies);
		}
        e.setDebug( debug);

        // create parameter object
        LevelsetSolveNonPDE solver = new LevelsetSolveNonPDE( e, ls, maxIter, spacing, null, preserveTopology);

		solver.setVerbose( getVerbose());
		solver.setDebug( debug);

		return solver;
	}

    /**
     * Construct an initial levelset function 
	 * <p>
	 * If both initBinImg and initLabelImg are supplied initLabelImg is prefered for initialization
	 *
	 * @param img Image to segment
	 * @param initBinImg Binary image for initialization, each connected component will yield one phase
	 * @param initLabelImg Label image for initialization, labels will be used as phases
	 * @param invalidImg Image of invalid pixels, 
	 * @param threshold Threshold to apply to img for initialization
	 * @see de.unihalle.informatik.MiToBo.underdevelopment.ruttkies.segmentation.levelset.nonPDE.MTBLevelsetMembership
     */
	// TODO: consider invalidImg if intiImg==null

    private MTBLevelsetMembership initLS() {
        if ( initLabelImg != null )  {

            return new MTBLevelsetMembership( initLabelImg, invalidImg, multiphase);
        } else if ( initBinImg != null )  {
			if ( multiphase ) {
				try {
            		// component laebelling with 4-neighborhod
            		LabelComponentsSequential regLabler= new LabelComponentsSequential( initBinImg, false);
            		regLabler.runOp( this);
            		MTBRegion2DSet regions= regLabler.getResultingRegions();
	
            		return new MTBLevelsetMembership( initBinImg.getSizeX(), initBinImg.getSizeY(), regions, invalidImg, multiphase);
				} catch (ALDException e) {
					e.printStackTrace();
					return null;
				}
			} else {
            	return new MTBLevelsetMembership( initBinImg, invalidImg, multiphase);
			}
        } else if ( threshold != null ) {
            if( getMultiphase() ) {
				try {
                	ImgThresh thresholder = new ImgThresh(inImg, threshold);
                	thresholder.runOp(null);
                	MTBImage binaryImg = thresholder.getResultImage();
	
                	LabelComponentsSequential labeler = new LabelComponentsSequential(binaryImg, false);
                	labeler.runOp(null);
                	labeler.getResultingRegions().size();
	
                	MTBImage labelImg = labeler.getLabelImage();

                	return new MTBLevelsetMembership(labelImg, invalidImg, multiphase);
				} catch (ALDException e) {
					e.printStackTrace();
					return null;
				}
            } else {
                return new MTBLevelsetMembership(inImg.getSizeX(), inImg.getSizeY(), inImg.getSizeZ(), inImg, threshold.intValue(), invalidImg);
            }
        } else {
			return new MTBLevelsetMembership( inImg.getSizeX(), inImg.getSizeY(), inImg.getSizeZ(), invalidImg);
        }
    }

	/** Create a new image where all contour pixels of img are set to 1,
	 *  all other pixels to 0.
	 */
    private static MTBImage getContourImg( MTBImage img) {
		int deltaX[] = new int[] {0,1,1,1,0,-1,-1,-1};
		int deltaY[] = new int[] {-1,-1,0,1,1,1,0,-1};

		MTBImage tmpImg = img.duplicate();
        for ( int y = 0 ; y < tmpImg.getSizeY() ; y++ ) {
            for ( int x = 0 ; x < tmpImg.getSizeX() ; x++ ) {
				int val = img.getValueInt( x, y);
				tmpImg.putValueInt( x, y, 0);
				for ( int n=0 ; n<8 ; n++ ) {
					if ( 0 <= x+deltaX[n] && x+deltaX[n] < tmpImg.getSizeX()  &&
					     0 <= y+deltaY[n] && y+deltaY[n] < tmpImg.getSizeY()  &&
                         val != img.getValueInt( x+deltaX[n], y+deltaY[n]) ) {

						tmpImg.putValueInt( x, y, 1);
						
						break;
                	}
                }
			}
		}

		return tmpImg;
	}

	/** Draw contours of resultImg and if non null initLabelImg 
	 *  into inIMg and return this result in a new image
	 */
    private MTBImage drawContours( MTBImage inImg, MTBImage resultImg, MTBImage initLabelImg) {
		double[] minMaxVal = inImg.getMinMaxDouble();
		// scale intenisites of input inage to output range
		double scale = 0.9*minMaxVal[1]/(double)inImg.getTypeMax(); //minMaxVal[1];

		// first draw contours found in the result
		// with intensity: max * 0.9
		MTBImage ctrImg = getContourImg( resultImg);
        for ( int y = 0 ; y < ctrImg.getSizeY() ; y++ ) {
            for ( int x = 0 ; x < ctrImg.getSizeX() ; x++ ) {
				if (  ctrImg.getValueInt( x, y) == 0 )  
					ctrImg.putValueInt( x, y, (int)(scale*(double)(inImg.getValueInt( x, y))));
				else
					ctrImg.putValueInt( x, y, (int)(0.9*ctrImg.getTypeMax()));
			}
		}

		// if initLabelImg is non null, draw contours of initLabelImg
		// with intensity: max * 0.8
		if ( initLabelImg != null ) {
			MTBImage ctrImg2 = getContourImg( initLabelImg);

            for ( int y = 0 ; y < ctrImg.getSizeY() ; y++ ) {
                for ( int x = 0 ; x < ctrImg.getSizeX() ; x++ ) {
					if (  ctrImg2.getValueInt( x, y) > 0 ) 
						ctrImg.putValueInt( x, y, (int)(0.8*ctrImg.getTypeMax()));
				}
			}
		}


		return ctrImg;
	}

    @Override
    protected void operate() throws ALDOperatorException,ALDProcessingDAGException {

		// create solver and set parameter and input arguments 
        LevelsetSolveNonPDE solver = initSolver();

		if ( getVerbose() ) 
			solver.print();
        // optimize
        solver.runOp( this);

		// retrieve outputs of solver and set this operator outputs
		setNumIterations( solver.getNumIterations());
		setResultImage( solver.getResultImage());
		setIntermediateLS( solver.getIntermediateLS());

		MTBImage ctrImg = drawContours( inImg, solver.getResultImage(), getInitLabelImg());
		setResultContourImage( ctrImg);
    }

	/** Get value of debug.
	  * Explanation: Bit mask for debugging of levelset solver.
	  * @return value of debug
	  */
	public java.lang.Integer getDebug(){
		return debug;
	}

	/** Set value of debug.
	  * Explanation: Bit mask for debugging of levelset solver.
	  * @param value New value of debug
	  */
	public void setDebug( java.lang.Integer value){
		this.debug = value;
	}

	/** Get value of lambdafg.
	  * Explanation: Weight for foreground homogenity.
	  * @return value of lambdafg
	  */
	public java.lang.Double getLambdafg(){
		return lambdafg;
	}

	/** Set value of lambdafg.
	  * Explanation: Weight for foreground homogenity.
	  * @param value New value of lambdafg
	  */
	public void setLambdafg( java.lang.Double value){
		this.lambdafg = value;
	}

	/** Get value of alpha.
	  * Explanation: Mean intensity for Gaussian length term.
	  * @return value of alpha
	  */
	public java.lang.Double getAlpha(){
		return alpha;
	}

	/** Set value of alpha.
	  * Explanation: Mean intensity for Gaussian length term.
	  * @param value New value of alpha
	  */
	public void setAlpha( java.lang.Double value){
		this.alpha = value;
	}

    /** Get value of threshold.
      * Explanation: Maximal number of iterations for optimization.
      * @return value of threshold
      */
    public Double getThreshold(){
            return threshold;
    }

    /** Set value of threshold.
      * Explanation: Maximal number of iterations for optimization.
      * @param value New value of threshold
      */
    public void setThreshold( Double value){
        this.threshold = value;
    }


	/** Get value of energytype.
	  * Explanation: Energy type.
	  * @return value of energytype
	  */
	public EnergyType getEnergytype(){
		return energytype;
	}

	/** Set value of energytype.
	  * Explanation: Energy type.
	  * @param value New value of energytype
	  */
	public void setEnergytype( EnergyType value){
		this.energytype = value;
	}

	/** Get value of kappa.
	  * Explanation: Weight for curvature term.
	  * @return value of kappa
	  */
	public java.lang.Double getKappa(){
		return kappa;
	}

	/** Set value of kappa.
	  * Explanation: Weight for curvature term.
	  * @param value New value of kappa
	  */
	public void setKappa( java.lang.Double value){
		this.kappa = value;
	}

	/** Get value of maxIter.
	  * Explanation: Maximal number of iterations for optimization.
	  * @return value of maxIter
	  */
	public java.lang.Integer getMaxIter(){
		return maxIter;
	}

	/** Set value of maxIter.
	  * Explanation: Maximal number of iterations for optimization.
	  * @param value New value of maxIter
	  */
	public void setMaxIter( java.lang.Integer value){
		this.maxIter = value;
	}

    /** Get value of preserveTopology.
      * Explanation: Preserve topology during optimization.
      * @return value of preserveTopology
      */
    public java.lang.Boolean getPreserveTopology(){
            return preserveTopology;
    }

    /** Set value of preserveTopology.
      * Explanation: Preserve topology during optimization.
      * @param value New value of preserveTopology
      */
    public void setPreserveTopology( java.lang.Boolean value){
            this.preserveTopology = value;
    }

	/** Get value of lambdabg.
	  * Explanation: Weight for background homogenity.
	  * @return value of lambdabg
	  */
	public java.lang.Double getLambdabg(){
		return lambdabg;
	}

	/** Set value of lambdabg.
	  * Explanation: Weight for background homogenity.
	  * @param value New value of lambdabg
	  */
	public void setLambdabg( java.lang.Double value){
		this.lambdabg = value;
	}

	/** Get value of nu.
	  * Explanation: Weight for area term.
	  * @return value of nu
	  */
	public java.lang.Double getNu(){
		return nu;
	}

	/** Set value of nu.
	  * Explanation: Weight for area term.
	  * @param value New value of nu
	  */
	public void setNu( java.lang.Double value){
		this.nu = value;
	}

	/** Get value of spacing.
	  * Explanation: Spacing of intermediate results (none if spacing == 0) solver.
	  * @return value of spacing
	  */
	public java.lang.Integer getSpacing(){
		return spacing;
	}

	/** Set value of spacing.
	  * Explanation: Spacing of intermediate results (none if spacing == 0) solver.
	  * @param value New value of spacing
	  */
	public void setSpacing( java.lang.Integer value){
		this.spacing = value;
	}

	/** Get value of sigma.
	  * Explanation: Standard deviation for Gaussian length term.
	  * @return value of sigma
	  */
	public java.lang.Double getSigma(){
		return sigma;
	}

	/** Set value of sigma.
	  * Explanation: Standard deviation for Gaussian length term.
	  * @param value New value of sigma
	  */
	public void setSigma( java.lang.Double value){
		this.sigma = value;
	}

	/** Get value of multiphase.
	  * Explanation: Use multiphase levelsets.
	  * @return value of multiphase
	  */
	public java.lang.Boolean getMultiphase(){
		return multiphase;
	}

	/** Set value of multiphase.
	  * Explanation: Use multiphase levelsets.
	  * @param value New value of multiphase
	  */
	public void setMultiphase( java.lang.Boolean value){
		this.multiphase = value;
	}

	/** Get value of mu.
	  * Explanation: Weight for length term.
	  * @return value of mu
	  */
	public java.lang.Double getMu(){
		return mu;
	}

	/** Set value of mu.
	  * Explanation: Weight for length term.
	  * @param value New value of mu
	  */
	public void setMu( java.lang.Double value){
		this.mu = value;
	}

	/** Get value of invalidImg.
	  * Explanation: Image of invalid pixels (pixels <> 0 are invalid.
	  * @return value of invalidImg
	  */
	public de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage getInvalidImg(){
		return invalidImg;
	}

	/** Set value of invalidImg.
	  * Explanation: Image of invalid pixels (pixels <> 0 are invalid.
	  * @param value New value of invalidImg
	  */
	public void setInvalidImg( de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage value){
		this.invalidImg = value;
	}

	/** Get value of initLabelImg.
	  * Explanation: Label image for initialization.
	  * @return value of initLabelImg
	  */
	public de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage getInitLabelImg(){
		return initLabelImg;
	}

	/** Set value of initLabelImg.
	  * Explanation: Label image for initialization.
	  * @param value New value of initLabelImg
	  */
	public void setInitLabelImg( de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage value){
		this.initLabelImg = value;
	}

	/** Get value of inImg.
	  * Explanation: Input image to segment
	  * @return value of inImg
	  */
	public de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage getInImg(){
		return inImg;
	}

	/** Set value of inImg.
	  * Explanation: Input image to segment
	  * @param value New value of inImg
	  */
	public void setInImg( de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage value){
		this.inImg = value;
	}

	/** Get value of initBinImg.
	  * Explanation: Binary image for initialization.
	  * @return value of initBinImg
	  */
	public de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage getInitBinImg(){
		return initBinImg;
	}

	/** Set value of initBinImg.
	  * Explanation: Binary image for initialization.
	  * @param value New value of initBinImg
	  */
	public void setInitBinImg( de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage value){
		this.initBinImg = value;
	}

	/** Get value of ResultContourImage.
	  * Explanation: Resulting contour image.
	  * @return value of ResultContourImage
	  */
	public de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage getResultContourImage(){
		return ResultContourImage;
	}

	/** Set value of ResultContourImage.
	  * Explanation: Resulting contour image.
	  * @param value New value of ResultContourImage
	  */
	public void setResultContourImage( de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage value){
		this.ResultContourImage = value;
	}

	/** Get value of ResultImage.
	  * Explanation: Resulting label image.
	  * @return value of ResultImage
	  */
	public de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage getResultImage(){
		return ResultImage;
	}

	/** Set value of ResultImage.
	  * Explanation: Resulting label image.
	  * @param value New value of ResultImage
	  */
	public void setResultImage( de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage value){
		this.ResultImage = value;
	}

	/** Get value of numIterations
	  * Explanation: Number of iterations performed
	  * @return value of numIterations
	  */
	public Integer getNumIterations(){
		return numIterations;
	}

	/** Set value of numIterations
	  * Explanation: Number of iterations performed
	  * @param New value of numIterations
	  */
	public void setNumIterations( Integer value){
		this.numIterations = value;;
	}

	/** Get value of intermediateLS.
	  * Explanation: Vector of intermediate results.
	  * @return value of intermediateLS
	  */
	public java.util.Vector getIntermediateLS(){
		return intermediateLS;
	}

	/** Set value of intermediateLS.
	  * Explanation: Vector of intermediate results.
	  * @param value New value of intermediateLS
	  */
	public void setIntermediateLS( java.util.Vector value){
		this.intermediateLS = value;
	}

}
