/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5355 $
 * $Date: 2012-04-04 15:32:46 +0200 (Mi, 04. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.energies;

import java.awt.geom.Point2D;
import java.util.Vector;

import Jama.Matrix;



import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.ALDParametrizedClass;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.MiToBo.core.datatypes.MTBPolygon2DException;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.segmentation.activecontours.energies.MTBGenericEnergyPDE_CVRegionFit;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.MTBSnake;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSinglePDE;

/**
 * Snake energy based on Chan-Vese region fitting.
 * <p>
 * This energy models the inner and outer regions of a snake by a Gaussian 
 * model, i.e. each region is specified by a mean intensity value and 
 * deviations from this value are penalized.
 * 
 * @author moeller
 */
@ALDDerivedClass
@ALDParametrizedClass

public class MTBSnakeE_CVRegionFit 
	extends MTBGenericEnergyPDE_CVRegionFit implements MTBSnakeEnergyPDE {

	/**
	 * Target interval boundaries of energy.
	 * <p>
	 * targetEnergyRange[0] -> left boundary
	 * targetEnergyRange[1] -> right boundary
	 * targetEnergyRange[2] -> interval width
	 */
	private final double [] targetEnergyRange = {-1.0, 1.0, 2.0};
	
	/**
	 * Interval boundaries of adapted energy interval.
	 * <p>
	 * energyRange[0] -> left boundary
	 * energyRange[1] -> right boundary
	 * energyRange[2] -> interval width
	 */
	private double [] energyRange = {0.0, 0.0, 0.0};
	
	/**
	 * Width of working image.
	 */
	private transient int iWidth;
	
	/**
	 * Height of working image. 
	 */
	private transient int iHeight;
	
	/**
	 * Scaling factor to rescale image coordinates in range [0,1] or several
	 * ranges to the original coordinates range like [1000, 1000] in a image of
	 * size 1000 x 1000. Default scaling factor is 1.
	 */
	protected double scaleFactor = 1.0;

	/**
	 * Current snake the energy is based on.
	 */
	private transient MTBSnake currentSnk;
	
	/**
	 * Global exclude mask for image to be segmented.
	 */
	private transient boolean [][] excludeMask;
	
	/**
	 * Default constructor.
	 */
	public MTBSnakeE_CVRegionFit() {
		// nothing to do here
	}
	
  /**
   * Default constructor.
   * 
   * @param im	Image to work on.
   * @param l_in	Weighting factor for inner region fit.
   * @param l_out	Weighting factor for outer region fit.
   * @throws ALDOperatorException 
   * @throws ALDProcessingDAGException 
   */
  public MTBSnakeE_CVRegionFit(MTBImage im, double lin, double lout) {
  	super(im,lin,lout);
  }
  
	@Override
  public boolean initEnergy(SnakeOptimizerSinglePDE opt) {
		if (this.inImg == null)
			this.inImg = opt.getInputImage();
		try {
	    super.init();
    } catch (ALDOperatorException e) {
	    // TODO Auto-generated catch block
	    e.printStackTrace();
    }
  	this.iWidth = this.inImg.getSizeX();
  	this.iHeight = this.inImg.getSizeY();
		opt.requireCounterClockwiseSnakePointOrder();
		opt.requireOverlapMask();
		// get energy minimum and maximum;
		// minimum of Chan-Vese energy is always smaller than zero and 
		// maximum always larger
		double eMin = this.getEnergyMinVal().doubleValue();
		double eMax = this.getEnergyMaxVal().doubleValue();
		if (Math.abs(eMin) > Math.abs(eMax)) {
			this.energyRange[0] = - Math.abs(eMin);
			this.energyRange[1] =   Math.abs(eMin);
		}
		else {
			this.energyRange[0] = - Math.abs(eMax);
			this.energyRange[1] =   Math.abs(eMax);
		}
		this.energyRange[2] = this.energyRange[1] - this.energyRange[0];
	  return true;
  }

	@Override
  public void updateStatus(SnakeOptimizerSinglePDE opt) {
		this.currentSnk = opt.getCurrentSnake();
		this.excludeMask = opt.getExcludeMask();
		// set excluded pixels invalid - if there are some...
    boolean [][] vMask = new boolean[this.iHeight][this.iWidth];
    if (this.excludeMask != null) {
    	for (int y=0;y<this.iHeight;++y) {
    		for (int x=0;x<this.iWidth;++x) {
    			vMask[y][x] = !this.excludeMask[y][x];
    		}
    	}
    }
    else {
    	// all pixels are valid, non is excluded
    	for (int y=0;y<this.iHeight;++y) {
    		for (int x=0;x<this.iWidth;++x) {
    			vMask[y][x] = true;
    		}
    	}    	
    }
		// fetch current overlap mask
    if (opt.getSnakeNumber() > 1) {
    	int[][] overlapMask = opt.getOverlapMask();
    	int[][] snakeMask;
    	snakeMask = this.currentSnk.getBinaryMask(opt.getInputImage().getSizeX(),
    			opt.getInputImage().getSizeY());	    
    	for (int y=0;y<this.iHeight;++y) {
    		for (int x=0;x<this.iWidth;++x) {
    			if (snakeMask[y][x] == 0 && overlapMask[y][x] != 0)
    				vMask[y][x] = false;
    		}
    	}
    }
		this.currentSnk.setValidmask(vMask);
		// do the actual update
		this.updateParameters(this.currentSnk);
	}
	
  /**
   * Updates the fitting term based on new region average values.
   */
	@Override
  public Matrix getLinearMatrix(SnakeOptimizerSinglePDE opt) {
		
		int snakePointNum = this.currentSnk.getPointNum();
		Vector<Point2D.Double> polyPoints = this.currentSnk.getPoints();
		
		Matrix A = new Matrix(snakePointNum * 2, snakePointNum * 2);
		for (int counter = 0; counter < snakePointNum - 1; ++counter) {
			Point2D.Double p = polyPoints.get(counter);
			double eDerive = this.getDerivative(this.currentSnk,
					                                (int)(p.x*this.scaleFactor), 
										 	 			 					    (int)(p.y*this.scaleFactor),0);
			// calculate normalized final derivative value
			double tau_i = (eDerive - this.energyRange[0])/this.energyRange[2]                                              
				* this.targetEnergyRange[2] + this.targetEnergyRange[0];
//			double oldVal= A.get(counter, snakePointNum + counter);
			A.set(counter, snakePointNum + counter, - tau_i);
//			oldVal= A.get(counter, snakePointNum + counter + 1);
			A.set(counter, snakePointNum + counter + 1, tau_i);
//			oldVal= A.get(snakePointNum + counter, counter);
			A.set(snakePointNum + counter, counter, tau_i);
//			oldVal= A.get(snakePointNum + counter, counter + 1);
			A.set(snakePointNum + counter, counter + 1, - tau_i);
		}
		// add last point
		Point2D.Double p = polyPoints.get(snakePointNum - 1);
		double eDerive = this.getDerivative(this.currentSnk,
				                                (int)(p.x*this.scaleFactor), 
				                                (int)(p.y*this.scaleFactor), 0);
		// normalize to range [-1,1]
		double tau_i = (eDerive - this.energyRange[0])/this.energyRange[2]                                              
			* this.targetEnergyRange[2] + this.targetEnergyRange[0];
		// fill matrix - upper right block
//		double oldVal = A.get(snakePointNum - 1, snakePointNum);
		A.set(snakePointNum - 1, snakePointNum, tau_i);
//		oldVal = A.get(snakePointNum - 1, snakePointNum * 2 - 1);
		A.set(snakePointNum - 1, snakePointNum * 2 - 1, - tau_i);
		// fill matrix - lower left block
//		oldVal = A.get(2 * snakePointNum - 1, 0);
		A.set(2 * snakePointNum - 1, 0, - tau_i);
//		oldVal = A.get(2 * snakePointNum - 1, snakePointNum - 1);
		A.set(2 * snakePointNum - 1, snakePointNum - 1, tau_i);
		return A;
	}

	@Override
  public Matrix getConstVector(SnakeOptimizerSinglePDE opt) {
	  return null;
  }

  @Override
  public double calcEnergy(SnakeOptimizerSinglePDE opt) {
  	
  	// TODO only partially implemented!
  	
  	MTBSnake snake= opt.getCurrentSnake();
  	
    // calculate region means, if not done before
//    if (this.c_in < 0 || this.c_out < 0)

    // result values
    double lengthTerm = 0;
    double curvatureTerm = 0;
    double regionHomoTerm = 0;

    // iterate over all points of the snake
//    Vector<Point2D.Double> spoints = snake.getPoints();
//    int pointNum = snake.getPointNum();

//    double x_p, y_p, x_i, y_i, x_n, y_n;
//    for (int i = 1; i < pointNum - 1; ++i) {
//
//      x_p = spoints.get(i - 1).x;
//      y_p = spoints.get(i - 1).y;
//      x_i = spoints.get(i).x;
//      y_i = spoints.get(i).y;
//      x_n = spoints.get(i + 1).x;
//      y_n = spoints.get(i + 1).y;

//      // length term
//      lengthTerm += alpha
//          * ((x_n - x_i) * (x_n - x_i) + (y_n - y_i) * (y_n - y_i));
//
//      // curvature term
//      curvatureTerm += beta
//          * ((x_n - 2 * x_i + x_p) * (x_n - 2 * x_i + x_p) + (y_n - 2 * y_i + y_p)
//              * (y_n - 2 * y_i + y_p));
//
//    }

    double c_in = this.means[1];
    double c_out = this.means[0];
    
    // region homogenity term
    double sumInside = 0, sumOutside = 0;
    for (int y = 0; y < this.iHeight; ++y) {
      for (int x = 0; x < this.iWidth; ++x) {
        // check if image point lies inside or outside
        int value = this.inImg.getValueInt(x, y);
        try {
          if (!snake.jni_containsPoint(x / this.scaleFactor, y
              / this.scaleFactor)) {
            sumInside += (value - c_in) * (value - c_in);
          } else {
            sumOutside += (value - c_out) * (value - c_out);
          }
        } catch (MTBPolygon2DException e) {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
    }
    regionHomoTerm = this.lambda_in * sumInside + this.lambda_out * sumOutside;
    return 0.5 * (lengthTerm + curvatureTerm) + regionHomoTerm;
  }

	@Override
  public double calcEnergy(SnakeOptimizerSinglePDE opt, int pointID) {
	  // TODO Auto-generated method stub
	  return 0;
  }

	/**
	 * Set the scaling factor.
	 * 
	 * @param s
	 *          new scaling factor.
	 */
	@Override
  public void setScaleFactor(double s) {
		this.scaleFactor = s;
	}

	/**
	 * Get scaling factor.
	 * 
	 * @return Scaling factor.
	 */
	@Override
  public double getScaleFactor() {
		return this.scaleFactor;
	}

	@Override
  public String toString() {
		return new String("MTBSnkEnergyPDE: Chan-Vese Region Fitting " + 
				"(lambda_in = " + this.lambda_in + 
				" , lamdba_out = " + this.lambda_out + ")");
  }

}
