/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5418 $
 * $Date: 2012-04-12 11:16:29 +0200 (Do, 12. Apr 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.optimize;

import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDOperatorException.*;


import de.unihalle.informatik.Alida.annotations.Parameter;
import de.unihalle.informatik.Alida.annotations.Parameter.ExpertMode;
import de.unihalle.informatik.MiToBo.core.datatypes.MTBPolygon2DSet;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImageByte;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImage.*;
import de.unihalle.informatik.MiToBo.core.datatypes.images.MTBImageRGB;
import de.unihalle.informatik.MiToBo.segmentation.snakes.datatypes.*;
import de.unihalle.informatik.MiToBo.tools.system.UserTime;

/**
 * <pre>
 * Image contour segmentation using parametric snakes.
 * </pre>
 * 
 * @author moeller, misiak
 */

public abstract class SnakeOptimizerSingle extends SnakeOptimizer {

	/**
	 * Target length of snake segments in resampling.
	 */
	@Parameter(label="Resample Segment Length", mode=ExpertMode.ADVANCED,
			direction=Parameter.Direction.IN, required=false, dataIOOrder = 11,
			description = "Desired resampling segment length.")
  protected Double resampleSegLength= new Double(5.0);

	/**
	 * Flag to enable/disable snake resampling.
	 */
	@Parameter(label="Do Resampling", mode=ExpertMode.ADVANCED,
			direction=Parameter.Direction.IN, required=false, dataIOOrder = 10,
			description = "Flag for snake resampling.")
	protected Boolean	doResampling= new Boolean(true);

  /*
   * Some internal convenience variables.
   */
  
	/**
	 * The snake on which the algorithm works.
	 */
	protected transient MTBSnake snake = null;

	/**
	 * Old snake of previous iteration (for evolution assessment).
	 */
	protected transient MTBSnake previousSnake = null;

	/**
	 * Flag for invoking energy calculations.
	 */
	protected transient boolean energyCalculationRequested = false;

	/**
	 * Current energy of the snake.
	 */
	protected transient double energy = -1.0;

	/**
	 * Snake energy in previous iteration (for change assessment).
	 */
	protected transient double previousEnergy = -1.0;

	/**
	 * Scale factor for snake normalization.
	 */
	protected transient double scaleFactor = 1.0;

  /**
   * Flag to indicate whether an energy requested overlap mask.
   */
  protected transient boolean overlapMaskRequested = false;
  
	/**
	 * Mask to indicate overlap regions between snakes.
	 */
	protected transient int[][] overlapMask = null;
	
//	/**
//	 * Flag for invoking energy calculations.
//	 */
//	protected boolean energyCalculationRequested = false;

	/************************************************************************/
	/*** Internal helpers for optimization procedure. ***/
	/************************************************************************/

//	/**
//	 * List of normalized energy weights.
//	 */
//	protected double [] energyWeightsNormed;

//	protected double minEnergy = Double.MAX_VALUE;
//	protected double minEnergyOld = Double.MAX_VALUE;

  /**
   * Timer object for time measurements.
   */
	protected transient UserTime timer = new UserTime();
  
  /**
   * Image for showing/saving intermediate results.
   */
  protected transient MTBImageRGB dispImg = null;
  
  /***********************************************************************/
  /*** Constructors and clone()-methods, validation routines. ***/
  /***********************************************************************/

	/**
	 * Default constructor.
	 * 
	 * @throws ALDOperatorException
	 */
	public SnakeOptimizerSingle() throws ALDOperatorException {
		// nothing to do here
	}
	
  /***********************************************************************/
  /*** Helper method for object serialization.                         ***/
  /***********************************************************************/
	
	/**
	 * Function for proper initialization of deserialized objects.
	 * <p>
	 * This function is called on an instance of this class being deserialized
	 * from file, prior to handing the instance over to the user. It takes care
	 * of a proper initialization of transient member variables as they are not
	 * initialized to the default values during deserialization. 
	 * @return
	 */
	private Object readResolve() {
		System.out.println("ReadResolve of SnakeOptimizerSingle called..");
		// super class members
		this.snakeNum= 1;
		this.itCounter = 0;
		this.counterClockwiseSnakePointOrderRequested = false;
		// members of this class
		this.energyCalculationRequested = false;
		this.energy = -1.0;
		this.previousEnergy = -1.0;
		this.scaleFactor = 1.0;
		this.overlapMaskRequested = false;
		this.timer = new UserTime();
		return this;
	}

  /***********************************************************************/
  /*** (Abstract) methods for operator functionality. ***/
  /***********************************************************************/

  @SuppressWarnings("unused")
  @Override
	protected void initOptimizer()
	throws MTBSnakeOptimizerException, ALDOperatorException {

		// image is required and should not be null
		if (this.inImg == null) {
			throw new ALDOperatorException(OperatorExceptionType.VALIDATION_FAILED,
			"SnakeOptimizerSingle - no image given!");
		}
		this.iHeight = this.inImg.getSizeY();
		this.iWidth = this.inImg.getSizeX();

		// init the output image
		if (   this.saveIntermediateResults.booleanValue() 
				|| this.showIntermediateResults.booleanValue()) {
			// initialize result image
			this.dispImg = (MTBImageRGB) MTBImage.createMTBImage(this.iWidth,
					this.iHeight, 1, 1, 1, MTBImageType.MTB_RGB);
		}
	}
	
  /**
   * Display intermediate results by overlaying the input image with 
   * the current snake result. 
   */
  @Override
  protected void showSnake() {
  	// convert input image to byte format
  	MTBImageByte background = 
  		(MTBImageByte)this.inImg.convertType(MTBImageType.MTB_BYTE, true);
  	
  	// plot it to the background of the disp image
  	int gray = 0, color = 0;
  	for (int y = 0; y < this.inImg.getSizeY(); ++y) {
  		for (int x = 0; x < this.inImg.getSizeX(); ++x) {
  			gray = background.getValueInt(x, y);
  			color = ((gray & 0xff) << 16) + ((gray & 0xff) << 8) + (gray & 0xff);
  			this.dispImg.putValueInt(x, y, color);
  		}
  	}
  	// plot the snake to the image
		this.snake.drawPolygon(this.dispImg);
		this.dispImg.updateAndRepaintWindow();
  	this.dispImg.show();
		this.outSnakesImg = (MTBImageRGB)this.dispImg.duplicate();
  }

  @Override
  protected void closeWindows() {
  	if (this.dispImg != null)
  		this.dispImg.close();
  }

  /**
   * Save intermediate results. 
   */
  @Override
  protected abstract void saveSnake();

	@Override
  public void printParams() {
		System.out.println("Snake number= " + this.snakeNum);
		/*
		 * TODO parameter wieder anzeigen ! Testen ob vorhanden oder nicht wird dazu
		 * noch nötig sein. Alle Terminations haben eine get methode für die params
		 * aus dem Konstruktor um maxIterations und die AreaFraction usw abzufragen,
		 * diese kann man dann ausgeben.
		 */
		// System.out.println("Iterations= " + this.getMaxIterations());
		// System.out.println("Min. energy improve= " + this.getEnergyImprove());
		// System.out.println("Point Fraction= " + this.getPointFraction());
		// System.out.println("Area Fraction= " + this.getAreaFraction());
		System.out.println("Resample snake= " + this.doResampling);
		System.out.println("Resampling segment length= " + this.resampleSegLength);
	}

  @Override
  public String toString() {
  	return new String("SnakeOptimizerSingle - abstract base class object");
  }

	/**
	 * Returns a set with current snake(!).
	 * <p>
	 * In case of the SnakeOptimizerSingle class this set will always have 
	 * exactly one element.
	 */
	@Override
  public MTBPolygon2DSet getCurrentSnakes() {
		if (this.snake == null)
			return null;
		MTBPolygon2DSet pset = new MTBPolygon2DSet(0,0,this.iWidth,this.iHeight);
		pset.add(this.snake.clone());
		return pset;
	}

  /***********************************************************************/
  /*** Getters and Setters. ***/
  /***********************************************************************/

	/**
	 * Enable energy calculation in each iteration.
	 */
	public void enableEnergyCalculation() {
		this.energyCalculationRequested = true;
	}
	
	/**
	 * Disable energy calculations completely.
	 */
	public void disableEnergyCalculation() {
		this.energyCalculationRequested = false;
	}

	/**
	 * Returns a copy of the current snake.
	 * @return Current snake.
	 */
  public MTBSnake getCurrentSnake() {
		if (this.snake == null)
			return null;
		return this.snake.clone();
	}

	/**
	 * Returns the previous snake.
	 * @return Snake contour of previous iteration.
	 */
	public MTBSnake getPreviousSnake() {
		if (this.previousSnake == null)
			return null;
		return this.previousSnake.clone();
	}

	/**
	 * Get current energy value for the snake.
	 * @return Energy value of the snake.
	 */
	public double getEnergyValue() {
		return this.energy;
	}

	/**
	 * Get old energy value for the snake at last iteration.
	 * @return Old snake energy value of the previous iteration.
	 */
	public double getPreviousEnergyValue() {
		return this.previousEnergy;
	}

	/**
	 * Returns desired segment length for resampling.
	 * @return Desired snake segment length.
	 */
	public Double getSegmentLength() {
		return this.resampleSegLength;
	}

	/**
	 * Returns true if snake should be resampled during optimization.
	 * @return True, if snake is to be resampled.
	 */
	public Boolean doResampleSnake() {
		return this.doResampling;
	}

	/**
	 * Specify new overlap mask.
	 */
	public void setOverlapMask(int [][] m) {
		this.overlapMask = m.clone();
	}

	/**
	 * Returns current overlap mask.
	 * @return Overlap mask when processing multiple snakes.
	 */
	public int [][] getOverlapMask() {
		if (this.overlapMask != null)
			return this.overlapMask.clone();
		return null;
	}

	/**
	 * Intended for energies to request overlap mask.
	 */
	public void requireOverlapMask() {
		this.overlapMaskRequested= true;
	}
	
	/**
	 * Returns true, if optimizer requires overlap mask.
	 * @return True, is any energy requires an overlap mask.
	 */
	public boolean doesRequireOverlapMask() {
		return this.overlapMaskRequested;
	}
	
	/***********************************************************************/
	/*** Protected routines. ***/
	/***********************************************************************/

	/**
	 * Resamples the snake using the specified segment length.
	 * <p>
	 * Note: if snake is normalized, segment length will also be normalized!
	 * Normalization of segment length is done by resampling!
	 */
	protected void resampleSnake() {
		this.snake.resample(this.resampleSegLength.doubleValue());
	}
}
