/*
 * This file is part of MiToBo, the Microscope Image Analysis Toolbox.
 *
 * Copyright (C) 2010 - 2012
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Fore more information on MiToBo, visit
 *
 *    http://www.informatik.uni-halle.de/mitobo/
 *
 */

/* 
 * Most recent change(s):
 * 
 * $Rev: 5236 $
 * $Date: 2012-03-22 13:07:18 +0100 (Do, 22. Mär 2012) $
 * $Author: moeller $
 * 
 */

package de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.termination;



import de.unihalle.informatik.Alida.annotations.ALDClassParameter;
import de.unihalle.informatik.Alida.annotations.ALDDerivedClass;
import de.unihalle.informatik.Alida.annotations.ALDParametrizedClass;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizerSingle;
import de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.SnakeOptimizer.Snake_status;

/**
 * Check upon snake termination by energy changes. 
 * <p>
 * Termination is done when the energy difference of the former and the current
 * snake is below a given fraction factor.
 * 
 * @author misiak
 */
@ALDDerivedClass
@ALDParametrizedClass

public class MTBTermEnergyDiff extends MTBTermination {

	/**
	 * Energy fraction factor.
	 */
	@ALDClassParameter(label="Energy Fraction")
	private double energyFraction;

	/**
	 * Default constructor.
	 */
	public MTBTermEnergyDiff() {
		// nothing to do here
	}

	/**
	 * Constructor with energy fraction factor below the snake is terminated.
	 */
	public MTBTermEnergyDiff(double _energyFraction) {
		this.energyFraction = _energyFraction;
	}

	@Override
	public boolean init(SnakeOptimizerSingle opt) {
		this.optimizer = opt;
		this.optimizer.enableEnergyCalculation();
		return true;
	}

	/**
	 * Termination is done when the energy difference of the former and the
	 * current snake is below a given fraction factor.
	 */
	@Override
	public Snake_status terminate() {
		double energyVal = this.optimizer.getEnergyValue();
		double oldEnergyVal = this.optimizer.getPreviousEnergyValue();
		/*
		 * Calculate current energy changes between former and current snake
		 * optimization step.
		 */
		System.out.println("    Stop at <= " + (this.energyFraction * 100)
				+ " % of the energy difference.");
		if (oldEnergyVal == -1.0)
			System.out.println("    Energy difference = NA in first iteration");
		else
			System.out.println("    Energy difference = "
					+ Math.abs((energyVal - oldEnergyVal) / oldEnergyVal));
		if (Math.abs((energyVal - oldEnergyVal) / oldEnergyVal) <= this.energyFraction) {
			System.out.println("\n--- Snake optimizer stopped: Snake done after "
					+ this.optimizer.getIterationCount()
					+ " step(s) [minimal energy difference reached].\n");
			return SnakeOptimizerSingle.Snake_status.SNAKE_DONE;
		}
		return SnakeOptimizerSingle.Snake_status.SNAKE_SUCCESS;
	}

	/**
	 * Get energy fraction factor.
	 * 
	 * @return Energy fraction factor.
	 */
	public double getEnergyFraction() {
		return this.energyFraction;
	}

	@Override
	public String toString() {
		return ("MTBTermEnergyDiff");
	}

	/**
	 * Clones this object.
	 * <p>
	 * Note that the {@link SnakeOptimizerSingle} object is left null! 
	 * You have to call {@link SnakeOptimizer.initOptimizer()} before 
	 * using the gamma updater.
	 */
	/* (non-Javadoc)
	 * @see de.unihalle.informatik.MiToBo.segmentation.snakes.optimize.stepsize.MTBGammaUpdate#clone()
	 */
	@Override
	@SuppressWarnings("unused")
	public MTBTermEnergyDiff clone() throws CloneNotSupportedException {
		MTBTermEnergyDiff newObj = 
			new MTBTermEnergyDiff(this.energyFraction);
		newObj.status = this.status;
		newObj.optimizer = null;
		return newObj;
	}
}
