package mtb_cell_motility;



import de.unihalle.informatik.Alida.exceptions.ALDOperatorException;
import de.unihalle.informatik.Alida.exceptions.ALDProcessingDAGException;
import de.unihalle.informatik.MiToBo.apps.scratchAssay.ScratchAssayAnalyzer;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.GenericDialog;
import ij.io.FileInfo;
import ij.io.OpenDialog;
import ij.plugin.PlugIn;

/**
 * class for analyzing several scratch assay images stored in the same directory
 * 
 * @author glass
 *
 */

public class ScratchAssay_Analysis implements PlugIn
{
	private int sigma;				// standard deviation of gauss filter
	private int entropyFilterSize;	// size of mask for entropy filtering
	private int closingMaskSize = 15; // size of mask for morphological closing
	private int maxIter;			// maximum number of iterations for level set segmentation
	private boolean isHorizontal;	// horizontally or vertically oriented scratch
	private boolean doBatch;		// analyse all images in the given directory
	private boolean noCheck;		// don't check for scratch presence prior to segmentation
	private boolean verbose;		// show additional information and intermediate results 
	private boolean silent;			// save results without user interaction
	private String svmFile = null;	// path to external SVM model file
	
	@Override
	public void run(String arg0)
	{
		ImagePlus imp = WindowManager.getCurrentImage();
		String directory = null;
		String fileName = null;
		
		if(imp != null)
		{
			FileInfo fileInfo = imp.getOriginalFileInfo();
			
			if(fileInfo != null)
			{
				directory = fileInfo.directory;
				fileName = fileInfo.fileName;
				
				if(directory == "" || fileName == "")
				{
					
					OpenDialog od = new OpenDialog("select first image","");
					directory = od.getDirectory();
					fileName = od.getFileName();
				}
			}
			else
			{
				OpenDialog od = new OpenDialog("select first image","");
				directory = od.getDirectory();
				fileName = od.getFileName();
			}
			
		}
		else
		{
			OpenDialog od = new OpenDialog("select first image","");
			directory = od.getDirectory();
			fileName = od.getFileName();
			
		}
		
		if(directory != null && fileName != null && parameterDialog())
		{
			
//			System.out.println("directory plugin: " + directory);
//			System.out.println("fileName plugin: " + fileName);
			
			try
			{
				String path = directory + fileName;
				
//				ScratchAssayAnalyzer analyzer = new ScratchAssayAnalyzer(path, sigma,
//																entropyFilterSize, closingMaskSize, maxIter,
//																isHorizontal, noCheck, doBatch);
				ScratchAssayAnalyzer analyzer = new ScratchAssayAnalyzer(path, sigma, entropyFilterSize,
																		 isHorizontal, noCheck, doBatch, maxIter);
				
				analyzer.setVerbose(verbose);
				analyzer.isSilent(silent);
				
				if(svmFile != null)
				{
					analyzer.setSVMFile(svmFile);
				}
				
				analyzer.runOp(null);
			} 
			catch(ALDOperatorException e)
			{
				e.printStackTrace();
			}
			catch(ALDProcessingDAGException e)
			{
				e.printStackTrace();
			}
			
		}
	}

	/**
	 * opens a dialog that queries for the required parameters
	 */
	private boolean parameterDialog()
	{
		String[] scratchOrientation = {"horizontal", "vertical"};
		
		GenericDialog gd = new GenericDialog("settings");
		gd.addChoice("scratch orientation", scratchOrientation, scratchOrientation[0]);
		gd.addNumericField("size of gauss filter (\u03c3)", 5, 0);
		gd.addNumericField("size of entropy filter", 31, 0);
		//gd.addNumericField("size of closing mask", 15, 0);
		gd.addNumericField("maximum number of iterations", 5000, 0);
		gd.addCheckbox("don't check for scratch presence", false);
		gd.addCheckbox("analyze all images in the current directory", true);
		gd.addCheckbox("use external svm model file", false);
		gd.addCheckbox("silent", true);
		gd.addCheckbox("verbose", false);
		gd.showDialog();
		
		if(gd.wasOKed())
		{
			String orientation = gd.getNextChoice();
			this.sigma = (int)gd.getNextNumber();
			this.entropyFilterSize = (int)gd.getNextNumber();
			//this.closingMaskSize = (int)gd.getNextNumber();
			//this.thresh = gd.getNextNumber() / 100;
			this.maxIter = (int)gd.getNextNumber();
			this.noCheck = gd.getNextBoolean();
			this.doBatch = gd.getNextBoolean();
			boolean useExtSVM = gd.getNextBoolean();
			
			if(useExtSVM)
			{
				OpenDialog od = new OpenDialog("open svm model file","");
				this.svmFile = od.getDirectory() + od.getFileName();
			}
			
			this.silent = gd.getNextBoolean();
			this.verbose = gd.getNextBoolean();
			
			if(orientation.equalsIgnoreCase(scratchOrientation[0]))	// horizontal scratch
			{
				this.isHorizontal = true;
			}
			else
			{
				this.isHorizontal = false;
			}
			
			return true;
		}
		else
		{
			return false;
		}
	}
}
